/*=======================================================
  Author:   R. Muth
  Version:  $Id: bbl.layout.c,v 1.28 1999/08/06 03:46:29 muth Exp $
  Comment:  This file contains the basic block layout stuff
=======================================================*/

#include "global.h"
#include "functions.h"
#include "flags.h"
#include "opcode.h"

/*=======================================================
=======================================================*/
extern INT64 AbblWeight(INDEX ibbl)
{
    return ABBL_COUNT(ibbl) * AbblNInss(ibbl);
}

/*=======================================================
=======================================================*/

static INDEX *SortedEdges = 0;

#define SORTED_EDGE(i)   SortedEdges[i]
#define AEDG_VALUE(i) AEDG_U1(i).quadword

#define FE_ONLY_IN    F_LOCAL_1
#define FE_ONLY_OUT   F_LOCAL_2
#define FE_FIXED      F_LOCAL_3


#define ABBL_CHAIN_MAX(i) ABBL(i).u.quadword
#define AFUN_BBL_HOT(i)   AFUN_BBL_FIRST(i)
#define AFUN_BBL_REST(i)  AFUN_BBL_LAST(i)
#define AFUN_BBL_ZERO(i)  AFUN(i).u.doubleword.d1


#define ABBL_CHAIN_WEIGHT(ibbl) ABBL_U1(ibbl).quadword

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "InitChain"

static void InitChain()
{
    /* the flag fiels are the set to zero by calloc */
    CASSERT( SortedEdges == 0,("weird\n"));
    SortedEdges = (INDEX *)calloc(CONFIG_UEDG,sizeof(INDEX));
    CASSERT( SortedEdges != 0,("Cannot Allocate Memory For SortedEdges\n"));
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ExitChain"

static void ExitChain()
{
    CASSERT( SortedEdges != 0,("weird\n"));
    free(SortedEdges);
    SortedEdges = 0;
}

/*=======================================================
@FUN
If there is a fallthru edge (ET_NORMAL or ET_FALSE) between
to nonadjacent bbls we need to an unconoditional branch

@END
=======================================================*/
#undef __PROC__
#define __PROC__ "LayoutCreateFallthruBbl"

static INDEX LayoutCreateFallthruBbl(INDEX iedg)
{
    INDEX hd = AEDG_HEAD(iedg);
    INDEX tl = AEDG_TAIL(iedg);
    INDEX ft;
    
    INDEX jedg;
    INDEX iins;

    if( ABBL_TYPE(hd) != BT_NORM )
    {
        ft = AbblNew();
        ABBL_COUNT(ft) = AEDG_COUNT(iedg);
        ABBL_FUN(ft) = ABBL_FUN(hd);
        jedg = AedgNewFancy(hd,ft,AEDG_TYPE(iedg));
        AEDG_COUNT(jedg) = AEDG_COUNT(iedg);
        AedgKillFancy(iedg);
        AedgLinkFancy(jedg);
    }
    else
    {
        AedgKillFancy(iedg);
        ft = hd;
    }
    
    ABBL_TYPE(ft) = BT_UNCOND;

    iins = AinsNew();
    AinsMakeBr(iins,0);
    AINS_BBL(iins) = ft;
    AinsAppend(iins);

    jedg = AedgNewFancy(ft,tl,ET_UNCOND);
    AEDG_COUNT(jedg) = AEDG_COUNT(iedg);
    AedgLinkFancy(jedg);
    return ft == hd ? INVALID : ft;
}

/*=======================================================
=======================================================*/

static void PrintList(INDEX ibbl)
{
    INDEX hd = ibbl;
    INDEX nhot = 0;
    INDEX ntot = 0;


    while( VALID(ibbl) )
    {
        if( ABBL_IS_HOT(ibbl) )
            nhot += AbblNInss(ibbl);

        ntot += AbblNInss(ibbl);
        ibbl = ABBL_NEXT(ibbl);
    }

    printf("## %5d %5d/%5d  ",hd,nhot,ntot);

    ibbl = hd;
    
    while( VALID(ibbl) )
    {
        if( ABBL_COUNT(ibbl) < 0 )
            fputc('I',stdout);
        else if( ABBL_IS_ZERO(ibbl)  )
            fputc('0',stdout);
        else if( ABBL_IS_HOT(ibbl) )
            fputc('*',stdout);
        else
            fputc('-',stdout);

        ibbl = ABBL_NEXT(ibbl);
    }
    
    printf("\n");
}

/*=======================================================
=======================================================*/

extern void PrintLayout(FILE *fp,INDEX xfun)
{
    INDEX ibbl;
    NUMBER count=0;

    FOREACH_FUN_BBL(xfun,ibbl)
    {
        if( AbblNInss(ibbl) == 0 ) continue;
        
        if( ABBL_COUNT(ibbl) < 0 )
            fputc('I',fp);
        else if( ABBL_TYPE(ibbl) == BT_UNCOND )
            fputc('U',fp);
        else if( ABBL_IS_ZERO(ibbl) )
            fputc('0',fp);
        else if( ABBL_IS_HOT(ibbl) )
            fputc('*',fp);
        else
            fputc('-',fp);
        count++;
        if( (count % 96) == 0 )
            fprintf(fp,"\n");
    }


    fprintf(fp,"\n");
    InfoStatus(fp,"\t[PrintLayout] bbls: %d\n",count);

}

/*=======================================================
=======================================================*/

extern void PrintChains(FILE *fp)
{
    INDEX ibbl;
    INDEX jbbl;
    INDEX ifun;
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( VALID(ABBL_PREV(ibbl)) ) continue;
        ifun = ABBL_FUN(ibbl);
        
        for(jbbl=ibbl; VALID(jbbl); jbbl = ABBL_NEXT(jbbl) )
        {
            if( AbblNInss(jbbl) == 0 ) continue;
        
            if( ABBL_COUNT(jbbl) < 0 )
                fputc('I',fp);
            else if( ABBL_TYPE(jbbl) == BT_UNCOND )
                fputc('U',fp);
            else if( ABBL_IS_ZERO(jbbl) )
                fputc('0',fp);
            else if( ABBL_IS_HOT(jbbl) )
                fputc('*',fp);
            else
                fputc('-',fp);
        }

        fprintf(fp," %d [%s]\n",ifun,AFUN_NAME(ifun));        
    }

    fprintf(fp,"\t[PrintChains] \n");

}

/*=======================================================
We try to group bbl from the same function next to each other.
=======================================================*/
#undef __PROC__
#define __PROC__ "LayoutBuildFunctionChain"

static void LayoutBuildFunctionChain(FILE *fp,BOOL fancy)
{
    INDEX ibbl;
    INDEX jbbl;
    INDEX ifun;

    INDEX first;
    INDEX last;
    INT64 max;
    INT64 min;
    
    clock_t time = clock();    
    
    FOREACH_LIVE_FUN(ifun)
    {
        AFUN_BBL_ZERO(ifun) = INVALID;
        AFUN_BBL_HOT(ifun) = INVALID;
        AFUN_BBL_REST(ifun) = INVALID;
    }
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_FLAGS(ibbl) & FB_PROG_ENTRY ) continue;
        if( VALID(ABBL_PREV(ibbl)) ) continue;
        ifun = ABBL_FUN(ibbl);

        if( fancy )
        {
            last = first = ibbl;

            for(max=0,min=0,jbbl=ibbl; VALID(jbbl); jbbl = ABBL_NEXT(jbbl))
            {
                last = jbbl;
                if( ABBL_COUNT(jbbl) > max )
                    max = ABBL_COUNT(jbbl);
                if( ABBL_COUNT(jbbl) < min )
                    min = ABBL_COUNT(jbbl);
            }

            if( max <= 0 )
            {
                if( VALID(AFUN_BBL_ZERO(ifun))  )
                {
                    ABBL_NEXT(AFUN_BBL_ZERO(ifun)) = first;
                    ABBL_PREV(first) = AFUN_BBL_ZERO(ifun);
                }
                AFUN_BBL_ZERO(ifun)  = last;
            }
            else if( max >= CONFIG_COUNT )
            {
                if( VALID(AFUN_BBL_HOT(ifun))  )
                {
                    ABBL_NEXT(AFUN_BBL_HOT(ifun)) = first;
                    ABBL_PREV(first) = AFUN_BBL_HOT(ifun);
            }
                AFUN_BBL_HOT(ifun)  = last;
            }
            else
            {
                if( VALID(AFUN_BBL_REST(ifun))  )
                {
                    ABBL_NEXT(AFUN_BBL_REST(ifun)) = first;
                    ABBL_PREV(first) = AFUN_BBL_REST(ifun);
                }
                AFUN_BBL_REST(ifun)  = last;
            }
        }
        else
        {
            first = last = ibbl;
            while( VALID(ABBL_NEXT(last)) )
                last = ABBL_NEXT(last);

            if( VALID(AFUN_BBL_REST(ifun))  )
            {
                ABBL_NEXT(AFUN_BBL_REST(ifun)) = first;
                ABBL_PREV(first) = AFUN_BBL_REST(ifun);
            }
            
            AFUN_BBL_REST(ifun)  = last;
        }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    InfoStatus(fp,"\t[LayoutBuildFunctionChain] {%d ms}\n",time);    
}

/*=======================================================
We try to group bbl from the same function next to each other.
=======================================================*/

static int cmp_chain(const void *x1,const void *x2)
{
    INT64 c1 = ABBL_CHAIN_WEIGHT(*(INDEX *)x1);
    INT64 c2 = ABBL_CHAIN_WEIGHT(*(INDEX *)x2);

    if( c1 < c2 )
        return 1;
    else if( c1 > c2 )
        return -1;
    else
        return 0;    
}

#undef __PROC__
#define __PROC__ "LayoutBuildSingleChain"


static void LayoutBuildSingleChain(FILE *fp,BOOL fancy)
{
    INDEX ibbl;
    INDEX jbbl;
    INDEX ibbl_entry;
    INT64 weight;
    INDEX *array;
    NUMBER nchains;
        
    clock_t time = clock();    

    array = calloc(sizeof(INDEX), CONFIG_NBBL);

    nchains = 0;

    if( fancy )
    {
        ibbl_entry = INVALID;
        
        FOREACH_LIVE_BBL(ibbl)
        {
            ABBL_CHAIN_WEIGHT(ibbl) = 0;
            
            if( ABBL_FLAGS(ibbl) & FB_PROG_ENTRY )
            {
                CERROR( VALID(ABBL_PREV(ibbl)) ,("weird\n"));
                CERROR( VALID(ibbl_entry),("weird\n"));
                ibbl_entry = ibbl;
                continue;
            }
            
            if( VALID(ABBL_PREV(ibbl)) ) continue;
            
            array[nchains] = ibbl;
            nchains++;
            weight = 0;
            for(jbbl=ibbl; VALID(jbbl); jbbl = ABBL_NEXT(jbbl))
            {
#if 1
                if( ABBL_COUNT(jbbl) > weight )
                    weight = ABBL_COUNT(jbbl);
#else
                weight += AbblWeight(jbbl);
#endif
            }
            
            ABBL_CHAIN_WEIGHT(ibbl) = weight;
        }
        
        qsort(array,nchains,sizeof(INDEX),cmp_chain);
        
        for(jbbl=0;jbbl<nchains;jbbl++)
        {
            while( VALID(ABBL_NEXT(ibbl_entry)) )
                ibbl_entry = ABBL_NEXT(ibbl_entry);
            
            ibbl = array[jbbl];       
            ABBL_NEXT(ibbl_entry) = ibbl;
            ABBL_PREV(ibbl) = ibbl_entry;
        }
        
        free(array);
    }
    else
    {
        ibbl_entry = INVALID;
        
        FOREACH_LIVE_BBL(ibbl)
        {
            if( ABBL_FLAGS(ibbl) & FB_PROG_ENTRY )
            {
                CERROR( VALID(ABBL_PREV(ibbl)) ,("prog entry has prev\n"));
                CERROR( VALID(ibbl_entry),("found another prog entry\n"));
                ibbl_entry = ibbl;
                continue;
            }
        }

        CASSERT( VALID(ibbl_entry),("could not find program entry\n"));

        jbbl = ibbl_entry;
        
        FOREACH_LIVE_BBL(ibbl)
        {
            if( ibbl == ibbl_entry ) continue;
            if( VALID(ABBL_PREV(ibbl)) ) continue;

            nchains++;
            while( VALID(ABBL_NEXT(jbbl)) )
                jbbl = ABBL_NEXT(jbbl);
            
            ABBL_NEXT(jbbl) = ibbl;
            ABBL_PREV(ibbl) = jbbl;
        }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;

    InfoStatus(fp,"\t[LayoutBuildSingleChain] chains: %d {%d ms}\n",nchains,time);
}

   
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "LayoutAddUnconditionalBranches"

static void LayoutAddUnconditionalBranches(FILE *fp)
{
    INDEX iedg;
    INDEX ibbl;
    INDEX hd;
    INDEX tl;
    NUMBER n=0;
    clock_t time = clock();    
    
    FOREACH_LIVE_EDG(iedg)
    {
        if( AEDG_TYPE(iedg) != ET_NORM &&
            AEDG_TYPE(iedg) != ET_FUN_LINK &&
            AEDG_TYPE(iedg) != ET_EXTRA &&
            AEDG_TYPE(iedg) != ET_FALSE ) continue;

        hd = AEDG_HEAD(iedg);
        tl = AEDG_TAIL(iedg);

        if(  AbblNInss(tl) == 0 )
        {
            WARNING(("unexpected empty bbl: %d  (edg: %d->%d)\n",tl,hd,tl));
        }
        

        if( tl == ABBL_NEXT(hd) ) continue;

        n++;
                
        ibbl = LayoutCreateFallthruBbl(iedg);

        if( VALID(ibbl) )
        {
            ABBL_NEXT(ibbl) = ABBL_NEXT(hd);
            ABBL_NEXT(hd) = ibbl;
            ABBL_PREV(ibbl) = hd;
            
            if( VALID(ABBL_NEXT(ibbl)) )
                ABBL_PREV( ABBL_NEXT(ibbl) ) = ibbl;
        }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;

    InfoStatus(fp,"\t[LayoutAddedUnconditionalBranches] Added branches: %d {%d ms}\n",n,time);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "cmp_edg"

static int cmp_edg(const void *x1,const void *x2)
{
    INDEX e1 = *(INDEX *)x1;
    INDEX e2 = *(INDEX *)x2;
    INT64   diff;
    
    if( e1 == INVALID && e2 == INVALID ) return 0;

    if( e1 == INVALID ) return 1;

    if( e2 == INVALID ) return -1;

    diff = AEDG_VALUE(e2) - AEDG_VALUE(e1);

    /* the following make qsort (which is not stable) more deterministic */

    if( diff  == 0 )
    {
        diff = AEDG_HEAD(e2) - AEDG_HEAD(e1);
        if( diff == 0 )
        {
            diff = AEDG_TAIL(e2) - AEDG_TAIL(e1);
        }
    }
    
    return diff;
    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "LayoutPreadjustAndSortEdges"

static void LayoutPreadjustAndSortEdges(FILE *fp, BOOL fancy)
{
    INDEX iedg;
    INDEX ibbl;
    INDEX hd,tl;
    
    for(iedg=0; iedg < CONFIG_UEDG; iedg++ )
    {
        SORTED_EDGE(iedg) = iedg;
    }

    /* mark small bbls that have only hot succs and preds */
       
    FOREACH_LIVE_BBL(ibbl)
    {
        ABBL_FLAGS(ibbl) &= ~F_MARKED;

        if( ABBL_IS_HOT(ibbl) ) continue;

        if( AbblNInss(ibbl) > 16 ) continue;
        
        FOREACH_BBL_SUCC(ibbl,iedg )
        {
            if( ABBL_IS_HOT(AEDG_TAIL(iedg)) || ABBL_TYPE(AEDG_TAIL(iedg)) == BT_RET )
                continue;
            break;
        }

        if( VALID(iedg) ) continue;

        FOREACH_BBL_PRED(ibbl,iedg )
        {
            if( ABBL_IS_HOT(AEDG_HEAD(iedg)) )
                continue;
            break;
        }

        if( VALID(iedg) ) continue;
                
        ABBL_FLAGS(ibbl) |= F_MARKED;
    }
    
    FOREACH_LIVE_EDG(iedg)
    {
        if( AEDG_TYPE(iedg) == ET_FALSE && AEDG_HEAD(iedg) == AEDG_TAIL(iedg) )
        {
            AbblBranchTwist(AEDG_HEAD(iedg));
        }
    }
    
    FOREACH_LIVE_EDG(iedg)
    {
        AEDG_FLAGS(iedg) &= ~(FE_ONLY_OUT|FE_ONLY_IN|FE_FIXED);
    }

    
    FOREACH_LIVE_EDG(iedg)
    {
        AEDG_FLAGS(iedg) &= ~(FE_ONLY_OUT|FE_ONLY_IN|FE_FIXED);
        if( AEDG_TYPE(iedg) == ET_TRUE && AEDG_HEAD(iedg) == AEDG_TAIL(iedg) )
        {
            AEDG_FLAGS(iedg) |= FE_FIXED;
            AEDG_FLAGS( ABBL_EDG_FALSE(AEDG_HEAD(iedg)) ) |= FE_ONLY_OUT;
        }

        if( AEDG_TYPE(iedg) == ET_FUN_LINK )
        {
            AEDG_FLAGS(iedg) |= FE_ONLY_OUT;
        }
    }
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( AbblNpreds(ibbl) == 1 )
        {
            AEDG_FLAGS(ABBL_PRED(ibbl)) |= FE_ONLY_IN;
        }

        if( AbblNsuccs(ibbl) == 1 )
        {
            AEDG_FLAGS(ABBL_SUCC(ibbl)) |= FE_ONLY_OUT;
        }
    }

    FOREACH_EDG(iedg)
    {
        if( AEDG_IS_KILLED(iedg) )
        {
            AEDG_VALUE(iedg) = 0;
            continue;
        }
        
        if( AEDG_TYPE(iedg) != ET_NORM &&
            AEDG_TYPE(iedg) != ET_TRUE &&
            AEDG_TYPE(iedg) != ET_FUN_LINK &&
            AEDG_TYPE(iedg) != ET_EXTRA &&
            AEDG_TYPE(iedg) != ET_FALSE )
        {
            AEDG_VALUE(iedg) = 0;
            continue;
        }

        if( AEDG_FLAGS(iedg) & FE_FIXED )
        {
            AEDG_VALUE(iedg) = 0;
            continue;            
        }
        
        if( fancy )
        {
            hd = AEDG_HEAD(iedg);
            tl = AEDG_TAIL(iedg);

#if 0
            if( (ABBL_IS_HOT(hd) && !ABBL_IS_MARKED(tl) && !ABBL_IS_HOT(tl)) ||
                (!ABBL_IS_HOT(hd) && !ABBL_IS_MARKED(hd) && ABBL_IS_HOT(tl)) )
            {
                    
                AEDG_VALUE(iedg) = 0;
            }
            else
#endif
            {
                AEDG_VALUE(iedg) = AEDG_COUNT(iedg) + 1;
                if( AEDG_VALUE(iedg) < 1 ) AEDG_VALUE(iedg) = 1;
            }

            
            if( AEDG_COUNT(iedg) * 32 < ABBL_COUNT(hd) )
            {
                AEDG_VALUE(iedg) = 0 ;
            }
            else 
            {
                AEDG_VALUE(iedg) = AEDG_COUNT(iedg) + 1;
            }
            
#if 0
            if( (ABBL_IS_HOT(hd) && !ABBL_IS_HOT(tl))  || (!ABBL_IS_HOT(hd) && ABBL_IS_HOT(tl)) )
            {
                if( AEDG_VALUE(iedg) != 0 )
                {
                    printf("NEW SCHEME USES EDGE\n");
                    PrintEdgSuper(stdout,iedg);
                }
                
            }
            else
            {
                if( AEDG_VALUE(iedg) == 0 )
                {
                    printf("NEW SCHEME DISCARDS EDGE %d %d\n",ABBL_IS_MARKED(hd),ABBL_IS_MARKED(tl));
                    PrintEdg(stdout,iedg);
                    PrintBblSuper(stdout,hd);
                    PrintBblSuper(stdout,tl);
                }
            }
#endif       
            
        }
        else
        {
            AEDG_VALUE(iedg) = 1024;
        }

#if 1
        if( AEDG_FLAGS(iedg) & (FE_ONLY_IN|FE_ONLY_OUT ) )
#else
            if( AEDG_FLAGS(iedg) & FE_ONLY_OUT )
#endif
        {
            AEDG_VALUE(iedg) *= 3;
            AEDG_VALUE(iedg) /= 2;
        }
    }

    qsort(SortedEdges,CONFIG_UEDG,sizeof(INDEX),cmp_edg);

}



     
/*=======================================================
=======================================================*/     
#undef __PROC__
#define __PROC__ "LayoutCreateChains"

static void LayoutCreateChains(FILE *fp)
{
    INDEX iedg;
    INDEX ibbl;
    INDEX i;
    INDEX hd;
    INDEX tl;
    clock_t time = clock();
                   
    /* create chains by decreasing value */

    for(i=0; i < CONFIG_UEDG; i++ )
    {
        iedg = SORTED_EDGE(i);
        
        if( AEDG_VALUE(iedg) == 0  )
            break; 

        CASSERT( !AEDG_IS_KILLED(iedg),("weird\n"));


        CERROR( AEDG_TYPE(iedg) != ET_NORM &&
                AEDG_TYPE(iedg) != ET_TRUE &&
                AEDG_TYPE(iedg) != ET_FUN_LINK &&
                AEDG_TYPE(iedg) != ET_EXTRA &&
                AEDG_TYPE(iedg) != ET_FALSE,("weird\n"));
        
        hd = AEDG_HEAD(iedg);
        tl = AEDG_TAIL(iedg);
        
        if( VALID(ABBL_NEXT(hd)) ) continue;
        if( VALID(ABBL_PREV(tl)) ) continue;
        
        /* check for circularity */

        for( ibbl = hd; VALID(ABBL_PREV(ibbl)); ibbl = ABBL_PREV(ibbl) )
            ;
        
        if( ibbl == tl )
        {
            continue;
        }
                
        /* link chains */
        
        ABBL_NEXT(hd) = tl;
        ABBL_PREV(tl) = hd;

        if( AEDG_TYPE(iedg) == ET_TRUE )
        {
            AbblBranchTwist(hd);
        }
#if 0
        printf("%3d->%3d %ld\n",AEDG_HEAD(iedg),AEDG_TAIL(iedg),AEDG_COUNT(iedg));
#endif
    }
    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;

    InfoStatus(fp,"\t[LayoutCreateChains] {%d ms}\n",time);
}

/*=======================================================
@FUN
Take all the basic blocks in the program and put them in
good sequence (using profiling information) and place this sequence
into xfun. Add unconditional branches to those basic blocks
of type BT_NORM that were not placed adjacent to their successor.
The type of such a basic block become BT_UNDOND.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "LayoutCode"

extern void LayoutCode(FILE *fp,INDEX xfun)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Layout");
    BOOL   fancy;
    INDEX ibbl;
    INDEX hd;
    INDEX tl;
    
    clock_t time = clock();

    OptUnlinkEmptyBbls(fp);

    InitChain();

    if( CONFIG_HAVE_BBL_PROFILE && !HasOption(DISABLED_OPT,"Layout") )
    {
        fancy = 1;
    }
    else
    {
        fancy = 0;
    }
    
    FOREACH_BBL(ibbl)
    {
        ABBL_NEXT(ibbl) = INVALID;
        ABBL_PREV(ibbl) = INVALID;
    }

    LayoutPreadjustAndSortEdges(fp,fancy);
    LayoutCreateChains(fp);


    if( verbose )
    {
        FOREACH_LIVE_BBL(ibbl)
        {
            if( VALID(ABBL_PREV(ibbl)) ) continue;
            PrintList(ibbl);
        }
    }

    LayoutBuildFunctionChain(fp,fancy);

    LayoutBuildSingleChain(fp,fancy);

    if( fancy )
    {
        FOREACH_LIVE_BBL(ibbl)
        {
            if( ABBL_TYPE(ibbl) != BT_BRANCH ) continue;
            if( AEDG_TAIL(ABBL_EDG_FALSE(ibbl)) == ABBL_NEXT(ibbl) ) continue;
            if( AEDG_COUNT(ABBL_EDG_FALSE(ibbl)) <= AEDG_COUNT(ABBL_EDG_TRUE(ibbl))) continue;
            
            HURRAH(("suboptimal bbl %d (%ld) detected: f %ld t %ld\n",
                    ibbl,ABBL_COUNT(ibbl),AEDG_COUNT(ABBL_EDG_FALSE(ibbl)),AEDG_COUNT(ABBL_EDG_TRUE(ibbl))));
            AbblBranchTwist(ibbl);
        }
    }
    

    LayoutAddUnconditionalBranches(fp);

        
    /* find binning and end of chain entry */

    hd = INVALID;
    tl = INVALID;
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( !VALID(ABBL_PREV(ibbl)) )
        {
            CASSERT( ABBL_FLAGS(ibbl) & FB_PROG_ENTRY,("non entry bbl with invalid prev encountered  %d\n",ibbl));
            CERROR( VALID(hd),("unexpected bbl with invalid prev: %d\n",ibbl));
            hd = ibbl;
        }

        if( !VALID(ABBL_NEXT(ibbl)) )
        {
            CERROR( VALID(tl),("unexpected bbl with invalid next: %d\n",ibbl));
            tl = ibbl;
        }
        
    }

    CASSERT( VALID(tl),("cannot find tail\n"));
    CASSERT( VALID(hd),("cannot find head\n"));

    AFUN_BBL_FIRST(xfun) = hd;
    AFUN_BBL_LAST(xfun) = tl;

    if( verbose )
        PrintLayout(stdout,xfun);

    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    
    ExitChain();

    InfoStatus(fp,"\t[LayoutCode] mode %s  {%d ms}\n",fancy ? "fancy" : "simple", time);

}

/*=======================================================
=======================================================*/

