/*=======================================================
  Author:   R. Muth
  Version:  $Id: binary.coff.c,v 1.9.1.33 1999/08/10 00:52:57 muth Exp $
  Comment:  This file contains high level procedures
            handling the binary
=======================================================*/


#include "global.h"
#include "functions.h"
#include "flags.h"
#include "system.coff.h"

#define ABBL_DFS_NUM(i) ABBL_U1(i).doubleword.d1
#define ABBL_NEW_ADDR(i) ABBL_U3(i).doubleword.d1

/*=======================================================
=======================================================*/
#if 0
#undef __PROC__
#define __PROC__ "IncreaseRconst"

static void IncreaseRconst(INT64 nBytes)
{
    INDEX isec_rconst = BIN_SECTION_MAP[R_SN_RCONST];
    INT8 *Data;
#if 0
    PrintSCNHDR(stdout,&ASEC_SCNHDR(isec_rconst));
#endif    
    Data = (INT8 *) calloc(nBytes + ASEC_SIZE(isec_rconst),1);
    memcpy( Data+nBytes,ASEC_DATA8(isec_rconst),ASEC_SIZE(isec_rconst));
    ASEC_SIZE(isec_rconst) += nBytes;
    ASEC_START(isec_rconst) -= nBytes;
    ASEC_STARTP(isec_rconst) -= nBytes;
    ASEC_DATA8(isec_rconst)  = Data;
#if 0
    PrintSCNHDR(stdout,&ASEC_SCNHDR(isec_rconst));
#endif    
    BIN_AOUTHDR.tsize += nBytes;
    BIN_AOUTHDR.text_start -= nBytes;
}
#endif
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "IncreaseData"

extern void IncreaseData(INT64 nBytes)
{
    INDEX isec_data = BIN_SECTION_MAP[R_SN_DATA];
    INT8 *Data;
#if 0
    PrintSCNHDR(stdout,&ASEC_SCNHDR(isec_data));
#endif    
    Data = (INT8 *) calloc(nBytes + ASEC_SIZE(isec_data),1);
    memcpy( Data+nBytes,ASEC_DATA8(isec_data),ASEC_SIZE(isec_data));
    ASEC_SIZE(isec_data) += nBytes;
    ASEC_START(isec_data) -= nBytes;
    ASEC_STARTP(isec_data) -= nBytes;
    ASEC_DATA8(isec_data)  = Data;
#if 0
    PrintSCNHDR(stdout,&ASEC_SCNHDR(isec_data));
#endif    
    BIN_AOUTHDR.dsize += nBytes;
    BIN_AOUTHDR.data_start -= nBytes;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "BinaryLoad"

extern void BinaryLoad()
{
    INDEX i;
    INDEX is;
    INT64 nreloc;
    
    FILE *fp=stdout;

    InfoPhase(fp,"Load");
    
    /* open file */
    
    fp = fopen(CONFIG_INFILE, "r");
    CFATAL( fp==0, ("cannot open input file %s\n",CONFIG_INFILE));
    
    /* read file header */
    
    fread(&BIN_FILHDR, sizeof(FILHDR), 1, fp);
    CERROR( BIN_FILHDR.f_magic != ALPHAMAGIC,("not an object file\n"));
    CERROR( BIN_FILHDR.f_opthdr != sizeof(AOUTHDR),("bad optional header size\n"));
    
    /* read optional file header (aouthdr) */
    
    fread(&BIN_AOUTHDR, sizeof(AOUTHDR), 1, fp);
    CFATAL( BIN_AOUTHDR.magic != ZMAGIC,("not a ZMAGIC file\n"));
    
    /* load section */
    
    for(i=0;i<BIN_NSECTIONS;i++)
    {
        is = AsecNew();
      
        fread(&ASEC_SCNHDR(is), sizeof(SCNHDR), 1, fp);
        CWARNING( ASEC_SIZE(is) & 0xf,("Section %.8s has odd length\n",ASEC_NAME(is)));
        
        if( ASEC_IS_TEXT(is) )   BIN_SECTION_MAP[R_SN_TEXT] = is;
        if( ASEC_IS_RDATA(is) )  BIN_SECTION_MAP[R_SN_RDATA] = is;
        if( ASEC_IS_DATA(is) )   BIN_SECTION_MAP[R_SN_DATA] = is;
        if( ASEC_IS_SDATA(is) )  BIN_SECTION_MAP[R_SN_SDATA] = is;
        if( ASEC_IS_SBSS(is) )   BIN_SECTION_MAP[R_SN_SBSS] = is;
        if( ASEC_IS_BSS(is) )    BIN_SECTION_MAP[R_SN_BSS] = is;
        if( ASEC_IS_INIT(is) )   BIN_SECTION_MAP[R_SN_INIT] = is;
        if( ASEC_IS_LIT8(is) )   BIN_SECTION_MAP[R_SN_LIT8] = is;
        if( ASEC_IS_LIT4(is) )   BIN_SECTION_MAP[R_SN_LIT4] = is;
        if( ASEC_IS_XDATA(is) )  BIN_SECTION_MAP[R_SN_XDATA] = is;
        if( ASEC_IS_PDATA(is) )  BIN_SECTION_MAP[R_SN_PDATA] = is;
        if( ASEC_IS_FINI(is) )   BIN_SECTION_MAP[R_SN_FINI] = is;
        if( ASEC_IS_LITA(is) )   BIN_SECTION_MAP[R_SN_LITA] = is;
        if( ASEC_IS_RCONST(is) ) BIN_SECTION_MAP[R_SN_RCONST] = is;
    }


    is = BIN_SECTION_MAP[R_SN_LITA];

    CFATAL(ASEC_SIZE(is) > 0xffff,("program has multiple global pointers (.lita size: %d\n",ASEC_SIZE(is)));

    nreloc = 0;
    
    FOREACH_SEC(is)
    {
        /* load section data if available (bss has no) */
#if 0
        printf("processing: %.8s\n",ASEC_NAME(is));
#endif   
        if( ASEC_SCNHDR(is).s_scnptr )
        {
            /* we allocate as much as there are instructions */

            if( ASEC_IS_TEXT(is) )
            {
                ASEC_DATA8(is) = calloc(4,CONFIG_NINS);
                CFATAL( CONFIG_NINS*4 < ASEC_SIZE(is),("text segement too big (%ld) - recompile alto\n",ASEC_SIZE(is)));                
            }
            else
            {
                ASEC_DATA8(is) = calloc(1,ASEC_SIZE(is));
            }
            

            
            fseek(fp,ASEC_SCNHDR(is).s_scnptr,SEEK_SET);
            fread(ASEC_DATA8(is),1,ASEC_SIZE(is),fp);
        }
        
        /* determine number of reloc entrSies */

        if( ASEC_HAS_RELOC_OVFL(is) )
        {
            RELOC reloc;
            WARNING(("Reloc overflow in section: %.8s\n",ASEC_NAME(is)));
            fseek(fp,ASEC_SCNHDR(is).s_relptr,SEEK_SET);
            fread(&reloc,1,sizeof(RELOC),fp);
            ASEC_NRELOCS(is) = reloc.r_vaddr;
        }
        else
        {
            ASEC_NRELOCS(is) = ASEC_SCNHDR(is).s_nreloc;
        }


        nreloc += ASEC_NRELOCS(is);
        
        /* load section relocation information if available */
        
        if( 0<ASEC_NRELOCS(is) )
        {
            ASEC_RELOC(is) = calloc(sizeof(RELOC),ASEC_NRELOCS(is));
            fseek(fp,ASEC_SCNHDR(is).s_relptr,SEEK_SET);
            fread(ASEC_RELOC(is),ASEC_NRELOCS(is),sizeof(RELOC),fp);
            
        }
        else
        {
            WARNING(("No reloc information found for section %.8s\n",ASEC_NAME(is)));
        }
        
    }

    CFATAL( nreloc == 0,("Program has not relocation  info\n"));
    
    /* load symbol table information if available */
    
    if( BIN_FILHDR.f_symptr )
    {
        
        fseek(fp,BIN_FILHDR.f_symptr,SEEK_SET);
        fread(&BIN_HDRR,sizeof(HDRR),1,fp);

#if 0        
        /* load procedure descriptor table */
        printf("loading npdt\n");
                
        if( BIN_SYMBOL_NPDT )
        {
            BIN_SYMBOL_PDT = calloc(BIN_SYMBOL_NPDT, sizeof(PDR));
            fseek(fp,BIN_HDRR.cbPdOffset,SEEK_SET);
            fread(BIN_SYMBOL_PDT,BIN_SYMBOL_NPDT,sizeof(PDR),fp);
        }
#endif
        
        /* load local symbol table */
                
        if( BIN_SYMBOL_NLST )
        {
            BIN_SYMBOL_LST = calloc(BIN_SYMBOL_NLST, sizeof(SYMR));
            fseek(fp,BIN_HDRR.cbSymOffset,SEEK_SET);
            fread(BIN_SYMBOL_LST,BIN_SYMBOL_NLST,sizeof(SYMR),fp);
        }
        
        /* load local symbol string */

        if( BIN_SYMBOL_NLSS )
        {
            BIN_SYMBOL_LSS = calloc(BIN_SYMBOL_NLSS, 1);
            fseek(fp,BIN_HDRR.cbSsOffset,SEEK_SET);
            fread(BIN_SYMBOL_LSS,BIN_SYMBOL_NLSS,1,fp);
        }

        /* load file descriptor table */

        if( BIN_SYMBOL_NFDT )
        {
            BIN_SYMBOL_FDT = calloc(BIN_SYMBOL_NFDT, sizeof(FDR));
            fseek(fp,BIN_HDRR.cbFdOffset,SEEK_SET);
            fread(BIN_SYMBOL_FDT,BIN_SYMBOL_NFDT,sizeof(FDR),fp);
        }

        /* load external symbol string */
                
        if( BIN_SYMBOL_NESS )
        {
            BIN_SYMBOL_ESS = calloc(BIN_SYMBOL_NESS, 1);
            fseek(fp,BIN_HDRR.cbSsExtOffset,SEEK_SET);
            fread(BIN_SYMBOL_ESS,BIN_SYMBOL_NESS,1,fp);
        }
        
        /* load external symbol table */

        if( BIN_SYMBOL_NEST )
        {
            BIN_SYMBOL_EST = calloc(BIN_SYMBOL_NEST, sizeof(EXTR));
            fseek(fp,BIN_HDRR.cbExtOffset,SEEK_SET);
            fread(BIN_SYMBOL_EST,BIN_SYMBOL_NEST,sizeof(EXTR),fp);
        }
        
        /* patch up things */

        FOREACH_FDR(i)
        {
            INDEX is;
            /*
            INDEX ip;
            */
            
            for(is=AFDR(i).isymBase; is<AFDR(i).isymBase+AFDR(i).csym; is++)
                ALSY(is).iss += AFDR(i).issBase;

#if 0
            for(ip=AFDR(i).ipdFirst; ip<AFDR(i).ipdFirst+AFDR(i).cpd; ip++)
            {
                APDR(ip).isym += AFDR(i).isymBase;
                APDR(ip).adr  = ASYM_VALUE(APDR(ip).isym);
            }
#endif            
        }

        
    }

    fclose(fp);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "BinaryDisassemble"

extern void BinaryDisassemble()
{
    INDEX is;
    FILE *fp=stdout;
    
    InfoPhase(fp,"Decode");    
    
    FOREACH_SEC(is)
    {
        
        CWARNING( ASEC_IS_INIT(is),("Binary has a init section\n"));
        CWARNING( ASEC_IS_FINI(is),("Binary has a fini section\n"));
      
        if( ASEC_IS_TEXT(is) || ASEC_IS_INIT(is)  || ASEC_IS_FINI(is) )
        {
            int i;
            
            for(i=0;  i<ASEC_SIZE(is)>>2; i++ )
            {
                INDEX iins = AinsNew();
                CASSERT( Ins2Addr(iins) == ASEC_START(is) + 4 * i,
                         ("assumption about intruction continuity violated\n"));
                Disassemble(ASEC_DATA32(is)[i],&AINS(iins));
            }
        }

        
        if( ASEC_IS_INIT(is)  || ASEC_IS_FINI(is) )
        {
            INDEX text = BIN_SECTION_MAP[R_SN_TEXT];
            
            ASEC_SIZE(text) +=ASEC_SIZE(is);
            ASEC_SIZE(is) = 0;
        }
        
    }

}
/*=======================================================
=======================================================*/
void Special()
{
    INDEX ifun;
    NUMBER count = 0 ;

    
    OptRemoveUnreachableCode(stdout);

    FOREACH_LIVE_FUN(ifun)
    {
        char fn[256];
        FILE *fp;
        NUMBER n,e;
        INDEX ibbl;
        
        if( !AfunIsLoopAnalyzable(ifun) ) continue;

        PrintFun(stdout,ifun);
        sprintf(fn,"g%04d.txt",count++);
        printf("writing %s\n",fn);
        n = 0;
        e = 0;
        FOREACH_FUN_BBL(ifun,ibbl)
        {
            INDEX iedg;
            ABBL_DFS_NUM(ibbl) = n;
            n++;
            FOREACH_BBL_SUCC(ibbl,iedg)
            {
                if( AEDG_IS_INTER(iedg) ) continue;
                e++;
                
            }
            
        }
        
        printf("nodes %d edges %d\n",n,e);
        fp = fopen(fn,"w");
        CASSERT( fp != 0 ,("xxxxxxxxxxxxxxxx\n"));
        
        fprintf(fp,"#beg %s\n",AFUN_NAME(ifun));
        fprintf(fp,"%d\n",n);
        fprintf(fp,"%d\n",e);
        fprintf(fp,"###########\n");
        FOREACH_FUN_BBL(ifun,ibbl)
        {
            INDEX iedg;
            FOREACH_BBL_SUCC(ibbl,iedg)
            {
                if( AEDG_IS_INTER(iedg) ) continue;
                fprintf(fp,"%-5d %5d\n",
                       ABBL_DFS_NUM(ibbl),ABBL_DFS_NUM(AEDG_TAIL(iedg)));
            }
            
        }
        fprintf(fp,"#eof %s\n",AFUN_NAME(ifun));
        fclose(fp);
    }
    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "BinaryConstructFlowGraph"

extern void BinaryConstructFlowGraph()
{
    FILE *fp = stdout;

    InfoPhase(fp,"Flow Graph Construction");
        
    CASSERT( ((4*CONFIG_NTAB+8*CONFIG_NCON) & 0xffffL) == 0,("bad CONFIG_TAB value\n"));
    IncreaseData(4*CONFIG_NTAB+8*CONFIG_NCON);

    Global.SwitchSectionStart = ASEC_START(BIN_SECTION_MAP[R_SN_DATA]);
    Global.ConstSectionStart  = ASEC_START(BIN_SECTION_MAP[R_SN_DATA]) + 4 * CONFIG_NTAB;

    MarkFunctions(fp);

    MarkBblocks(fp);

    SysMarkRelocAndSymbol(fp);

    CreateFunctionsAndBblocks(fp);
    
    CreateEdges(fp);

    KillRelocInstructions(fp);

}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "BinaryTrivialOpts"

extern void BinaryTrivialOpts()
{
    FILE *fp = stdout;

    InfoPhase(fp,"Preprocess");

    
    /* the following optimization is necessary
       otherwise the GlobalPointer-optimizer does not work!
       I guess this needs some work */ 

    OptNormalize(fp);

    SpecialGlobalPointers(fp);

    SpecialEliminateHeadFunctions(fp);

    SpecialSetJmpLongJmp(fp);

    /* the following optimizations just do some cleanup */
    
    OptRemoveUnreachableCode(fp);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "BinaryCompress"

extern void BinaryCompress()
{
    FILE *fp = stdout;

#if 1
    ComputeSavedChangedRegisters(fp);
    ComLiveness(fp);
    FactorPrologs(fp);

    ComputeSavedChangedRegisters(fp);
    ComLiveness(fp);


    NormEpilogs(fp);    
    FactorEpilogs(fp);

    
    FactorCleanupDataField(fp);

    FactorSeparateBranches(fp);

    ComputeSavedChangedRegisters(fp);
    ComLiveness(fp);
    FactorRegRealloc(fp,0);

    ComputeSavedChangedRegisters(fp);
    ComLiveness(fp);
    FactorRegRealloc(fp,1);


    ComputeSavedChangedRegisters(fp);
    ComLiveness(fp);
    FactorRegRealloc(fp,2);

    ComputeSavedChangedRegisters(fp);
    ComLiveness(fp);
    FactorRegRealloc(fp,3);





    ComputeSavedChangedRegisters(fp);
    ComLiveness(fp);
    FactorBbls(fp);

    OptimizeEasy(fp,1);
#endif
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "BinaryProcess"

extern void BinaryProcess()
{
    NUMBER i;
    FILE *fp = stdout;
    BOOL compress = CONFIG_COMPRESS;

    /* generate bbl instrumentation if requested */
       
    if( HasOption(CONFIG_PROFILE_WHAT, "Bbl") ||
        HasOption(CONFIG_PROFILE_WHAT, "Edg") )
    {
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
	ComInsertProfilingInformation(stdout);
	return;  /* don't optimize instrumented exe any further */
    }
    
    /* Read bbl profiles if available and wanted*/

    if( !HasOption(DISABLED_OPT, "Profiles") )
    {
        CONFIG_HAVE_BBL_PROFILE = ProfReadAltoCounts(fp);
        if( !CONFIG_HAVE_BBL_PROFILE )
            CONFIG_HAVE_BBL_PROFILE = ProfReadPixieCounts(fp);
        if( CONFIG_HAVE_BBL_PROFILE )
            ProfCheckCounts(fp);
    }

    /* base optimizations */
      
    if( HasOption(PRINT_OPT_WHEN,"Preopt") )
        PrintAll(fp);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
    
    for(i=0;i<CONFIG_ROUNDS; i++ )
    {
        /***************************************************/    
        ComputeCountCutOff(fp,50,0.60,3 * CONFIG_ICACHE_SIZE / 4);        
        InfStatResource(fp);
        /***************************************************/

        if( !OptimizeEasy(fp,compress) )
            break;
    }

    CWARNING( i == CONFIG_ROUNDS,("OptimizeEasy might need more rounds\n"));
    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    
    ComputeCountCutOff(fp,50,0.40,0);        
    InfStatResource(fp);
    /***************************************************/    

    OptimizeHard(fp,compress);

#if 0
    OptimizeEasy(fp);   

    if( HasOption(CONFIG_PROFILE_WHAT, "Value") )
    {
	ComInsertProfilingInformation(fp);
	return;  /* don't further optimize profile */
    }
    else if( !HasOption(DISABLED_OPT, "Specialize") )
    {
	ComReadValueProfilingInformation(fp);
            
        if( CONFIG_NVALUES > 0 ) 
        {
            HURRAH(("Using value profiling info\n"));
            OptRemoveUnreachableCode(fp);
            ComputeSavedChangedRegisters(fp);
            ComLiveness(fp);
            ComStackInfo(fp);
            ComUseDefDag(fp);
            ComUseDefConstPot(fp);
            OptValProfSpecialize(fp); 
        }
    }
#endif


    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
    
    /* base optimizations */
    
    for(i=0;i<CONFIG_ROUNDS; i++ )
    {
        /***************************************************/    
        ComputeCountCutOff(fp,50,0.60, 3 * CONFIG_ICACHE_SIZE / 4);        
        InfStatResource(fp);
        /***************************************************/

        if( !OptimizeEasy(fp,compress) )
            break;

    }

    CWARNING( i == CONFIG_ROUNDS,("OptimizeEasy might need more rounds\n"));

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

        
    if( HasOption(PRINT_OPT_WHEN,"Postopt") )
        PrintAll(fp);
    
    if( compress )
        BinaryCompress();
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "BinarySave"

int cmp(const void *x1,const void *x2)
{
    unsigned long addr1 =((STRUCT_SEC *) x1)->scnhdr.s_vaddr;
    unsigned long addr2 =((STRUCT_SEC *) x2)->scnhdr.s_vaddr;

    /*
      some of the s_vaddr's are zero
      they need to go to the end of the file,
      so we make them big !!!
      */

    if( addr1 == 0 ) addr1 -= 1;
    if( addr2 == 0 ) addr2 -= 1;
    


    if( addr1 > addr2 ) return  1;
    if( addr1 < addr2 ) return -1;
    return 0;
}

extern void BinarySave()
{
    INDEX is;
    INT64 fileptr;
    FILE *fp=stdout;
    InfoPhase(fp,"Save");

#if 0    
    /* probably not needed */

    is = BIN_SECTION_MAP[R_SN_INIT];
 
     ASEC_START(is) = 0;
     ASEC_STARTP(is) = 0;
 
     is = BIN_SECTION_MAP[R_SN_FINI];
 
     ASEC_START(is) = 0;
     ASEC_STARTP(is) = 0;
#endif
    /* patch file headers */

    /* there will be no symbol table */
    
    BIN_FILHDR.f_symptr = 0;
    BIN_FILHDR.f_nsyms = 0;
    BIN_FILHDR.f_flags = F_RELFLG | F_EXEC | F_LNNO;
    
    /* patch section headers */

    fileptr = sizeof(FILHDR) + sizeof(AOUTHDR) + BIN_NSECTIONS * sizeof(SCNHDR);

    qsort(aSec,BIN_NSECTIONS,sizeof(STRUCT_SEC),&cmp);
    
    FOREACH_SEC(is)
    {

        ASEC_SCNHDR(is).s_nreloc = 0;
        ASEC_SCNHDR(is).s_relptr = 0;

        /* round up to multiple of 16 */
        
        fileptr = (fileptr+0xf) & ~0xf;   

        if( ASEC_IS_DATA(is) )
            fileptr = (fileptr+0x2000) & ~(0x2000-1);
        
        if( ASEC_IS_TEXT(is) )
        {
            CASSERT(fileptr==ASEC_SCNHDR(is).s_scnptr,
                    ("text segemnt has been moved in file (is %lx was %lx)\n",fileptr,ASEC_SCNHDR(is).s_scnptr));
            /* patch tsize in aoutheader */
        }

        if( ASEC_SCNHDR(is).s_scnptr )
        {
            ASEC_SCNHDR(is).s_scnptr = fileptr;
            fileptr += ASEC_SCNHDR(is).s_size;
        }
    }

    
    fp = fopen(CONFIG_OUTFILE, "w+");
    CASSERT(fp!=0,("Cannot open file: %d\n",CONFIG_OUTFILE));
    
    fwrite(&BIN_FILHDR, sizeof(FILHDR), 1, fp);

    fwrite(&BIN_AOUTHDR, sizeof(AOUTHDR), 1, fp);

    /* write section headers */
    
    FOREACH_SEC(is)
    {
        fwrite(&ASEC_SCNHDR(is), sizeof(SCNHDR), 1, fp);
    }

    /* write section data */
    
    FOREACH_SEC(is)
    {
        if( ASEC_SCNHDR(is).s_scnptr )
        {
            fseek(fp,ASEC_SCNHDR(is).s_scnptr,SEEK_SET );
            fwrite(ASEC_DATA8(is),ASEC_SIZE(is),1,fp );
        }
    }

    

    /* this makes linux alpha happy */
    
    for(is=0;is<16*1024;is++)
        fputc(0,fp);
    
 
    fclose( fp );

}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "BinaryAssemble"

extern void BinaryAssemble()
{
    INDEX xfun;
    FILE *fp = stdout;
    BOOL compress = CONFIG_COMPRESS;
    
    InfoPhase(fp,"Assemble");

    /***************************************************/    
    ComputeCountCutOff(fp,50,0.666, CONFIG_ICACHE_SIZE);        
    InfStatResource(fp);
    /***************************************************/

    /**********************************************************************/
    /*
      the scheduler needs liveness but it is not a good idea
      to compute it after the layout because then, all the functions
      have disappeared.
      */
    /**********************************************************************/
    
    OptRemoveUnreachableCode(fp);
    ComStackInfo(fp);
    ComUseDefDagLoadStoreExtBbl(fp);
    ComAliasAuxInfo(fp);
    
    ComputeSavedChangedRegisters(fp);
    ComLiveness(fp);
    
    /**********************************************************************/
    /* reallocate registers to help scheduler */
    /**********************************************************************/
    
#ifndef NO_REG_REALLOC_TWO
     if( !HasOption(DISABLED_OPT,"Regrealloctwo")  && 0==strcmp("",CONFIG_PROFILE_WHAT) )
     {
         OptRegisterReallocTwo(fp);
     }
#endif

     /**********************************************************************/
     /* create layout reintroduce unconditional branches where necessary */
     /**********************************************************************/
     
    xfun = AfunNewFancy();
    AFUN_NAME(xfun) = "|collect|";

    if( CONFIG_HAVE_BBL_PROFILE &&  !HasOption(DISABLED_OPT,"Icache") )
    {
        ComputeEdgeCounts(fp);
        LayoutCode(fp,xfun);
    }
    else
    {
        LayoutCode(fp,xfun);
    }

    InfStatEdges(fp);

     
    if( CONFIG_HAVE_BBL_PROFILE &&  !HasOption(DISABLED_OPT,"Align") && !compress)
    {
        ComputeEdgeCounts(fp);
        EmitComputeAlignment(fp,xfun);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Preschedule") )
        PrintAll(fp);

    /**********************************************************************/
    /* create layout reintroduce unconditional branches where necessary */
    /**********************************************************************/

    if( !HasOption(DISABLED_OPT,"Schedule")  )
    {
        Schedule(fp,xfun); 
    }

#if 0
    {
        INDEX ibbl;
        FOREACH_LIVE_BBL(ibbl)
        {
            if( !ABBL_IS_HOT(ibbl) ) continue;
            PrintBblSuper(stdout,ibbl);
        }
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postschedule") )
        PrintAll(fp);
#endif

    /**********************************************************************/
    /* generate object code */
    /* second call to EmitGenerateCode is essentially a back patching pass */
    /**********************************************************************/

    EmitGenerateCode(fp,xfun);
    EmitRecomputeAddresses(fp,xfun);

    if( strcmp("",CONFIG_PROFILE_WHAT)  )
    {
	ComInsertInsMap(fp);
    }
    
    EmitGenerateCode(fp,xfun);

}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ComputeNumberOps"

extern void ComputeNumberOps(FILE *fp) 
{
    INDEX ibbl;
    INDEX iins;
    INT64 ildcount = 0;
    INT64 fldcount = 0;
    INT64 brcount = 0;
    INT64 jmpcount = 0;
    INT64 cyclecount = 0;
    
 
    FOREACH_BBL(ibbl) {
	FOREACH_BBL_INS(ibbl, iins) {
	    if (AINS_IS_INT_LOAD(iins)) {
		ildcount += ABBL_COUNT(ibbl);
	    }
	    else if (AINS_IS_FLT_LOAD(iins)) {
		fldcount += ABBL_COUNT(ibbl);
	    }
	    else if (AINS_TYPE(iins) == IT_IBR) {
		brcount += ABBL_COUNT(ibbl);
	    }
	    else if (AINS_TYPE(iins) == IT_FBR) {
		brcount += ABBL_COUNT(ibbl);
	    }
	    else if (AINS_TYPE(iins) == IT_BRA) {
		brcount += ABBL_COUNT(ibbl);
	    }
	    else if (AINS_TYPE(iins) == IT_JMP) {
		jmpcount += ABBL_COUNT(ibbl);
	    }
	    cyclecount += ABBL_COUNT(ibbl);
	}
    }
    fprintf(fp, "grepthis ild : %ld   fld : %ld  br: %ld  jmp: %ld cycle: %ld\n", ildcount, fldcount, brcount, jmpcount, cyclecount);
}
/*=======================================================
=======================================================*/
