/*=======================================================
  Author:   R. Muth
  Version:  $Id: binary.elf.c,v 1.1 1998/02/25 18:14:01 saw Exp $
  Comment:  This file contains high level procedures
            handling the binary
=======================================================*/


#include "global.h"
#include "functions.h"
#include "flags.h"
#include "system.elf.h"



#include <string.h>

INT64 oldsize;

/*=======================================================
=======================================================*/
#if 0
#undef __PROC__
#define __PROC__ "IncreaseRconst"

static void IncreaseRconst(INT64 nBytes)
{
    INDEX isec_rconst = BIN_SECTION_MAP[R_SN_RCONST];
    INT8 *Data;
#if 0
    PrintSCNHDR(stdout,&ASEC_SCNHDR(isec_rconst));
#endif    
    Data = (INT8 *) calloc(nBytes + ASEC_SIZE(isec_rconst),1);
    memcpy( Data+nBytes,ASEC_DATA8(isec_rconst),ASEC_SIZE(isec_rconst));
    ASEC_SIZE(isec_rconst) += nBytes;
    ASEC_START(isec_rconst) -= nBytes;
    ASEC_DATA8(isec_rconst)  = Data;
#if 0
    PrintSCNHDR(stdout,&ASEC_SCNHDR(isec_rconst));
    BIN_AOUTHDR.tsize += nBytes;
    BIN_AOUTHDR.text_start -= nBytes;
#endif    
}
#endif
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "IncreaseData"

static void IncreaseData(INT64 nBytes)
{
    INDEX isec_data = BIN_SECTION_MAP[R_SN_DATA];
    INT8 *Data;

#if 0
    PrintSCNHDR(stdout,&ASEC_SCNHDR(isec_data));
#endif    

    Data = (INT8 *) calloc(nBytes + ASEC_SIZE(isec_data),1);
    memcpy( Data+nBytes,ASEC_DATA8(isec_data),ASEC_SIZE(isec_data));
    ASEC_SIZE(isec_data) += nBytes;

    printf("munging data section from start %lx ", ASEC_START(isec_data));
    ASEC_START(isec_data) -= nBytes;
    ASEC_SCNHDR(isec_data).sh_offset -= nBytes;
    printf("to %lx \n", ASEC_START(isec_data));

    ASEC_DATA8(isec_data)  = Data;


#if 0
    PrintSCNHDR(stdout,&ASEC_SCNHDR(isec_data));
#endif    

}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "BinaryLoad"

extern void BinaryLoad()
{
    INDEX i;
    INDEX is;
    
    FILE *fp;

    PRINT_FUN_BEG1();
    
    /* open file */
    
    fp = fopen(CONFIG_INFILE, "r");
    printf("opening %s\n", CONFIG_INFILE);
    CERROR( fp==0, ("cannot open input file %s\n",CONFIG_INFILE));
    
    /* read file header */
    
    fread(&BIN_FILHDR, sizeof(FILHDR), 1, fp);

    /* now make sure it's the kind of file we want */
    /* should probably check each byte separately, but for now... */
    CERROR( strncmp((char *)BIN_FILHDR.e_ident, (char *)ELFMAG, 4) != 0, 
	    ("Not an elf object file") );
    CERROR( BIN_FILHDR.e_ident[EI_CLASS] != ELFCLASS64 , ("Not a 64-bit elf file") );
    BIN_SECTION_MAP = (UINT8 *)malloc(sizeof(UINT8) * BIN_NSECTIONS);
    
    /* elf section headers are not necessarily right after the 
       file header */
    fseek(fp, BIN_FILHDR.e_shoff, SEEK_SET);
    
    /* load section headers */
    for( i=0; i < BIN_NSECTIONS ; i++)
    {
        is = AsecNew();

        fread(&ASEC_SCNHDR(is), sizeof(SCNHDR), 1, fp);
        CWARNING( ASEC_SIZE(is) & 0xf,("Section %d has odd length\n", is) );
    }

#ifdef ELF
    /* read section header string table for mapping purposes */
    fseek(fp, ASEC_SCNHDR(theBinary.filhdr.e_shstrndx).sh_offset, SEEK_SET);
    BIN_SHDRSTRTAB = (char *)malloc(ASEC_SCNHDR(theBinary.filhdr.e_shstrndx).sh_size);
    fread(BIN_SHDRSTRTAB, 1, ASEC_SCNHDR(theBinary.filhdr.e_shstrndx).sh_size, fp);
#endif
    /* map section headers */
    for( i=0; i < BIN_NSECTIONS ; i++)
    {
	is = i;
	printf("section %d is %s\n", is, &BIN_SHDRSTRTAB[ASEC_SCNHDR(is).sh_name]);
	printf("flags are : %lx\n",ASEC_SCNHDR(is).sh_flags);
        if( ASEC_IS_TEXT(is) ) {  
	    BIN_SECTION_MAP[R_SN_TEXT] = is;
	    printf("section %d is a text section\n", is);
	}

        if( ASEC_IS_RDATA(is) )  BIN_SECTION_MAP[R_SN_RDATA] = is;
        if( ASEC_IS_DATA(is) )   BIN_SECTION_MAP[R_SN_DATA] = is;
        if( ASEC_IS_SDATA(is) )  BIN_SECTION_MAP[R_SN_SDATA] = is;
        if( ASEC_IS_SBSS(is) )   BIN_SECTION_MAP[R_SN_SBSS] = is;
        if( ASEC_IS_BSS(is) )    BIN_SECTION_MAP[R_SN_BSS] = is;
        if( ASEC_IS_INIT(is) )   BIN_SECTION_MAP[R_SN_INIT] = is;
        if( ASEC_IS_FINI(is) )   BIN_SECTION_MAP[R_SN_FINI] = is;
        if( ASEC_IS_RCONST(is) ) BIN_SECTION_MAP[R_SN_RCONST] = is;
	if( ASEC_IS_SYMTAB(is) ) BIN_SECTION_MAP[R_SN_SYMTAB] = is;
	if( ASEC_IS_STRTAB(is) ) BIN_SECTION_MAP[R_SN_STRTAB] = is;
	if( ASEC_IS_GOT(is) )    BIN_SECTION_MAP[R_SN_GOT] = is;
	if( ASEC_IS_SHDRSTRTAB(is) ) BIN_SECTION_MAP[R_SN_SHDRSTRTAB] = is;
	if( ASEC_IS_REL(is) ) {
	    ASEC_NRELOCS(is) = ASEC_SCNHDR(is).sh_size / ASEC_SCNHDR(is).sh_entsize;
	    printf("relocation section, size is %lx, number of relocs is %lx\n", ASEC_SCNHDR(is).sh_size, (long)ASEC_NRELOCS(is) );
	    printf("relocation associated w/symbol table %d, applied to %d\n",
		   ASEC_SCNHDR(is).sh_link, ASEC_SCNHDR(is).sh_info);
	}
    }

    
    FOREACH_SEC(is)
    {
        /* load section data if available (bss has no) */
	/* so, we should check to see if the section is the bss section? */
        
        if( is != BIN_SECTION_MAP[R_SN_BSS] )
        {

            /* we allocate as much as there are instructions */
#if 1
            if( ASEC_IS_TEXT(is) )
                ASEC_DATA8(is) = calloc(4,CONFIG_NINS);
            else
#endif
                ASEC_DATA8(is) = calloc(1,ASEC_SIZE(is));

            fseek(fp,ASEC_SCNHDR(is).sh_offset,SEEK_SET);
            fread(ASEC_DATA8(is),1,ASEC_SIZE(is),fp);

	}

    }
    /* set the string table */
    BIN_STRING_TAB = (char *)ASEC_DATA8(BIN_SECTION_MAP[R_SN_STRTAB]);
        

    theBinary.PHDRS = (Elf64_Phdr *)malloc(BIN_FILHDR.e_phnum * BIN_FILHDR.e_phentsize);
    fseek(fp, BIN_FILHDR.e_phoff, SEEK_SET);
    fread(theBinary.PHDRS, BIN_FILHDR.e_phnum, BIN_FILHDR.e_phentsize, fp);

    fclose(fp);
    PRINT_FUN_END1();
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "BinaryDisassemble"

extern void BinaryDisassemble()
{
    INDEX is;
    unsigned long tempstart = 0;
    unsigned long addition = 0;

    PRINT_FUN_BEG1();
    
    FOREACH_SEC(is)
    {
        
        CWARNING( ASEC_IS_INIT(is),("Binary has a init section, %d\n", is));
        CWARNING( ASEC_IS_FINI(is),("Binary has a fini section, %d\n", is));
	if (ASEC_IS_INIT(is)) {
	    tempstart = ASEC_START(BIN_SECTION_MAP[R_SN_TEXT]);
	    ASEC_START(BIN_SECTION_MAP[R_SN_TEXT]) = ASEC_START(is);
	}


        if( ASEC_IS_TEXT(is) || ASEC_IS_INIT(is)  || ASEC_IS_FINI(is) )
        {
            int i;
            
            for(i=0;  i<ASEC_SIZE(is)>>2; i++ )
            {
                INDEX iins = AinsNew();
/*  		CASSERT( Ins2Addr(iins) == ASEC_START(is) + 4 * i,  */
/*  			 ("assumption about intruction continuity violated\n"));   */
		if (ASEC_DATA32(is)[i] == 0x1f041f04) {
		    Disassemble(0x47ff041f, &AINS(iins));
		}
		else {
		    Disassemble(ASEC_DATA32(is)[i],&AINS(iins));
		}
            }
        }


        /* if it's an init section, let's fill in the unused space between 
	 * the init and the text section */
	/* XXXXX assumes text section is next after init section */
	printf("ASEC_START is %lx, ASEC_SIZE is %lx, text start is %lx\n",
	       ASEC_START(is), ASEC_SIZE(is), tempstart);
	if (ASEC_IS_INIT(is) && 
	    ((ASEC_START(is) + (ASEC_SIZE(is)>>2) < tempstart - 4))) 
	    {
		INDEX iins;
		printf("using up space between sections \n");
		iins = AinsNew();
		Disassemble(0x47ff041f, &AINS(iins));
		while (Ins2Addr(iins) < tempstart - 4) {
		    iins = AinsNew();
		    Disassemble(0x47ff041f, &AINS(iins));
		    PrintIns(stdout,iins);
		}
	}
        if( ASEC_IS_INIT(is)  || ASEC_IS_FINI(is) )
        {
	    /* add this section size to the text section size (later) */
	    addition += ASEC_SIZE(is);
	    ASEC_SIZE(is) = 0;
        }
        
    }

    printf("adjusting text size\n");
    {
	INDEX text = BIN_SECTION_MAP[R_SN_TEXT];
	
	ASEC_SIZE(text) += addition;
	ASEC_SCNHDR(text).sh_offset = ASEC_SCNHDR(text-1).sh_offset;
    }

    SysComputeGlobalPointer();

    PRINT_FUN_END1();
}
/*=======================================================
=======================================================*/
void Special()
{
    INDEX ifun;
    NUMBER count = 0 ;

    
    OptRemoveUnreachableCode(stdout);

    FOREACH_LIVE_FUN(ifun)
    {
        char fn[256];
        FILE *fp;
        NUMBER n,e;
        INDEX ibbl;
        
        if( !AfunIsLoopAnalyzable(ifun) ) continue;

        PrintFun(stdout,ifun);
        sprintf(fn,"g%04d.txt",count++);
        printf("writing %s\n",fn);
        n = 0;
        e = 0;
        FOREACH_FUN_BBL(ifun,ibbl)
        {
            INDEX iedg;
            ABBL_DFS_NUM(ibbl) = n;
            n++;
            FOREACH_BBL_SUCC(ibbl,iedg)
            {
                if( AEDG_IS_INTER(iedg) ) continue;
                e++;
                
            }
            
        }
        
        printf("nodes %d edges %d\n",n,e);
        fp = fopen(fn,"w");
        CASSERT( fp != 0 ,("xxxxxxxxxxxxxxxx\n"));
        
        fprintf(fp,"#beg %s\n",AFUN_NAME(ifun));
        fprintf(fp,"%d\n",n);
        fprintf(fp,"%d\n",e);
        fprintf(fp,"###########\n");
        FOREACH_FUN_BBL(ifun,ibbl)
        {
            INDEX iedg;
            FOREACH_BBL_SUCC(ibbl,iedg)
            {
                if( AEDG_IS_INTER(iedg) ) continue;
                fprintf(fp,"%-5ld %5ld\n",
                       ABBL_DFS_NUM(ibbl),ABBL_DFS_NUM(AEDG_TAIL(iedg)));
            }
            
        }
        fprintf(fp,"#eof %s\n",AFUN_NAME(ifun));
        fclose(fp);
    }
    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "BinaryConstructFlowGraph"

extern void BinaryConstructFlowGraph()
{
    FILE *fp = stdout;
    
    PRINT_FUN_BEG1();

    CASSERT( ((4*CONFIG_NTAB) & 0xffffL) == 0,("bad CONFIG_TAB value\n"));
    IncreaseData(4*CONFIG_NTAB);
    
    MarkFunctions(fp);
    
    MarkBblocks(fp);
    
    SysMarkRelocAndSymbol(fp); 
    
    CreateFunctionsAndBblocks(fp); 
    
    CreateEdges(fp); 
    
    BinaryTrivialOpts();
    
    PRINT_FUN_END1();
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "BinaryTrivialOpts"

/* These used to be done during flow graph construction.
 * Moved because we may want to simply profile this.
 */
extern void BinaryTrivialOpts() {
    FILE *fp = stdout;
    
    PRINT_FUN_BEG1();

    ComputeCountCutOff(fp);
    
    /* the following optimization is necessary
       otherwise the GlobalPointer-optimizer does not work!
       I guess this needs some work */

    OptNormalizeInstructions(fp); 
    
    SpecialGlobalPointers(fp);
    
    SpecialEliminateHeadFunctions(fp); 
    
    SpecialSetJmpLongJmp(fp);
    
}



/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "BinaryProcess"

extern void BinaryProcess()
{
    NUMBER i;
    FILE *fp = stdout;
    
    PRINT_FUN_BEG1();

/*     if (CONFIG_INSERT_PROFILE) return; */

    if( HasOption(PRINT_OPT_WHEN,"Preopt") )
        PrintAll(fp);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
    
    for(i=0;i<CONFIG_ROUNDS; i++ )  
    {  
	fflush(fp); 
          if( !OptimizeEasy(fp) )  
              break;  
    }  

     CWARNING( i == CONFIG_ROUNDS,("OptimizeEasy might need more rounds\n")); 
     if( HasOption(PRINT_OPT_WHAT,"Check") ) 
         CheckAll(fp);

       OptimizeHard(fp);   

     if( HasOption(PRINT_OPT_WHAT,"Check") )
         CheckAll(fp);

     for(i=0;i<CONFIG_ROUNDS; i++ ) 
     { 
	 if( !OptimizeEasy(fp) ) 
	     break;  
     }  

     CWARNING( i == CONFIG_ROUNDS,("OptimizeEasy might need more rounds\n")); 

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

        
/*     if( HasOption(PRINT_OPT_WHEN,"Postopt") ) */
/*         PrintAll(fp); */
            
    PRINT_FUN_END1();
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "BinarySave"

int cmp(const void *x1,const void *x2)
{
    STRUCT_SEC *s1=(STRUCT_SEC *) x1;
    STRUCT_SEC *s2=(STRUCT_SEC *) x2;
    
    if( s1->scnhdr.sh_offset > s2->scnhdr.sh_offset ) return  1;
    if( s1->scnhdr.sh_offset < s2->scnhdr.sh_offset ) return -1;
    return 0;
}

extern void BinarySave()
{
    INDEX is;
    INT64 fileptr;
    INT64 address;
    INT64 newAddress;
    INT32 diff;
    INT64 *gaddr;
    INDEX goff;
    FILE *fp;
    INDEX isec_text;
    INDEX iins;

    PRINT_FUN_BEG1();
    
    /* patch file headers */

    /* there will be no symbol table */
    
#ifdef COFF
    BIN_FILHDR.f_symptr = 0;
    BIN_FILHDR.f_nsyms = 0;
    BIN_FILHDR.f_flags = F_RELFLG | F_EXEC | F_LNNO;
#endif

    
    /* patch section headers */

    fileptr = sizeof(FILHDR);

/*     qsort(aSec,BIN_NSECTIONS,sizeof(STRUCT_SEC),&cmp); */
    

    FOREACH_SEC(is)
    {

#ifdef COFF
        ASEC_SCNHDR(is).s_nreloc = 0;
        ASEC_SCNHDR(is).s_relptr = 0;
#endif

        /* round up to multiple of 16 */
        
        fileptr = (fileptr+0xf) & ~0xf;   

	/* leave the data section where it is */
	if( ASEC_IS_DATA(is) ) 
	    fileptr = ASEC_SCNHDR(is).sh_offset;
        
        if( ASEC_IS_TEXT(is) )
        {
	     /* patch tsize in aoutheader */
	     /* round up to multiple of 64 */
	     fileptr = (fileptr+0xff) & ~0xff;   

             CASSERT(fileptr==ASEC_SCNHDR(is).sh_offset ,
                     ("text segemnt has been moved in file was %lx, is now %lx\n", fileptr, ASEC_SCNHDR(is).sh_offset)); 
        }

/* test for bss again? */
        if( is != BIN_SECTION_MAP[R_SN_BSS] ) {
/* 	    ASEC_SCNHDR(is).sh_offset = fileptr; */
            fileptr += ASEC_SCNHDR(is).sh_size;
        }
    }

    
    fp = fopen(CONFIG_OUTFILE, "w+");
    CASSERT(fp!=0,("Cannot open file: %d\n",CONFIG_OUTFILE));

    /* need to update the shoffset here, put them at the end of the file.
       Make sure we round to 64 alignment. */
    fileptr = (fileptr+0xff) & ~0xff;   
    BIN_FILHDR.e_shoff = fileptr;
    printf("moving section headers to %lx\n", BIN_FILHDR.e_shoff);

    /* also need to update the entry point somewhere */
    address = SysGetProgramEntry();
    printf("address is %lx, translation is %lx, and entry is %lx\n", 
	   address, TranslateAddress(address), BIN_FILHDR.e_entry); 
    BIN_FILHDR.e_entry = TranslateAddress(address); 
    printf("after address is %lx, translation is %lx, and entry is %lx\n", 
	   address, TranslateAddress(address), BIN_FILHDR.e_entry); 

    /* now we need to patch the GP value for the binary based on the 
       original address - the new address */
    diff = address - TranslateAddress(address);
    printf("difference is %d\n", diff);
    isec_text = BIN_SECTION_MAP[R_SN_TEXT]; 
    iins = (TranslateAddress(address) - ASEC_START(isec_text)) >> 2;

    printf("old instruction is %x, new instruction is %x\n", ((INT32 *)ASEC_DATA32(isec_text))[iins], 0x47ff040f); 
    ((INT32 *)ASEC_DATA32(isec_text))[iins] = 0x47ff040f; 

    iins += 3;
    printf("text section is %d, starts at %lx\n",isec_text, ASEC_START(isec_text));
    printf("iins is %d\n", iins);
    printf("old instruction is %x  ", ((INT32 *)ASEC_DATA32(isec_text))[iins]); 
    ((INT32 *)ASEC_DATA32(isec_text))[iins] = diff + ((INT32 *)ASEC_DATA32(isec_text))[iins];
    
    printf("new instruction is %x\n", ((INT32 *)ASEC_DATA32(isec_text))[iins]);
    

    AINS(iins).Data += diff;

    fwrite(&BIN_FILHDR, sizeof(FILHDR), 1, fp);

    /* reset the function offsets in the symbol table */
    for (is=0; is < BIN_SYMBOL_NENT; is++) {
	SYMBOL *s = &SYM(is);
	if( SYM_IS_PROC(s) && SYM_VALUE(is) != 0 )
	{
	    SYM_ADDRESS(is) = TranslateAddress(SYM_ADDRESS(is));
	}
    }
    
    /* reset the function addresses in the global offset table */
    printf("************************************\n");
    printf(" munging the global offset table now\n");
    printf("************************************\n");

    /* store the section index and address here so we don't have to 
     * keep looking it up. */
    is = BIN_SECTION_MAP[R_SN_GOT];
    gaddr = ASEC_DATA64(is);

    for (goff = 0; goff < ASEC_SCNHDR(is).sh_size / 8; goff++) {
	address = gaddr[goff];
	/* if the address is in the text section, we need to patch
	   it. */
	printf("virtual address is currently %lx\n", ASEC_SCNHDR(is).sh_addr + (goff * 8));
	printf("relocating address %lx\n", address);
	if ((address >= ASEC_SCNHDR(isec_text).sh_addr) &&
	    (address <= ASEC_SCNHDR(isec_text).sh_addr + oldsize)) {
	    newAddress = TranslateAddress(address);
	    while (newAddress < 1) {
		address += 4;
		newAddress = TranslateAddress(address);
	    }
	    printf("old global offset table entry is %lx, new is %lx\n", gaddr[goff], newAddress);
	    gaddr[goff] = newAddress;
	}
    }
    fseek(fp, BIN_FILHDR.e_phoff, SEEK_SET);
    fwrite(theBinary.PHDRS, 2, BIN_FILHDR.e_phentsize, fp);
    
    

    /* write section data */
    FOREACH_SEC(is)
	{
	    /* test for bss again? We don't need to write the BSS section 
	       into the object file */
	    if( is != BIN_SECTION_MAP[R_SN_BSS] ) {
		printf("section header %d says section is at offset %lx\n", is, ASEC_SCNHDR(is).sh_offset);
		fseek(fp,ASEC_SCNHDR(is).sh_offset,SEEK_SET );
		fwrite(ASEC_DATA8(is),1, ASEC_SIZE(is),fp );
	    }
	}
    

    /* write section headers */
    
    printf("writing section headers at %lx\n", BIN_FILHDR.e_shoff);
    fseek(fp,BIN_FILHDR.e_shoff,SEEK_SET );
    FOREACH_SEC(is)
    {
	printf("section header %d says section is at offset %lx\n", is, ASEC_SCNHDR(is).sh_offset);
        fwrite(&ASEC_SCNHDR(is), 1, sizeof(SCNHDR), fp);
    }


    fclose( fp );

    PRINT_FUN_END1();
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "BinaryAssemble"

extern    void BinaryAssemble()
{
    INDEX xfun;
    INDEX isec = SysGetTextSection();

    PRINT_FUN_BEG1();

    /* create unconditional branches where necessary */
    
    xfun = AfunNewFancy();
    AFUN_NAME(xfun) = "|collect|";

    if( CONFIG_PROFILE &&  !HasOption(DISABLED_OPT,"Icache") )
    {
        LayoutCodeFancy(stdout,xfun);
    }
    else
    {
        LayoutCodeSimple(stdout,xfun);
    }
    
    ComputeEdgeCounts(stdout);
    EmitComputeAlignment(stdout,xfun);

#if 0
    Schedule(stdout,xfun);
#else
    printf("SCHEDULING HAS BEEN DISABLED\n");
#endif
    
    /* we'll be changing the size of the text section, so remember the
       old size here */
    oldsize = ASEC_SCNHDR(isec).sh_size;
    EmitGenerateCode(stdout,xfun);
    printf("XXXXXXXXYYYYYYYYXXXXXXX\n");
    EmitRecomputeAddresses(stdout,xfun);
    EmitGenerateCode(stdout,xfun);

    PRINT_FUN_END1();
}

/*=======================================================
=======================================================*/




