/*=======================================================
  Author:   R. Muth
  Version:  $Id: def.c,v 1.8 1999/04/18 19:06:21 muth Exp $
  Comment:  Single UseDef Chain Creation
=======================================================*/


#include "global.h" 
#include "functions.h"
#include "opcode.h"
#include "regs.h"
#include "flags.h"

#include <string.h>

/*=======================================================
=======================================================*/

#define ABBL_USEDEF_IN(i)  ABBL_U1(i).doubleword.d1
#define ABBL_USEDEF_DEF(i) ABBL_U1(i).doubleword.d2

#define USEDEF_BOT -69
#define USEDEF_TOP -1

#define FI_HAS_BEEN_PROCESSED F_LOCAL_1
#define FI_CANNOT_BE_PROCESSED F_LOCAL_2

#define AINS_USE_DEF_POTENTIAL(i) AINS_U1(i).quadword
/*=======================================================
=======================================================*/
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "UseDefComputeJoin"

static void UseDefComputeJoin(INDEX ibbl, INDEX def)
{
    if( def > CONFIG_NINS + (CONFIG_UBBL * 64) + 63 )
        printf("weird def %x propagated to bbl %d\n",def,ibbl);
    
        
    if( def == ABBL_USEDEF_IN(ibbl) ) return;    
    if( def == USEDEF_BOT ) return;
    if( ABBL_USEDEF_IN(ibbl) == USEDEF_TOP ) return;

    ABBL_MARK(ibbl);
    
    if( def == USEDEF_TOP )
    {
        ABBL_USEDEF_IN(ibbl) = def;
        return;
    }

    if( ABBL_USEDEF_IN(ibbl) == USEDEF_BOT )
    {
        ABBL_USEDEF_IN(ibbl) = def;
        return;
    }

    ABBL_USEDEF_IN(ibbl) = USEDEF_TOP;
}

/*=======================================================
=======================================================*/

#define CFA1

#undef __PROC__
#define __PROC__ "UseDefPropagation"

static void UseDefPropagation(REG reg, BOOL fun_args)
{
    BOOL  flag;
    INDEX iins;
    INDEX ibbl;
    INDEX ifun;
    
    INDEX iedg;
    INDEX def;
        
    /* Initialize Fixpoint Computation */

    FOREACH_LIVE_BBL(ibbl)
    {
        ifun = ABBL_FUN(ibbl);
        
        if( AFUN_IS_PSEUDO(ifun) )
        {
            ABBL_USEDEF_IN(ibbl) = USEDEF_TOP;
            ABBL_USEDEF_DEF(ibbl) = USEDEF_TOP;
        }
        else if( AbblNpreds(ibbl) == 0 )
        {
            ABBL_USEDEF_IN(ibbl) = USEDEF_TOP;
            ABBL_USEDEF_DEF(ibbl) = USEDEF_BOT;
        }
        else
        {
            if( fun_args && ABBL_IS_FIRST(ibbl) )
            {
                    ABBL_USEDEF_IN(ibbl) =   CONFIG_NINS + (ibbl * 64) + reg; /* same trick as with aliasing */
            }
            else
            {
                ABBL_USEDEF_IN(ibbl)  = USEDEF_BOT;
            }
            
            ABBL_USEDEF_DEF(ibbl) = USEDEF_BOT;

            FOREACH_BBL_INS_R(ibbl,iins)
            {
                if( AINS_REG_C(iins) == reg )
                {
                    ABBL_USEDEF_DEF(ibbl) = iins;
                    break;
                }
            }
        }

        def =  ABBL_USEDEF_DEF(ibbl);
        if( def > CONFIG_NINS + (CONFIG_UBBL * 64) + 63 )
            printf("reg: %d weird def %x in bbl %d\n",reg,def,ibbl);

        def = ABBL_USEDEF_IN(ibbl);
        if( def > CONFIG_NINS + (CONFIG_UBBL * 64) + 63 )
            printf("reg: %d weird in %x in bbl %d should be  %d\n",reg,def,ibbl,(def-CONFIG_NINS)/64);
        
        ABBL_MARK(ibbl);
    }

    do
    {
        flag = 0;

        FOREACH_LIVE_BBL(ibbl)
        {
            
            if( !ABBL_IS_MARKED(ibbl) ) continue;
            
            ABBL_UNMARK(ibbl);
            flag = 1;


            /* COMPUTE OUT SET */

            def = ABBL_USEDEF_DEF(ibbl);
            if( def == USEDEF_BOT )
                def = ABBL_USEDEF_IN(ibbl);

            if( def == USEDEF_BOT ) continue;
            
            if( def > CONFIG_NINS + (CONFIG_UBBL * 64) + 63 )
                printf("weird def %x propagated from bbl %d\n",def,ibbl);
            
    
            /* PROPAGATE OUT SET TO INSETS */

            FOREACH_BBL_SUCC(ibbl,iedg)
            {
                INDEX tl = AEDG_TAIL(iedg);

                if( fun_args )
                {
                    if( ABBL_IS_FIRST(tl) )
                        continue;  /* do not change function init bbls */
                }
                
#ifdef CFA1
                /* propagate unchanged regs */
                   
                if( AEDG_TYPE(iedg) == ET_FUN_LINK )
                {
                    INDEX CallFun = ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl)));
                    REGLIST NotChanged = ~AFUN_REGS_CHANGED(CallFun);

                    if( ABBL_IS_CRIT_PATH(ibbl) || (REG2BIT(reg) & NotChanged) )
                    {
                        UseDefComputeJoin(tl,def);
                    }
                    continue;
                }
  
                /* propagate changed regs only */
                
                if( AEDG_TYPE(iedg) == ET_FUN_RETURN )
                {
                    INDEX CallFun = ABBL_FUN(ibbl);
                    REGLIST Changed = AFUN_REGS_CHANGED(CallFun);
                    if( ABBL_IS_CRIT_PATH(tl) || (REG2BIT(reg) & Changed) )
                    {
                        UseDefComputeJoin(tl,def);
                    }
                    
                    continue;
                }
#else
                if( AEDG_TYPE(iedg) == ET_FUN_LINK )
                {
                    continue;
                }
                
#endif
                UseDefComputeJoin(tl,def);
            }
            
        }
    }
    while( flag );

    /* update field */

    FOREACH_LIVE_FUN(ifun)
    {
        flag = 0;
        FOREACH_FUN_BBL(ifun,ibbl)
        {
            def = ABBL_USEDEF_IN(ibbl);

            if( def < -1 || def > CONFIG_NINS + (CONFIG_UBBL * 64) + 63 )
            {
                if( ABBL_TYPE(ibbl) != BT_EXIT )
                {
                    /* sometimes exit bbls become essentially unreachable */
                    
                    ERROR(("reg: %d weird def %8x  in bbl: %d\n",reg,def,ibbl));
                    flag = 1;
                }
            }
            
            FOREACH_BBL_INS(ibbl,iins)
            {
                if( AINS_REG_A(iins) == reg && !(AINS_FLAGS(iins) & FI_EVAL_A) )
                    AINS_A_EVAL(iins) = def;
                
                if( AINS_REG_B(iins) == reg && !(AINS_FLAGS(iins) & FI_EVAL_B) )
                    AINS_B_EVAL(iins) = def;
                
                if( AINS_REG_C(iins) == reg  )
                    def = iins;
            }
        }

        if( flag )
            PrintFunSuper(stdout,ifun);
    }    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ComUseDef"

extern void ComUseDef(FILE *fp,BOOL fun_args)
{
    NUMBER count=0;
    INDEX  iins;
    
    REG    reg;
    REGLIST reglist;
    
    clock_t time = clock();

    FOREACH_LIVE_INS(iins)
    {
        if( !(AINS_FLAGS(iins) & FI_EVAL_A) )
            AINS_A_EVAL(iins) = USEDEF_TOP;
        
        if( !(AINS_FLAGS(iins) & FI_EVAL_B) )
            AINS_B_EVAL(iins) = USEDEF_TOP;
    }

    reglist = REGS_ALL_BUT_ZERO;
    
    for(reg=0;reg<64;reg++)
    {
        if( REG2BIT(reg) & reglist )
        {
#if 0
            printf("Doing register %d\n",reg);
#endif       
            UseDefPropagation(reg,fun_args);
        }
        
    }

    FOREACH_LIVE_INS(iins)
    {
        if( !(AINS_FLAGS(iins) & FI_EVAL_A) )
            if( AINS_A_EVAL(iins) >=  0 )
                count++;
        
        if( !(AINS_FLAGS(iins) & FI_EVAL_B) )
            if( AINS_B_EVAL(iins) >=  0 )
                count++;
    }
        
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    

    InfoStatus(fp,"\t[ComUsedef] Nontop Usedefs: %d  {%d ms}\n",count,time);
}

/*=======================================================
=======================================================*/

static INT32 AinsUseDefPotential(INDEX iins)
{
    switch( AINS_TYPE(iins) )
    {
    case IT_ILD:   
    case IT_FLD:
    case IT_ULD:
    case IT_LLD:          

    case IT_IST:
    case IT_FST:
    case IT_UST:
    case IT_CST:
#if 0
    return 1;
#else
    return 0;
#endif
    
    case IT_ICM:    /* the potential actually depends on which reg is involved */
    case IT_FCM:

    case IT_ICP:
    case IT_FCP:
            
    case IT_LDA:

    case IT_IOP:
        
    case IT_F2I:
    case IT_I2F:
    case IT_LOG:
    case IT_SFT:
    case IT_EXT:
        return 2;

    case IT_MSC:
    case IT_BRA:
    case IT_PAL:
    case IT_F2R:
    case IT_RX:
    case IT_MEM:
        return 0;
        
    case IT_IBR:
    case IT_FBR:
    case IT_JMP:
        return 10;
        
    case IT_FM4: 
    case IT_FM8:
    case IT_FOP:
    case IT_FCY:
    return 4;
        
    case IT_FD4:
        return 30;
                
    case IT_FD8:
        return 60;
        
    case IT_MLL:
    case IT_MLQ:
    case IT_MLH:
        return 16;
        
    default:
        ERROR(("unexpected instruction type\n"));
        PrintIns(stdout,iins);
        return 0;
    }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsIsInteresting"

static BOOL AinsIsInteresting(INDEX iins)
{
    INDEX def;
    
    if( AINS_REG_C(iins) == REG_ZERO || AINS_REG_C(iins) == REG_FZERO )
        return 0;
    if( !(AINS_FLAGS(iins) & FI_EVAL_A) && (AINS_FLAGS(iins) & FI_EVAL_B) )
    {
        def = AINS_A_EVAL(iins);
        
        if( VALID(def) ) return 0;
    }
    else if( (AINS_FLAGS(iins) & FI_EVAL_A) && !(AINS_FLAGS(iins) & FI_EVAL_B) )
    {
        def = AINS_B_EVAL(iins);
        if( VALID(def) ) return 0;
    }
    else if( !(AINS_FLAGS(iins) & FI_EVAL_A) && !(AINS_FLAGS(iins) & FI_EVAL_B) && AINS_REG_A(iins) == AINS_REG_B(iins) )
    {
        def = AINS_B_EVAL(iins);
        if( VALID(def) ) return 0;
    }

    return 1;
}




/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ComInsDefPotential"

extern void ComInsDefPotential(FILE *fp, BOOL fun_args )
{
    INDEX iins;
    INDEX ibbl;
    INDEX ifun;
    INDEX def;
    NUMBER  count_unprocessed;
    NUMBER  count_processed;
    REG reg;
#if 0
    int i;
    INDEX jins;
#endif

    clock_t time = clock();

#if 1
    WARNING(("INIT\n"));
#endif

    FOREACH_LIVE_BBL(ibbl)
    {
        FOREACH_BBL_INS(ibbl,iins)
        {
            AINS_USE_DEF_POTENTIAL(iins) =  AinsUseDefPotential(iins) * ABBL_COUNT(ibbl);
            AINS_FLAGS(iins) &= ~FI_HAS_BEEN_PROCESSED;
        }
    }
    


#if 1
    WARNING(("PROPAGATE\n"));
#endif
    
    do
    {
#if 1
        WARNING(("PROPAGATE CLEAR\n"));
#endif
        /* clear */

        FOREACH_LIVE_INS_R(iins)
        {
            if( AINS_FLAGS(iins) & FI_HAS_BEEN_PROCESSED ) continue;
            AINS_FLAGS(iins) &= ~FI_CANNOT_BE_PROCESSED;            
        }

#if 1
        WARNING(("PROPAGATE CANDIDATES\n"));
#endif
        /* compute candidates */

        FOREACH_LIVE_INS_R(iins)
        {
            if( AINS_FLAGS(iins) & FI_HAS_BEEN_PROCESSED ) continue;

            if( !(AINS_FLAGS(iins) & FI_EVAL_A) && (AINS_FLAGS(iins) & FI_EVAL_B) )
            {
                def = AINS_A_EVAL(iins);
            }
            else if( (AINS_FLAGS(iins) & FI_EVAL_A) && !(AINS_FLAGS(iins) & FI_EVAL_B) )
            {
                def = AINS_B_EVAL(iins);
            }
            else if( !(AINS_FLAGS(iins) & FI_EVAL_A) && !(AINS_FLAGS(iins) & FI_EVAL_B) && AINS_REG_A(iins) == AINS_REG_B(iins) )
            {
                def = AINS_B_EVAL(iins);
            }
            else
            {
                def = INVALID;
            }
            
                        
            if( def <  CONFIG_NINS ) 
                if( VALID(def) && !(AINS_FLAGS(def) & FI_HAS_BEEN_PROCESSED) )
                {
                    AINS_FLAGS(def) |= FI_CANNOT_BE_PROCESSED;
                }
        }

#if 1
        WARNING(("PROPAGATE PROPAGATE\n"));
#endif
        FOREACH_LIVE_INS_R(iins)
        {
            /* if an instruction has not been processed its value has not stabalized
               and we can therfore not propagate that value
               */
            
            if( !(AINS_FLAGS(iins) & FI_HAS_BEEN_PROCESSED) ) continue;

            if( !(AINS_FLAGS(iins) & FI_EVAL_A) && (AINS_FLAGS(iins) & FI_EVAL_B) )
            {
                def = AINS_A_EVAL(iins);
            }
            else if( (AINS_FLAGS(iins) & FI_EVAL_A) && !(AINS_FLAGS(iins) & FI_EVAL_B) )
            {
                def = AINS_B_EVAL(iins);
            }
            else if( !(AINS_FLAGS(iins) & FI_EVAL_A) && !(AINS_FLAGS(iins) & FI_EVAL_B) && AINS_REG_A(iins) == AINS_REG_B(iins) )
            {
                def = AINS_B_EVAL(iins);
            }
            else
            {
                def = INVALID;
            }
            

            if( def < CONFIG_NINS ) 
                if( VALID(def) && !(AINS_FLAGS(def) & FI_HAS_BEEN_PROCESSED) && !(AINS_FLAGS(def) & FI_CANNOT_BE_PROCESSED) )
                {
                    /* we do not propagate the benefit of load upward because usually knowing the
                       address of a load does not help us finding the value stored at that address */
                       
                    if( AINS_TYPE(iins) != IT_ILD &&
                        AINS_TYPE(iins) != IT_FLD &&
                        AINS_TYPE(iins) != IT_ULD &&
                        AINS_TYPE(iins) != IT_LLD &&
                        AINS_TYPE(iins) != IT_IST &&
                        AINS_TYPE(iins) != IT_FST &&
                        AINS_TYPE(iins) != IT_UST &&
                        AINS_TYPE(iins) != IT_CST )
                        AINS_USE_DEF_POTENTIAL(def) += AINS_USE_DEF_POTENTIAL(iins);
                }
        }

#if 1
        WARNING(("PROPAGATE CLEANUP\n"));
#endif

        count_processed = 0 ;
        count_unprocessed = 0;
        

        FOREACH_LIVE_INS_R(iins)
        {
            if( AINS_FLAGS(iins) & FI_HAS_BEEN_PROCESSED ) continue;
            if( AINS_FLAGS(iins) & FI_CANNOT_BE_PROCESSED )
            {
                count_unprocessed++;

                continue;
            }
            
            AINS_FLAGS(iins) |= FI_HAS_BEEN_PROCESSED;
            count_processed++;
        }

        printf("unpr %d   prc %d\n",count_unprocessed,count_processed);
        
        if( count_unprocessed > 0 && count_processed == 0 )
        {
            printf("circle detected\n");
	    break;
        }
        
    }
    while( count_unprocessed > 0 );

    
#if 1
    WARNING(("PROPAGATE FUN_ARG\n"));
#endif
    
    /* now handle functions arg seperately */
        
    if( fun_args )
    {
        FOREACH_LIVE_FUN(ifun)
        {
            
            for( reg=0; reg <64; reg++ )
            {
                AFUN_USE_DEF_ARG_POTENTIAL(ifun,reg) = 0;
            }
        }
        
        FOREACH_LIVE_INS(iins)
        {
            if( AINS_TYPE(iins) == IT_ILD ||
                AINS_TYPE(iins) == IT_FLD ||
                AINS_TYPE(iins) == IT_ULD ||
                AINS_TYPE(iins) == IT_LLD ||
                AINS_TYPE(iins) == IT_IST ||
                AINS_TYPE(iins) == IT_FST ||
                AINS_TYPE(iins) == IT_UST ||
                AINS_TYPE(iins) == IT_CST )
            continue;
                    
            if( !(AINS_FLAGS(iins) & FI_EVAL_A) && (AINS_FLAGS(iins) & FI_EVAL_B) )
            {
                def = AINS_A_EVAL(iins);
            }
            else if( (AINS_FLAGS(iins) & FI_EVAL_A) && !(AINS_FLAGS(iins) & FI_EVAL_B) )
            {
                def = AINS_B_EVAL(iins);
            }
            else if( !(AINS_FLAGS(iins) & FI_EVAL_A) && !(AINS_FLAGS(iins) & FI_EVAL_B) && AINS_REG_A(iins) == AINS_REG_B(iins) )
            {
                def = AINS_B_EVAL(iins);
            }
            else
            {
                def = INVALID;
            }
            
            if( !VALID(def) ) continue;
            if( def <  CONFIG_NINS )  continue;
            
            ibbl = (def - CONFIG_NINS) / 64;

            reg  = (def - CONFIG_NINS) % 64;
            ifun = ABBL_FUN(ibbl);
            
            AFUN_USE_DEF_ARG_POTENTIAL(ifun,reg) += AINS_USE_DEF_POTENTIAL(iins);
        }
    }
        

#if 1
    WARNING(("ACCOUNT FOR LOADS/STORES\n"));
#endif

    /* we do not want to propagate the potential for load stores since this
       would make it appear that we could statically evaluate eg load chains
       which we most likely cannot.
       However there is a benefit of of knowing the base address of a load or store
        for aliasing!
        */

    FOREACH_LIVE_INS(iins)
    {
        AINS_USE_DEF_POTENTIAL(iins) -=  AinsUseDefPotential(iins) * ABBL_COUNT(AINS_BBL(iins));
            
        if( AINS_TYPE(iins) != IT_ILD &&
            AINS_TYPE(iins) != IT_FLD &&
            AINS_TYPE(iins) != IT_ULD &&
            AINS_TYPE(iins) != IT_LLD &&
            AINS_TYPE(iins) != IT_IST &&
            AINS_TYPE(iins) != IT_FST &&
            AINS_TYPE(iins) != IT_UST &&
            AINS_TYPE(iins) != IT_CST )
            continue;
            
        if( !(AINS_FLAGS(iins) & FI_EVAL_A) && (AINS_FLAGS(iins) & FI_EVAL_B) )
        {
            def = AINS_A_EVAL(iins);
        }
        else if( (AINS_FLAGS(iins) & FI_EVAL_A) && !(AINS_FLAGS(iins) & FI_EVAL_B) )
        {
            def = AINS_B_EVAL(iins);
        }
        else if( !(AINS_FLAGS(iins) & FI_EVAL_A) && !(AINS_FLAGS(iins) & FI_EVAL_B) && AINS_REG_A(iins) == AINS_REG_B(iins) )
        {
            def = AINS_B_EVAL(iins);
        }
        else
        {
            def = INVALID;
        }
        
        if( !VALID(def) ) continue;

        if( def <  CONFIG_NINS )
        {
            AINS_USE_DEF_POTENTIAL(def) += ABBL_COUNT(AINS_BBL(iins));
        }
        else
        {
            ibbl = (def - CONFIG_NINS) / 64;
            
            reg  = (def - CONFIG_NINS) % 64;
            ifun = ABBL_FUN(ibbl);
            AFUN_USE_DEF_ARG_POTENTIAL(ifun,reg) += ABBL_COUNT(AINS_BBL(iins));
        }
        
    }
        
#if 1
    WARNING(("PRINT INTERESTING INSS AND FARGS\n"));
#endif
    
    FOREACH_LIVE_INS(iins)
    {
        AINS_FLAGS(iins) &= ~F_MARKED;
    }


    FOREACH_LIVE_FUN(ifun)
    {
        for( reg=0; reg <64; reg++ )
            AFUN_ARG_FLAGS(ifun,reg) &= ~F_MARKED;
    }
    

    
    FOREACH_LIVE_FUN(ifun)
    {
        BOOL flag = 0;

        for( reg=0; reg <64; reg++ )
        {
            if( reg==REG_SP || reg==REG_ZERO || reg==REG_FZERO) continue;
            
            if( AFUN_USE_DEF_ARG_POTENTIAL(ifun,reg) / (2+ABBL_COUNT(AFUN_BBL_FIRST(ifun))) < 25 ) continue;

            AFUN_ARG_FLAGS(ifun,reg) |= F_MARKED;
            
            flag = 1;
            
            printf("\t[*]fun: %d [%s] reg %d  pot: %ld\n",
                   ifun,AFUN_NAME(ifun),reg,AFUN_USE_DEF_ARG_POTENTIAL(ifun,reg));
        }

        FOREACH_FUN_BBL(ifun,ibbl)
        {
            if( !ABBL_IS_HOT(ibbl) ) continue;
        
            FOREACH_BBL_INS(ibbl,iins)
            {
                if( AINS_USE_DEF_POTENTIAL(iins) / (2+ABBL_COUNT(ibbl)) < 25 ) continue;
                if ( !AinsIsInteresting(iins) ) continue;

                AINS_FLAGS(iins) |= F_MARKED;
                printf("\t[*]fun: %d [%s] ins %d    bbl: %d   pot:  %ld\n",
                       ifun,AFUN_NAME(ifun),iins,ibbl,AINS_USE_DEF_POTENTIAL(iins));
                PrintIns(stdout,iins);
                
                flag = 1;
            }
        }
        
/*         if( flag ) */
/*             PrintFunSuper(stdout,ifun); */
    }

    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
}

/*=======================================================
=======================================================*/

static INDEX FindRootDef(INDEX iins)
{
    if( iins >=  CONFIG_NINS )
        return iins;
    
    if( !(AINS_FLAGS(iins) & FI_EVAL_A) && (AINS_FLAGS(iins) & FI_EVAL_B) )
    {
        return FindRootDef( AINS_A_EVAL(iins) );
    }
    else if( (AINS_FLAGS(iins) & FI_EVAL_A) && !(AINS_FLAGS(iins) & FI_EVAL_B) )
    {
        return FindRootDef( AINS_B_EVAL(iins) );
    }
    else if( !(AINS_FLAGS(iins) & FI_EVAL_A) && !(AINS_FLAGS(iins) & FI_EVAL_B) && AINS_REG_A(iins) == AINS_REG_B(iins) )
    {
        return FindRootDef( AINS_B_EVAL(iins) );
    }
    else
    {
        return iins;
    }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ValProfMarkImportantUsesTransitive"

extern void ValProfMarkImportantUsesTransitive(INDEX valprof)
{

    INDEX iins;

    FOREACH_LIVE_INS(iins)
    {
        AINS_FLAGS(iins) &= ~F_MARKED;
    }
    
    FOREACH_LIVE_INS(iins)
    {
        if( !(AINS_FLAGS(iins) & FI_EVAL_A) && VALID(AINS_A_EVAL(iins)) )
        {
            if( valprof != FindRootDef(AINS_A_EVAL(iins)) )
                continue;
            
            AINS_FLAGS(iins) |= F_MARKED;
        }
        
        if( !(AINS_FLAGS(iins) & FI_EVAL_B) && VALID(AINS_B_EVAL(iins)) )
        {
            if( valprof != FindRootDef(AINS_B_EVAL(iins)) )
                continue;
            
            AINS_FLAGS(iins) |= F_MARKED;
        }
    }
}

/*=======================================================
=======================================================*/
