/*=======================================================
=======================================================*/

#include "global.h"
#include "flags.h"
#include "functions.h"
#include "regs.h"
#include "opcode.h"

#define ABBL_SORTED(i)       ABBL_U1(i).doubleword.d1
#define ABBL_ID_EXACT(i)     ABBL_U2(i).doubleword.d1
#define ABBL_ID_SIMILAR(i)   ABBL_U2(i).doubleword.d2
#define ID_EXACT_N(i)    ABBL_U3(i).doubleword.d1
#define ID_SIMILAR_N(i)  ABBL_U3(i).doubleword.d2

#define ABBL_REGS_SCRATCH(i) ABBL_U3(i).quadword

/*=======================================================
This needs to be transitive !!!!
  =======================================================*/
#undef __PROC__
#define __PROC__ "AbblCmp"

static int AbblCmp(INDEX ibbl, INDEX jbbl)
{
    INDEX iins;
    INDEX jins;
    NUMBER n;
    
    n = AbblNInss(ibbl) - AbblNInss(jbbl);
    if( n != 0 ) return n;

    for(iins=ABBL_INS_FIRST(ibbl),jins=ABBL_INS_FIRST(jbbl); VALID(iins); iins=AINS_NEXT(iins),jins=AINS_NEXT(jins) )
    {
        if( AINS_CODE(iins) != AINS_CODE(jins) )
            return AINS_CODE(iins) - AINS_CODE(jins);
    }

    for(iins=ABBL_INS_FIRST(ibbl),jins=ABBL_INS_FIRST(jbbl); VALID(iins); iins=AINS_NEXT(iins),jins=AINS_NEXT(jins) )
    {
        if( AINS_DATA(iins) != AINS_DATA(jins) )
            return AINS_DATA(iins) - AINS_DATA(jins);
    }
    
    for(iins=ABBL_INS_FIRST(ibbl),jins=ABBL_INS_FIRST(jbbl); VALID(iins); iins=AINS_NEXT(iins),jins=AINS_NEXT(jins) )
    {
        if( AINS_REG_A(iins) != AINS_REG_A(jins) )
            return AINS_REG_A(iins) - AINS_REG_A(jins);
    }

    for(iins=ABBL_INS_FIRST(ibbl),jins=ABBL_INS_FIRST(jbbl); VALID(iins); iins=AINS_NEXT(iins),jins=AINS_NEXT(jins) )
    {
        if( AINS_REG_B(iins) != AINS_REG_B(jins) )
            return AINS_REG_B(iins) - AINS_REG_B(jins);
    }

    for(iins=ABBL_INS_FIRST(ibbl),jins=ABBL_INS_FIRST(jbbl); VALID(iins); iins=AINS_NEXT(iins),jins=AINS_NEXT(jins) )
    {
        if( AINS_REG_C(iins) != AINS_REG_C(jins) )
            return AINS_REG_C(iins) - AINS_REG_C(jins);
    }
    
    return 0;
}

     

/*=======================================================
  This needs to be transitive !!!!
  It took me quite a while to get it transitive so don't
  make changes without thinking real hard!
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblCmpUsedef"

static int AbblCmpUsedef(INDEX ibbl, INDEX jbbl)
{
    INDEX iins;
    INDEX jins;
    NUMBER n;
    
    NUMBER   idef[65];  /* includes REG_LIT */
    NUMBER   jdef[65];  /* includes REG_LIT */

    n = AbblNInss(ibbl) - AbblNInss(jbbl);

    if( n != 0 ) return n;
    

    for(iins=ABBL_INS_FIRST(ibbl),jins=ABBL_INS_FIRST(jbbl); VALID(iins); iins=AINS_NEXT(iins),jins=AINS_NEXT(jins) )
    {
        if( AINS_CODE(iins) != AINS_CODE(jins) )
            return AINS_CODE(iins) - AINS_CODE(jins);
    }

    for(iins=ABBL_INS_FIRST(ibbl),jins=ABBL_INS_FIRST(jbbl); VALID(iins); iins=AINS_NEXT(iins),jins=AINS_NEXT(jins) )
    {
        if( AINS_DATA(iins) != AINS_DATA(jins) )
            return AINS_DATA(iins) - AINS_DATA(jins);
    }

    /* check renaming */
    
    for( n=0; n<65; n++)
        idef[n] = -1;

    idef[REG_ZERO] = -2;
    idef[REG_FZERO] = -3;
    idef[REG_LIT] = -4;
    
    for( n=0; n<65; n++)
        jdef[n] = -1;
    
    jdef[REG_ZERO] = -2;
    jdef[REG_FZERO] = -3;
    jdef[REG_LIT] = -4;
    
    for(n=0,iins=ABBL_INS_FIRST(ibbl),jins=ABBL_INS_FIRST(jbbl); VALID(iins); n++,iins=AINS_NEXT(iins),jins=AINS_NEXT(jins) )
    {
        if( idef[AINS_REG_A(iins)] != jdef[AINS_REG_A(jins)] )
            return idef[AINS_REG_A(iins)] - jdef[AINS_REG_A(jins)];
        
        if( idef[AINS_REG_B(iins)] != jdef[AINS_REG_B(jins)] )
            return idef[AINS_REG_B(iins)] - jdef[AINS_REG_B(jins)];
        
        if( AINS_REG_C(iins) != REG_ZERO &&  AINS_REG_C(iins) != REG_FZERO )
        {
            idef[AINS_REG_C(iins)] = n;
        }

        if( AINS_REG_C(jins) != REG_ZERO &&  AINS_REG_C(jins) != REG_FZERO )
        {
            jdef[AINS_REG_C(jins)] = n;
        }
    }

    return 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "cmp_similar"

static int cmp_similar(const void *x1,const void *x2)
{
    int ret = AbblCmpUsedef( ((QUADWORD *)x1)->doubleword.d1, ((QUADWORD *)x2)->doubleword.d1);
    return  ret != 0 ? ret : AbblCmp( ((QUADWORD *)x1)->doubleword.d1, ((QUADWORD *)x2)->doubleword.d1);
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "cmp_count"

static int cmp_count(const void *x1,const void *x2)
{
    INDEX ibbl;
    INDEX jbbl;

    ibbl = ((QUADWORD *)x1)->doubleword.d1;
    jbbl = ((QUADWORD *)x2)->doubleword.d1;

    if( ABBL_ID_SIMILAR(ibbl) != ABBL_ID_SIMILAR(jbbl) )
    {
        return ABBL_ID_SIMILAR(ibbl) - ABBL_ID_SIMILAR(jbbl);
    }

    if( ID_EXACT_N(ABBL_ID_EXACT(ibbl)) != ID_EXACT_N(ABBL_ID_EXACT(jbbl)) )
    {
        return ID_EXACT_N(ABBL_ID_EXACT(jbbl)) - ID_EXACT_N(ABBL_ID_EXACT(ibbl));
    }
    
    if( ABBL_ID_EXACT(ibbl) != ABBL_ID_EXACT(jbbl) )
    {
        return ABBL_ID_EXACT(ibbl) - ABBL_ID_EXACT(jbbl);
    }

    return 0;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FactorSeparateBranches"

extern void FactorSeparateBranches(FILE *fp)
{
    INDEX ibbl;
    INDEX jbbl;
    INDEX iins;
    NUMBER ncount = 0;
    BOOL verbose = HasOption(PRINT_OPT_VERBOSE,"Factorseparatebranches");

    /* PrintFunSuper(stdout,378); */

    FOREACH_LIVE_BBL_R(ibbl)
    {
        if( AbblNInss(ibbl) < 2 ) continue;
            
        switch (ABBL_TYPE(ibbl))
        {
        case BT_BRANCH:
        case BT_PAL:
        case BT_JUMP:   /* but we do not touch switches !!! */
        case BT_CALL:
            if( verbose )
            {
                INDEX ifun = ABBL_FUN(ibbl);
                fprintf(fp,"\t[*] type %c  bbl %d  fun %d [%s]\n",ABBL_TYPE(ibbl),ibbl,ifun,AFUN_NAME(ifun));
            }
#if 0            
            PrintBblSuper(stdout,ibbl);
#endif       
            ncount++;
            iins = ABBL_INS_LAST(ibbl);

            jbbl = AbblNew();
            ABBL_TYPE(jbbl) = ABBL_TYPE(ibbl);
            ABBL_TYPE(ibbl) = BT_NORM;
            ABBL_FUN(jbbl) = ABBL_FUN(ibbl);
            AbblInsertAfter(jbbl,ibbl);
            ABBL_COUNT(jbbl) = ABBL_COUNT(ibbl);
            AbblMoveSuccEdges(ibbl, jbbl);
            AedgLinkFancy(AedgNewFancy(ibbl, jbbl, ET_NORM));
            AinsMove(iins, jbbl);
#if 0
            PrintBblSuper(stdout,ibbl);
            PrintBblSuper(stdout,jbbl);
#endif
            break;
        }
    }
    
    InfoStatus(fp,"\t[Factorseparatebranches] %d]\n",ncount);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblCreateFactorFun"

static INDEX AbblCreateFactorFun(INDEX ibbl, REG reg)
{
    INDEX ifun = AfunNewFancy();
    INDEX init;
    INDEX iins;

    AFUN_NAME(ifun) = "bblfactor";
    
    CASSERT( ABBL_TYPE(ibbl) == BT_NORM,("weird\n"));
    
    init = AbblDup(ibbl);
    ABBL_FUN(init) = ifun;
    ABBL_TYPE(init) = BT_RET;
    ABBL_COUNT(init) = 0;   /* NEEDS MORE WORK fix */
    
    iins = AinsNew();
    AinsMakeRet(iins,reg);
    AINS_BBL(iins) = init;
    
    AinsAppend(iins);
    AbblAppendFancy(init);
    
    AedgLinkFancy( AedgNewFancy(init, init, ET_EXIT_LINK) );
 
    return ifun;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FactorCleanupDataField"

extern void FactorCleanupDataField(FILE *fp)
{
    INDEX iins;
    
    FOREACH_LIVE_INS(iins)
    {
        switch( AINS_TYPE(iins) )
        {
        case IT_BRA:
        case IT_FBR:
        case IT_IBR:
        case IT_JMP:
            AINS_DATA(iins) = 0;
            break;
        }
    }

    InfoStatus(fp,"\t[Factorcleanupdatafield]: \n");
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FactorSortBblsCheckTansitivity"

static void FactorSortCheck(NUMBER n)
{
    NUMBER i;
    int ret;
    INDEX ibbl;
    INDEX jbbl;
    
    
    for( i=0; i<n-1; i++ )
    {
        ibbl = ABBL_SORTED(i);
        jbbl = ABBL_SORTED(i+1);
        ret = AbblCmpUsedef(ibbl,jbbl);
        if( ret == 0 )
        {
            if(  ABBL_ID_SIMILAR(ibbl) != ABBL_ID_SIMILAR(jbbl) )
            {
                ERROR(("weird %d %d -> %d\n",ABBL_ID_SIMILAR(ibbl),ABBL_ID_SIMILAR(jbbl),ret));
            }
            
        }
        else if( ret < 0 )
        {
            if( ABBL_ID_SIMILAR(ibbl) == ABBL_ID_SIMILAR(jbbl) )
            {
                ERROR(("weird %d %d -> %d\n",ABBL_ID_SIMILAR(ibbl),ABBL_ID_SIMILAR(jbbl),ret));
            }
        }
        else
        {
            ERROR(("weird %d %d -> %d\n",ABBL_ID_SIMILAR(ibbl),ABBL_ID_SIMILAR(jbbl),ret));
        }
    }
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FactorSortBbls"

static NUMBER FactorSortBbls(FILE *fp)
{
    NUMBER n;
    NUMBER id_exact;     
    NUMBER id_similar;
    NUMBER i;
    INDEX ibbl;
    INDEX jbbl;
    
    clock_t time = clock();
        
    n=0;
    FOREACH_LIVE_BBL(ibbl)
    {
        ABBL_SORTED(n) = ibbl;
        n++;
    }

    qsort( &ABBL_U1(0),n,sizeof(QUADWORD),cmp_similar);

    ABBL_SORTED(n) = 0;  /* end marker */
    
    id_exact = 0;
    id_similar = 0;
    ABBL_ID_EXACT(ABBL_SORTED(0)) = id_exact;
    ABBL_ID_SIMILAR(ABBL_SORTED(0)) = id_similar;

    for(i=0; i<n; i++)
    {
        ID_SIMILAR_N(i) = 0;
        ID_EXACT_N(i) = 0;
    }
    
    for(i=1; i<n; i++)
    {
        ibbl = ABBL_SORTED(i);
        jbbl = ABBL_SORTED(i-1);

        if( AbblCmp(ibbl,jbbl) )
            id_exact++;
        ABBL_ID_EXACT(ibbl) = id_exact;

        ID_EXACT_N(id_exact)++;
        
        
        if( AbblCmpUsedef(ibbl,jbbl) )
            id_similar++;
        ABBL_ID_SIMILAR(ibbl) = id_similar;

        ID_SIMILAR_N(id_similar)++;
                
        if( ABBL_ID_EXACT(ibbl) == ABBL_ID_EXACT(jbbl) && ABBL_ID_SIMILAR(ibbl) != ABBL_ID_SIMILAR(jbbl) )
        {
            ERROR(("weird\n"));
#if 1
            PrintBblSuper(stdout,ibbl);
            PrintBblSuper(stdout,jbbl);
#endif
        }         
     }

    FactorSortCheck(n);
        
    /* now sort again to make id_exact with a high count appear before those with a low count */

    n=0;
    FOREACH_LIVE_BBL(ibbl)
    {
        ABBL_SORTED(n) = ibbl;
        n++;
    }

    qsort( &ABBL_U1(0),n,sizeof(QUADWORD),cmp_count);
    FactorSortCheck(n);

#if 0
    for(i=1; i<n; i++)
    {
        ibbl = ABBL_SORTED(i);
        if( AbblNInss(ibbl) < 5 ) continue;
        printf("%5d ins %2d  bbl %5d  similar %5d (%5d)  exact %5d (%5d)\n",
               i,AbblNInss(ibbl),ibbl,
               ABBL_ID_SIMILAR(ibbl), ID_SIMILAR_N(ABBL_ID_SIMILAR(ibbl)),
               ABBL_ID_EXACT(ibbl), ID_EXACT_N(ABBL_ID_EXACT(ibbl)));
    }
#endif
    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    InfoStatus(fp,"\t[FactorSortbbls]:  bbls %d  identical %d  similar %d {%ldms}\n",n,id_exact,id_similar,time);
    return n;
}


/*=======================================================
@FUN
Make the register allocation of ibbl identical to jbbl
allowing at most N move instruction to be inserted
requires liveness!
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblRegallocAssimilate"

#define AINS_SIBLING(i) AINS_U1(i).doubleword.d1

static NUMBER AbblRegallocAssimilate(INDEX ibbl, INDEX jbbl, NUMBER maxins)
{
    NUMBER nbypass = 0;
    NUMBER npostmove = 0;
    NUMBER npremove = 0;
    REG premove[64];
    REG postmove[64];
    REG bypass[64];
    REG ireg;
    REG jreg;
    REGLIST regs;
    REGLIST avail;
    INDEX iins;
    INDEX jins;
    INDEX nbbl;
    INDEX idef[64];
    INDEX jdef[64];


    for(ireg=0; ireg<64; ireg++ )
    {
        premove[ireg] = INVALID;
        postmove[ireg] = INVALID;
        bypass[ireg] = INVALID;
        idef[ireg] = INVALID;
        jdef[ireg] = INVALID;
    }
    
    for(iins=ABBL_INS_FIRST(ibbl),jins=ABBL_INS_FIRST(jbbl); VALID(iins); iins=AINS_NEXT(iins),jins=AINS_NEXT(jins) )
    {
        AINS_SIBLING(iins) = jins;

        /* do register A */
        
        ireg = AINS_REG_A(iins);
        jreg = AINS_REG_A(jins);

        if( ireg == REG_FZERO  || ireg == REG_ZERO || ireg == REG_LIT )
        {
            if( jreg != ireg )
            {
                printf("BAD: reg a (admin) problem in ins %d/%d\n",iins,jins);
                return INVALID;
            }
        }
        else
        {
            if( VALID(idef[ireg]) &&  VALID(jdef[jreg]) )
            {
                if( AINS_SIBLING(idef[ireg]) != jdef[jreg] )
                {
                    printf("BAD: reg a (different defs) problem in ins %d/%d    %d/%d=%d \n",
                           iins,jins,idef[ireg],jdef[jreg], AINS_SIBLING(idef[ireg]));
                    return INVALID;
                }
                
            }
            else if( !VALID(idef[ireg]) &&  !VALID(jdef[jreg]) )
            {
                if( VALID(premove[jreg]) )
                {
                    if( premove[jreg] != ireg )
                    {
                        return INVALID;
                    }
                }
                else 
                {
                    premove[jreg] = ireg;
                }
                
            }
            else
            {
                printf("BAD: reg a problem in ins %d/%d\n",iins,jins);
                return INVALID;
            }
        }
        
        /* do register B */

        ireg = AINS_REG_B(iins);
        jreg = AINS_REG_B(jins);
        
        if( ireg == REG_FZERO  || ireg == REG_ZERO || ireg == REG_LIT )
        {
            if( jreg != ireg )
            {
                printf("BAD: reg b (admin) problem in ins %d/%d\n",iins,jins);
                return INVALID;
            }
        }
        else
        {
            if( VALID(idef[ireg]) &&  VALID(jdef[jreg]) )
            {
                if( AINS_SIBLING(idef[ireg]) != jdef[jreg] )
                {
                    printf("BAD: reg b (different defs) problem in ins %d/%d    %d/%d=%d \n",
                           iins,jins,idef[ireg],jdef[jreg], AINS_SIBLING(idef[ireg]));
                    return INVALID;
                }
            }
            else if( !VALID(idef[ireg]) &&  !VALID(jdef[jreg]) )
            {
                if( VALID(premove[jreg]) )
                {
                    if( premove[jreg] != ireg )
                    {
                        return INVALID;
                    }
                    
                }
                else
                {
                    premove[jreg] = ireg;
                }
                
            }
            else
            {
                printf("BAD: reg b problem in ins %d/%d\n",iins,jins);
                return INVALID;
            }
        }


        idef[AINS_REG_C(iins)]= iins;
        jdef[AINS_REG_C(jins)]= jins;
    }

    
    /*
      Now analyze the data collected and determine how many
      move instructions need to be inserted
      */
    
    for(ireg=0; ireg<64; ireg++ )
    {
        iins = idef[ireg];
        if( !VALID(iins) ) continue;
        if( !(REG2BIT(ireg) & ABBL_OUT(ibbl)) ) continue;
        jins = AINS_SIBLING(iins);
        jreg = AINS_REG_C(jins);
        if( jdef[jreg] != jins )
        {
            printf("BAD: reg %d (defined by %d) is overwritten later\n",jreg,jins);
            return INVALID;
        }

        /* if( !(REG2BIT(jreg) & ABBL_OUT(jbbl)) ) continue; */
        if( ireg == jreg ) continue;
        postmove[ireg] = jreg;
        npostmove++;
    }

    for(ireg=0; ireg<64; ireg++ )
    {
        if( !VALID(postmove[ireg]) ) continue;
        if( !VALID(postmove[postmove[ireg]]) ) continue;
        printf("BAD: chain problem %d -> %d -> %d\n",ireg,postmove[ireg],postmove[postmove[ireg]]);
        return INVALID;
    }
        
    /* pre move */
    for(ireg=0; ireg<64; ireg++ )
    {
        if( !VALID(premove[ireg]) ) continue;
        if( premove[ireg] == ireg )
        {
            premove[ireg] = INVALID;
            continue;
        }
        
        if( AbblRegsLiveIn(ibbl) & REG2BIT(ireg) )
        {
            printf("BAD: pre move %d->%d\n",premove[ireg],ireg);
            return INVALID;
        }
        npremove++;
    }

    for(ireg=0; ireg<64; ireg++ )
    {
        if( !VALID(premove[ireg]) ) continue;
        if( !VALID(premove[premove[ireg]]) ) continue;
        printf("BAD: chain problem %d -> %d -> %d\n",premove[premove[ireg]],premove[ireg],ireg);
        return INVALID;
    }

    /* by pass contains those register whose live range spans all of
       ibbl and which are defined in jbbl.
       they need to be moved to other registers when ibbl is assimilated to jbbl
       */
    
    regs = REGS_INT_BUT_ZERO & ABBL_OUT(ibbl) & ~AbblRegsDef(ibbl) & AbblRegsDef(jbbl);
    avail = REGS_INT_BUT_ZERO & ~AbblRegsDef(jbbl) & ~ABBL_OUT(ibbl) & ~AbblRegsUse(jbbl);
             
    for(ireg=0; ireg<32; ireg++ )
    {
        if( REG2BIT(ireg) & regs )
        {
            nbypass++;
            if( avail == REGS_NONE )
            {
                printf("BAD: no bypas reg found for %d\n",ireg);
                return INVALID;
            }

            jreg = RegsGetFirst(avail);
            bypass[ireg] = jreg;
            avail &= ~REG2BIT(jreg);
        }
    }

    regs = REGS_FLT_BUT_ZERO & ABBL_OUT(ibbl) & ~AbblRegsDef(ibbl) & AbblRegsDef(jbbl);
    avail = REGS_FLT_BUT_ZERO & ~AbblRegsDef(jbbl) & ~ABBL_OUT(ibbl) & ~AbblRegsUse(jbbl);
             
    for(ireg=32; ireg<64; ireg++ )
    {
        if( REG2BIT(ireg) & regs )
        {
            nbypass++;
            if( avail == REGS_NONE )
            {
                printf("BAD: no bypas reg found for %d\n",ireg);
                return INVALID;
            }

            jreg = RegsGetFirst(avail);
            bypass[ireg] = jreg;
            avail &= ~REG2BIT(jreg);
        }
    }


    if( nbypass+nbypass+npremove+npostmove > maxins )
    {
        printf("BAD: too many moves needed\n");
        return INVALID;
    }

#if 0

    /* PrintBblSuper(stdout,ibbl);*/


    for(ireg=0; ireg<64; ireg++ )
    {
        if( !VALID(postmove[ireg]) ) continue;
        if( postmove[ireg] == ireg ) continue;

        n-=1;
    }
#endif
    
    printf("#######################################\n");

    if( nbypass + npremove > 0 )
    {
        if( ABBL_IS_FIRST(ibbl) )
        {
            POTENTIAL(("not implemented because somewhat tricky\n"));
            return INVALID;
        }
        
        nbbl = AbblNew();
        ABBL_FUN(nbbl) = ABBL_FUN(ibbl);
        ABBL_COUNT(nbbl) = ABBL_COUNT(ibbl);
        ABBL_TYPE(nbbl) = BT_NORM;
        
        AbblMigrateRelocInfo(ibbl,nbbl);   /* very important */        
        AbblMovePredEdges(ibbl,nbbl);
        AbblInsertAfter(nbbl,ABBL_PREV(ibbl));
        AedgLinkFancy(AedgNewFancy(nbbl, ibbl, ET_NORM));

        /* bypass regs first !!! */
        
        for(ireg=0; ireg<64; ireg++ )
        {
            if( !VALID(bypass[ireg]) ) continue;
            iins = AinsNew();
            AINS_BBL(iins) = nbbl;
            if( REG_IS_INT(ireg) )
                AinsMakeLda(iins,ireg,bypass[ireg],0);
            else
                AinsMakeCpys(iins,ireg,ireg,bypass[ireg]);
            AinsAppend(iins);
            printf("bypass %2d -> %2d\n",ireg,bypass[ireg]);
        }

        for(ireg=0; ireg<64; ireg++ )
        {
            if( !VALID(premove[ireg]) ) continue;
            iins = AinsNew();
            AINS_BBL(iins) = nbbl;
            AinsMakeLda(iins,premove[ireg],ireg,0);
            AinsAppend(iins);
            printf("premove  %2d -> %2d\n",premove[ireg],ireg);
        }
    }

    if( nbypass + npostmove > 0 )
    {
        nbbl = AbblNew();
        ABBL_FUN(nbbl) = ABBL_FUN(ibbl);
        ABBL_COUNT(nbbl) = ABBL_COUNT(ibbl);
        ABBL_TYPE(nbbl) = BT_NORM;
        AbblInsertAfter(nbbl,ibbl);
        AbblMoveSuccEdges(ibbl, nbbl);
        AedgLinkFancy(AedgNewFancy(ibbl, nbbl, ET_NORM));

        /* bypass regs last !!! */
        
        for(ireg=0; ireg<64; ireg++ )
        {
            if( !VALID(postmove[ireg]) ) continue;
            iins = AinsNew();
            AINS_BBL(iins) = nbbl;
            AinsMakeLda(iins,postmove[ireg],ireg,0);
            AinsAppend(iins);
            printf("postmove %2d -> %2d\n",postmove[ireg],ireg);
        }

        for(ireg=0; ireg<64; ireg++ )
        {
            if( !VALID(bypass[ireg]) ) continue;
            iins = AinsNew();
            AINS_BBL(iins) = nbbl;
            if( REG_IS_INT(ireg) )
                AinsMakeLda(iins,bypass[ireg],ireg,0);
            else
                AinsMakeCpys(iins,bypass[ireg],bypass[ireg],ireg);
            AinsAppend(iins);
            printf("bypass %2d -> %2d\n",bypass[ireg],ireg);
        }
    }
    

    printf("#######################################\n");


    HURRAH(("renaming nins %d\n",AbblNInss(ibbl)));
    for(iins=ABBL_INS_FIRST(ibbl),jins=ABBL_INS_FIRST(jbbl); VALID(iins); iins=AINS_NEXT(iins),jins=AINS_NEXT(jins) )
    {
        AINS_REG_A(iins) = AINS_REG_A(jins);
        AINS_REG_B(iins) = AINS_REG_B(jins);
        AINS_REG_C(iins) = AINS_REG_C(jins);
    }
           
         
    return nbypass+nbypass+npremove+npostmove;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FactorRegRealloc"

extern void FactorRegRealloc(FILE *fp, NUMBER maxmoves)
{
    NUMBER n;
    NUMBER id_similar;
    NUMBER id_exact;    
    NUMBER i;
    NUMBER leader;
    NUMBER j;
    INDEX ibbl;
    INDEX jbbl;
    NUMBER x;
    NUMBER nmoves=0;
    NUMBER nbbls=0;

    BOOL verbose = HasOption(PRINT_OPT_VERBOSE,"Factorregrealloc");
    
    clock_t time = clock();

    /* sort bbls */

    n = FactorSortBbls(fp);

    id_similar = INVALID;
    leader = INVALID;
    
    for(i=0; i<n; i++ )
    {
        ibbl = ABBL_SORTED(i);
        if( ABBL_ID_SIMILAR(ibbl) != id_similar )
            leader = i;
        id_similar =  ABBL_ID_SIMILAR(ibbl);
        id_exact =  ABBL_ID_EXACT(ibbl);
        
        
        CASSERT( ID_SIMILAR_N(id_similar) > 0,("weird\n"));
        CASSERT( ID_EXACT_N(id_exact) > 0,("weird\n"));
        CASSERT( ID_EXACT_N(id_exact) <=  ID_SIMILAR_N(id_similar) ,("weird\n"));

        if( ABBL_TYPE(ibbl) != BT_NORM )
            continue;
                
        if( ID_SIMILAR_N(id_similar) == 1 )
            continue;

        if( ID_EXACT_N(id_exact) != 1 )
            continue;
        

#if 0
        if( maxmoves != 3 ) continue;

        if( AbblNInss(ibbl) != 7 ) continue;
                
        if( id_exact > PRINT_OPT_NUM ) continue;
        

        if( AbblNInss(ibbl) < maxmoves+3+6 )
            continue;
#endif
        if( AbblNInss(ibbl) < maxmoves+3 )
            continue;
        printf("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
        printf("Doing ibbl %d  ins %2d   similar  %4d  (%4d)   exact  %4d  (%4d)\n",
               ibbl, AbblNInss(ibbl), id_similar,ID_SIMILAR_N(id_similar),id_exact,ID_EXACT_N(id_exact));
        printf("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
        
        /* PrintBblSuper(stdout,ibbl); */
        
        for(j=leader;j<n ;j++)
        {
            jbbl = ABBL_SORTED(j);
            if( ABBL_ID_SIMILAR(jbbl) != id_similar ) break;
            if( ABBL_ID_EXACT(jbbl) == id_exact ) continue;
            if( ibbl == jbbl ) continue;
            CASSERT( ABBL_ID_EXACT(jbbl) != id_exact,("weird\n"));

            /* PrintBblSuper(stdout,jbbl); */
            
            x = AbblRegallocAssimilate(ibbl, jbbl, maxmoves);
            if( VALID(x) )
            {
                ID_EXACT_N( ABBL_ID_EXACT(ibbl) ) --;
                ID_EXACT_N( ABBL_ID_EXACT(jbbl) ) ++;
                ABBL_ID_EXACT(ibbl) = ABBL_ID_EXACT(jbbl);

                nmoves += x;
                nbbls++;
                break;
            }
        }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    InfoStatus(fp,"\t[FactorRegRealloc]: Bbls %d  Realloc %d  Moves %d Maxmoves %d {%ldms}\n",n,nbbls,nmoves,maxmoves,time);

}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FactorBbls"

extern void FactorBbls(FILE *fp)
{
    NUMBER n;
    NUMBER id;
    NUMBER i;
    NUMBER j;
    INDEX ifun;
    INDEX ibbl;
    INDEX jbbl;
    INDEX iins;
    NUMBER nfun=0;
    NUMBER nbbl=0;
    NUMBER nins=0;
    NUMBER the_regs[64];
    
    REGLIST  regs;
    REG reg;
    REG best;

    BOOL verbose = HasOption(PRINT_OPT_VERBOSE,"Factorbbls");
    
    clock_t time = clock();

    /* sort bbls */

    n = FactorSortBbls(fp);


    /* unmark unprocessed bbls */
    
    FOREACH_LIVE_BBL(ibbl)
    {
        ABBL_FLAGS(ibbl) &= ~F_MARKED;
    }

    /* find available return regs */
    
    FOREACH_LIVE_BBL(ibbl)
    {
        ABBL_REGS_SCRATCH(ibbl) = REGS_INT_BUT_ZERO & ~ABBL_OUT(ibbl) & ~AbblRegsDef(ibbl) & ~AbblRegsUse(ibbl);
    }
    
    /* outline */
    
    i=0;
    
    while( i<n )
    {        
        ibbl = ABBL_SORTED(i);
        id = ABBL_ID_EXACT(ibbl);

        if( AbblNInss(ibbl) > 2 &&  ABBL_TYPE(ibbl) == BT_NORM )
        {
            for(;;)
            {
                for(reg=0;reg<64;reg++)
                    the_regs[reg] = 0;
                
                for(j=i+1; j<n; j++ )
                {
                    ibbl = ABBL_SORTED(j);
                    if( ABBL_ID_EXACT(ibbl) != id ) break;
                    if( ABBL_FLAGS(ibbl) & F_MARKED ) continue;
                    
                    regs = ABBL_REGS_SCRATCH(ibbl);
                    for(reg=0;reg<64;reg++)
                    {
                        if( REG2BIT(reg) & regs )
                        {
                            the_regs[reg]++;
                        }
                    }
                }

                best = 0;
                for(reg=1;reg<64;reg++)
                {
                    if( the_regs[reg] > the_regs[best] )
                        best = reg;
                }
                
                if( the_regs[best] < 2 ) break;

                nfun++;
                nbbl += the_regs[best];

                ibbl =  ABBL_SORTED(i);
                printf("creating factor fun for id %d  bbl %d  iins %d\n",ABBL_ID_EXACT(ibbl),ibbl, AbblNInss(ibbl) + 1);
                
                ifun = AbblCreateFactorFun(ibbl, best);

                nins -= AbblNInss(ibbl) + 1;
                
                
                printf("doing bbls %d\n",the_regs[best]);
                for(j=i+1; j<n; j++ )
                {
                    ibbl = ABBL_SORTED(j);
                    if( ABBL_ID_EXACT(ibbl) != id ) break;
                    if( ABBL_FLAGS(ibbl) & F_MARKED ) continue;
                    if( !(REG2BIT(best) &  ABBL_REGS_SCRATCH(ibbl)) ) continue;

                    nins += AbblNInss(ibbl) - 1;

                    printf("doing bbl %d  id %d  iins %d\n",ibbl, ABBL_ID_EXACT(ibbl),AbblNInss(ibbl)-1);

                    FOREACH_BBL_INS(ibbl,iins)
                    {
                        AinsKill(iins);
                    }

                    
                    ABBL_TYPE(ibbl) = BT_CALL;
                    iins = AinsNew();
                    AinsMakeBsr(iins,best);
                    AINS_BBL(iins) = ibbl;
                    AinsAppend(iins);
                    jbbl = AEDG_TAIL(ABBL_EDG_NORM(ibbl));
                    AedgKillFancy(ABBL_EDG_NORM(ibbl));
                    AedgLinkFancy( AedgNewFancy(ibbl,jbbl,ET_FUN_LINK));
                    AedgLinkFancy( AedgNewFancy(ibbl,AFUN_BBL_FIRST(ifun),ET_FUN_CALL));
                    ABBL_FLAGS(ibbl)  |= F_MARKED;
                }
            }
        }

        /* skip to next id */
        
        for(i+=1; i < n && ABBL_ID_EXACT(ABBL_SORTED(i)) == id; i++ )
            ;        
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    InfoStatus(fp,"\t[FactorBbl] funs %d  bbls %d  nins %d {%ldms}\n",nfun,nbbl,nins,time);
}
/*=======================================================
=======================================================*/


