/*=======================================================
=======================================================*/

#include "global.h"
#include "flags.h"
#include "functions.h"
#include "regs.h"
#include "opcode.h"

#define ABBL_REGS_SAVED(i) ABBL_U1(i).quadword
#define ABBL_REGS_RESTORED(i) ABBL_U1(i).quadword
#define ABBL_REGS_DEAD_AT_INIT(i) ABBL_U2(i).quadword
/*=======================================================
=======================================================*/

static struct
{
    NUMBER Offset;
    REG    Prev;
    REG    Next;
}
SaveReg[32] =
{
    { INVALID, INVALID, INVALID},   /*  0 */
    { INVALID, INVALID, INVALID},   /*  1 */
    { INVALID, INVALID, INVALID},   /*  2 */
    { INVALID, INVALID, INVALID},   /*  3 */
    { INVALID, INVALID, INVALID},   /*  4 */
    { INVALID, INVALID, INVALID},   /*  5 */
    { INVALID, INVALID, INVALID},   /*  6 */
    { INVALID, INVALID, INVALID},   /*  7 */
    { INVALID, INVALID, INVALID},   /*  8 */
    { 8,       26,      10     },   /*  9 */
    { 16,       9,      11     },   /* 10 */
    { 24,      10,      12     },   /* 11 */
    { 32,      11,      13     },   /* 12 */
    { 40,      12,      14     },   /* 13 */
    { 48,      13,      15     },   /* 14 */
    { 56,      14,      INVALID},   /* 15 */
    { INVALID, INVALID, INVALID},   /* 16 */
    { INVALID, INVALID, INVALID},   /* 17 */
    { INVALID, INVALID, INVALID},   /* 18 */
    { INVALID, INVALID, INVALID},   /* 19 */
    { INVALID, INVALID, INVALID},   /* 20 */
    { INVALID, INVALID, INVALID},   /* 21 */
    { INVALID, INVALID, INVALID},   /* 22 */
    { INVALID, INVALID, INVALID},   /* 23 */
    { INVALID, INVALID, INVALID},   /* 24 */
    { INVALID, INVALID, INVALID},   /* 25 */
    { 0,       INVALID, 9      },   /* 26 */
    { INVALID, INVALID, INVALID},   /* 27 */
    { INVALID, INVALID, INVALID},   /* 28 */
    { INVALID, INVALID, INVALID},   /* 29 */
    { INVALID, INVALID, INVALID},   /* 30 */
    { INVALID, INVALID, INVALID},   /* 31 */
};

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "RegsHighestSavedRegByCallingConvention"

static REG RegsHighestSavedRegByCallingConvention(REGLIST regs)
{
    REG reg;

    for(reg=0;reg<32;reg++)
    {
        if( !(REG2BIT(reg) & regs) ) continue;
        if( !VALID(SaveReg[reg].Next) ) break;
        if( !(REG2BIT(SaveReg[reg].Next) & regs) ) break;
    }
    
    CASSERT( reg != 32,("weird\n")); 
    return reg;    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "RegsCleanupCallingConvention"

static REGLIST RegsCleanupCallingConvention(REGLIST regs)
{
    REG reg;
    BOOL change;
    
    do
    {
        change = 0;
        for(reg=0;reg<32;reg++)
        {
            REG prev;
            
            if( !(REG2BIT(reg) & regs) ) continue;
            prev = SaveReg[reg].Prev;
            
            if( VALID(prev) &&  !(REG2BIT(prev) & regs) )
            {
                change = 1;
                regs &= ~REG2BIT(reg);
            }
        }
    }
    while( change );

    return regs;
}

     
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunSavedRegsCallingConvention"

static REGLIST AbblSavedRegsCallingConvention(INDEX ibbl)
{
    REGLIST  regs_def = REGS_NONE;
    REGLIST  regs_saved = REGS_NONE;
    INDEX iins;
    REG  reg;

#if 0
    if( ifun < 410 ) return REGS_NONE;
    if( ifun > PRINT_OPT_NUM ) return REGS_NONE;
#endif
        
    FOREACH_BBL_INS(ibbl, iins)
    {        
        if( AinsDef(iins) & REG2BIT(REG_SP) )
        {
            CASSERT( AINS_CODE(iins) == I_lda &&
                     AINS_REG_B(iins) == REG_SP &&
                     AINS_REG_C(iins) == REG_SP,("weird\n"));
            
            if( regs_def & REG2BIT(REG_SP) ) 
                return REGS_NONE;
        }
        else if( AinsUse(iins) & REG2BIT(REG_SP) )
        {
            if( !(regs_def & REG2BIT(REG_SP)) ) 
                return REGS_NONE;
        }

        if( AINS_CODE(iins) == I_stq && AINS_REG_B(iins) == REG_SP )
        {
            reg = AINS_REG_A(iins);
            if( !(REG2BIT(reg) & regs_def) && AINS_DATA(iins) == SaveReg[reg].Offset )
            {
                regs_saved |= REG2BIT(reg);
            }
        }
        
        regs_def |= AinsDef(iins);
    }
    
    return RegsCleanupCallingConvention(regs_saved);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblKillSavedRegsCallingConvention"

static void AbblKillSavedRegsCallingConvention(INDEX ibbl,REGLIST regs)
{
    INDEX iins;
    REGLIST regs_def = REGS_NONE;
    REG reg;

    FOREACH_BBL_INS(ibbl, iins)
    {
        if( AINS_CODE(iins) == I_stq && AINS_REG_B(iins) == REG_SP )
        {
            reg = AINS_REG_A(iins);
            if( !(REG2BIT(reg) & regs_def) && (regs & REG2BIT(reg)) && AINS_DATA(iins) == SaveReg[reg].Offset )
            {
                regs &= ~REG2BIT(reg);
                AinsKill(iins);
                continue;
            }
            
        }
        
        regs_def |= AinsDef(iins);
    }

    CASSERT( regs == REGS_NONE,("weird\n"));
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblRestoresRegsCallingConvention"

static REGLIST AbblRestoredRegsCallingConvention(INDEX ibbl)
{
    REGLIST  regs_def = REGS_NONE;
    REGLIST  regs_use = REGS_NONE;
    REGLIST  regs_restored = REGS_NONE;
    INDEX iins;
    REG  reg;

#if 0
    if( ifun < 410 ) return REGS_NONE;
    if( ifun > PRINT_OPT_NUM ) return REGS_NONE;
#endif
        
    FOREACH_BBL_INS_R(ibbl, iins)
    {
        if( AINS_CODE(iins) == I_ret ) continue;  /* ignore return !*/
        
        if( AinsDef(iins) & REG2BIT(REG_SP) )
        {
            if( AINS_CODE(iins) != I_lda ||
                AINS_REG_B(iins) != REG_SP ||
                AINS_REG_C(iins) != REG_SP )
            {
                INDEX ifun = ABBL_FUN(ibbl);
                WARNING(("Strange stack adjustment in fun %d [%s]\n",ifun,AFUN_NAME(ifun)));
                return REGS_NONE;
            }
            
            
            if( regs_use & REG2BIT(REG_SP) ) 
                return REGS_NONE;
            if( regs_def & REG2BIT(REG_SP) ) 
                return REGS_NONE;
        }

        if( AINS_CODE(iins) == I_ldq && AINS_REG_B(iins) == REG_SP )
        {
            reg = AINS_REG_C(iins);
            if( !(REG2BIT(reg) & regs_def) &&  !(REG2BIT(reg) & regs_use) && AINS_DATA(iins) == SaveReg[reg].Offset )
            {
                regs_restored |= REG2BIT(reg);
            }
        }
        
        regs_def |= AinsDef(iins);
        regs_use |= AinsUse(iins);
    }

    return RegsCleanupCallingConvention(regs_restored);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblKillRestoresRegdCallingConvention"

static void AbblKillRestoredRegsCallingConvention(INDEX ibbl,REGLIST regs)
{
    INDEX iins;
    REGLIST regs_mod = REGS_NONE;
    REG reg;

    FOREACH_BBL_INS_R(ibbl, iins)
    {
        if( AINS_CODE(iins) == I_ldq && AINS_REG_B(iins) == REG_SP )
        {
            reg = AINS_REG_C(iins);
            if( !(REG2BIT(reg) & regs_mod) && (regs & REG2BIT(reg)) && AINS_DATA(iins) == SaveReg[reg].Offset )
            {
                regs &= ~REG2BIT(reg);
                AinsKill(iins);
                continue;
            }
            
        }
        
        regs_mod |= AinsUse(iins) | AinsDef(iins);
    }

    CASSERT( regs == REGS_NONE,("weird\n"));
}


/*=======================================================
@FAC
Needs liveness
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "FactorPrologs"

extern void FactorPrologs(FILE *fp)
{
    NUMBER benefit[32];
    NUMBER count[32];
    
    INDEX ifun;
    INDEX ibbl;
    INDEX jbbl;
    INDEX iins;
    INDEX factorbbl[32];

    REG the_reg;
    REG reg;
    BOOL verbose = HasOption(PRINT_OPT_VERBOSE,"Factorprologs");

    NUMBER pcount = 0;   /* factored prologs */
    NUMBER scount = 0;   /* factores saves */ 
    NUMBER acount = 0;
    NUMBER icount = 0;     

    clock_t time = clock();
 
    /* mark those functions where we cannot factor save regs */

    verbose = 1;


#if 0
    FOREACH_LIVE_FUN(ifun)
    {
        REGLIST regs;
        
        if (AFUN_IS_PSEUDO(ifun)) continue;

        printf("FUN %5d [%s]\n",ifun,AFUN_NAME(ifun));


        ibbl = AFUN_BBL_FIRST(ifun);
        printf("SAVE: ");
        regs = AbblSavedRegsCallingConvention(ibbl);
        for( reg=0; reg<64; reg++ )
        {
            if( REG2BIT(reg) & regs )
            {
                printf("%2d ",reg);
            }
            
        }
        printf("\n");

        FOREACH_FUN_BBL(ifun,ibbl)
        {
            if( ABBL_TYPE(ibbl) != BT_RET ) continue;
            printf("REST: ");
            regs = AbblRestoredRegsCallingConvention(ibbl);
            for( reg=0; reg<64; reg++ )
            {
                if( REG2BIT(reg) & regs )
                {
                    printf("%2d ",reg);
                }
                
            }
            printf("\n");
        }
    }
#endif
    
    
    FOREACH_LIVE_FUN(ifun)
    {
        ibbl = AFUN_BBL_FIRST(ifun);
        ABBL_REGS_SAVED(ibbl) = REGS_NONE;

        if (AFUN_IS_PSEUDO(ifun)) continue;

        ABBL_REGS_SAVED(ibbl) = AbblSavedRegsCallingConvention(ibbl);
        if(   BinaryBits(ABBL_REGS_SAVED(ibbl)) < 2 )
        {
            ABBL_REGS_SAVED(ibbl) = REGS_NONE;
            continue;
        }
        
        ABBL_REGS_DEAD_AT_INIT(ibbl) = REGS_INT_BUT_ZERO & ~AbblRegsLiveIn(ibbl); 
    }    
    
    for(;;)
    {
        WARNING(("round: %d\n",acount));
        for( reg=0; reg<32; reg++)
        {
            benefit[reg] = 0;
            benefit[reg] = 0;
            factorbbl[reg] = 0;
        }
        
        FOREACH_LIVE_FUN(ifun)
        {
            ibbl = AFUN_BBL_FIRST(ifun);
            if( ABBL_REGS_SAVED(ibbl) == REGS_NONE )  continue;
            
            for(reg=0;reg<32;reg++)
            {
                if( !(REG2BIT(reg) & ABBL_REGS_DEAD_AT_INIT(ibbl)) ) continue;
                benefit[reg] += BinaryBits( ABBL_REGS_SAVED(ibbl) );
                count[reg] += 1;
            }
        }

        the_reg = INVALID;
        
        for( reg=0; reg<32; reg++)
        {
            if( benefit[reg] == 0 ) continue;
            if( count[reg] < 2 ) continue;
            
            if( !VALID(the_reg) || benefit[reg] > benefit[the_reg] )
            {
                the_reg = reg;
            }
        }

        if( !VALID(the_reg) ) break;
        
        acount++;

        /* create new factor functions and bbls for register the_reg */
        
        for( reg = 0; reg<32; reg++ )
        {
            if( !VALID(SaveReg[reg].Offset) ) continue;
            
            ifun = AfunNewFancy();
            AFUN_NAME(ifun) = "$FactorSave";

            ibbl = AbblNew();
            ABBL_FUN(ibbl) = ifun;
            if( VALID(SaveReg[reg].Prev) )
                ABBL_TYPE(ibbl) = BT_NORM;
            else
                ABBL_TYPE(ibbl) = BT_RET;
            ABBL_COUNT(ibbl) = 1;                          /* NEEDS TO BE FIXED !!! */
            AbblAppendFancy(ibbl);
  
            iins = AinsNew();
            AinsMakeStq(iins, reg, REG_SP, SaveReg[reg].Offset);
            AINS_BBL(iins) = ibbl;
            AinsAppend(iins);
            icount--;

            
            if( !VALID(SaveReg[reg].Prev) )
            {
                iins = AinsNew();
                AinsMakeRet(iins, the_reg);
                AINS_BBL(iins) = ibbl;
                AinsAppend(iins);
                icount--;
            }
            
            factorbbl[reg] = ibbl;
        }

        /* create edges */
        
        for( reg=0; reg<32; reg++ )
        {
            if( !VALID(SaveReg[reg].Offset) ) continue;

            if( VALID(SaveReg[reg].Prev) )
                AedgLinkFancy(AedgNewFancy(factorbbl[reg], factorbbl[SaveReg[reg].Prev], ET_NORM));
            else
                AedgLinkFancy( AedgNewFancy(factorbbl[reg], factorbbl[reg], ET_EXIT_LINK) );
        }
    
        scount += benefit[the_reg];

        FOREACH_LIVE_FUN(ifun)
        {
            ibbl = AFUN_BBL_FIRST(ifun);
            if( ABBL_REGS_SAVED(ibbl) == REGS_NONE )  continue;
            if( !(REG2BIT(the_reg) & ABBL_REGS_DEAD_AT_INIT(ibbl)) ) continue;

            pcount++;

            if( verbose )
            {
                fprintf(fp,"\t[*] fun %d [%s] saved %016lx (%d) -> dead %d  (%ld)\n",
                        ifun,AFUN_NAME(ifun),
                        ABBL_REGS_SAVED(ibbl),
                        RegsHighestSavedRegByCallingConvention(ABBL_REGS_SAVED(ibbl)),
                        the_reg,
                        ABBL_COUNT(ibbl));
            }

#if 0
            PrintFunSuper(stdout,ifun);
#endif
            /* now do the outlining for ifun */

            /* find the right target function init bbl */
            
            reg = RegsHighestSavedRegByCallingConvention(ABBL_REGS_SAVED(ibbl));
            
            /* kill stores, split bbl */

            AbblKillSavedRegsCallingConvention(ibbl,ABBL_REGS_SAVED(ibbl));
            icount+=BinaryBits(ABBL_REGS_SAVED(ibbl));
            
            jbbl = AbblNew();
            ABBL_FUN(jbbl)  = ABBL_FUN(ibbl);
            ABBL_TYPE(jbbl) = ABBL_TYPE(ibbl);
            ABBL_COUNT(jbbl) = ABBL_COUNT(ibbl);
            AbblInsertAfter(jbbl,ibbl);

            AbblMoveSuccEdges(ibbl, jbbl);
                  
            FOREACH_BBL_INS(ibbl, iins)
            {
                if( AINS_CODE(iins)==I_lda && AINS_REG_B(iins)==REG_SP && AINS_REG_C(iins)==REG_SP )  continue;
                AinsMove(iins, jbbl);
            }

            iins = AinsNew();
            AinsMakeBsr(iins, the_reg);
            AINS_BBL(iins) = ibbl;
            AinsAppend(iins);
            icount--;
            
            ABBL_TYPE(ibbl) = BT_CALL;

            AedgLinkFancy(AedgNewFancy(ibbl, jbbl, ET_FUN_LINK));
            AedgLinkFancy( AedgNewFancy(ibbl, factorbbl[reg], ET_FUN_CALL) );

            ABBL_REGS_SAVED(ibbl) = REGS_NONE;   /* exclude this bbl from further processing */

#if 0
            PrintFunSuper(stdout,ifun);
#endif
        }        
    }
        
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    InfoStatus(fp,"\t[FactorProlog]: removed prologs %d  removed saves %d  added funs %d  nins %d {%ldms}\n",
               pcount, scount, acount,icount,time);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FactorEpilogs"

extern void FactorEpilogs(FILE *fp)
{
    INDEX ifun;
    INDEX ibbl;
    INDEX iins;
    INDEX factorbbl[32];
    NUMBER rcount = 0;
    NUMBER ecount = 0;
    NUMBER icount = 0;

    BOOL verbose = HasOption(PRINT_OPT_VERBOSE,"Factorepilogs");
    
    REG reg;

    clock_t time = clock();

    verbose  = 1;
    
    /* create factor functions */

    for( reg = 0; reg<32; reg++ )
    {
        factorbbl[reg ] = INVALID;
        
        if( !VALID(SaveReg[reg].Offset) ) continue;

        ifun = AfunNewFancy();
        AFUN_NAME(ifun) = "$FactorRestore";
        
        ibbl = AbblNew();
        ABBL_FUN(ibbl) = ifun;
        if( VALID(SaveReg[reg].Prev) )
            ABBL_TYPE(ibbl) = BT_NORM;
        else
            ABBL_TYPE(ibbl) = BT_RET;
        ABBL_COUNT(ibbl) = 1;                          /* NEEDS TO BE FIXED !!! */
        AbblAppendFancy(ibbl);
        
        if( !VALID(SaveReg[reg].Prev) )
        {
            iins = AinsNew();
            AinsMakeAddq(iins, REG_SP, REG_RA, REG_SP);  /* $sp := $26 + $sp */
            AINS_BBL(iins) = ibbl;
            AinsAppend(iins);
            icount -= 1;        
            
            iins = AinsNew();
            AinsMakeSubq(iins, REG_SP, REG_RA, REG_RA);  /* $26 := $sp - $26 */
            AINS_BBL(iins) = ibbl;
            AinsAppend(iins);
            icount -= 1;        
                    
            iins = AinsNew();
            AinsMakeLdq(iins, REG_RA, REG_RA, SaveReg[reg].Offset);        /* $26 := 0($26) */
            AINS_BBL(iins) = ibbl;
            AinsAppend(iins);
            icount -= 1;        
            
            iins = AinsNew();
            AinsMakeRet(iins, REG_RA);                   /* ret ($26) */
            AINS_BBL(iins) = ibbl;
            AinsAppend(iins);
            icount -= 1;        
        }
        else
        {
          iins = AinsNew();
          AinsMakeLdq(iins, REG_SP, reg, SaveReg[reg].Offset);
          AINS_BBL(iins) = ibbl;
          AinsAppend(iins);
          icount -= 1;        
      }
        
        factorbbl[reg] = ibbl;
  }
    
  /* create corresponding edges */
        
    for( reg=0; reg<32; reg++ )
    {
        if( !VALID(SaveReg[reg].Offset) ) continue;
        
        if( VALID(SaveReg[reg].Prev) )
            AedgLinkFancy(AedgNewFancy(factorbbl[reg], factorbbl[SaveReg[reg].Prev], ET_NORM));
        else
            AedgLinkFancy( AedgNewFancy(factorbbl[reg], factorbbl[reg], ET_EXIT_LINK) );
    }

    /* factor "good" return bbls */
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_RET ) continue;
        iins = ABBL_INS_LAST(ibbl);
        CASSERT( AINS_CODE(iins) == I_ret,("weird\n"));
        if( AINS_REG_B(iins) != REG_RA ) continue;
        ABBL_REGS_RESTORED(ibbl) = AbblRestoredRegsCallingConvention(ibbl);
#if 0
        PrintBblSuper(stdout,ibbl);
        printf("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ %016lx\n",ABBL_REGS_RESTORED(ibbl));
#endif   
        if( BinaryBits(ABBL_REGS_RESTORED(ibbl)) == 0 ) continue;
        
        ecount++;
        rcount += BinaryBits( ABBL_REGS_RESTORED(ibbl));

        reg = RegsHighestSavedRegByCallingConvention( ABBL_REGS_RESTORED(ibbl));

        if( verbose )
        {
            ifun = ABBL_FUN(ibbl);
            fprintf(fp,"\t[*] bbl %d  fun %d [%s] restored %016lx (%d)  (%ld)\n",
                    ibbl,
                    ifun,AFUN_NAME(ifun),
                    ABBL_REGS_RESTORED(ibbl),
                    RegsHighestSavedRegByCallingConvention(ABBL_REGS_RESTORED(ibbl)),
                    ABBL_COUNT(ibbl));
        }

        AinsKill(iins);
        icount+=1;

        CASSERT( AbblNsuccs(ibbl) == 1,("too manu outgoing edges in BT_RET\n"));
        AedgKillFancy( ABBL_SUCC(ibbl) );

        AbblKillRestoredRegsCallingConvention(ibbl,ABBL_REGS_RESTORED(ibbl));
        icount += BinaryBits(ABBL_REGS_RESTORED(ibbl));
                
        FOREACH_BBL_INS_R(ibbl, iins)
        {
            if( AINS_CODE(iins) == I_lda && AINS_REG_B(iins) == REG_SP && AINS_REG_C(iins) == REG_SP)
            {
                AinsMakeLda(iins, REG_ZERO, REG_RA, AINS_DATA(iins));
                break;
            }
        }

        ABBL_TYPE(ibbl) = BT_NORM;
        AedgLinkFancy(AedgNewFancy(ibbl,factorbbl[reg], ET_NORM));
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    InfoStatus(fp,"\t[FactorEpilog]: removed epilogs %d  removed restores: %d  nins %d {%ldms}\n", ecount, rcount, icount, time);
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "NormEpilogs"

extern void NormEpilogs(FILE *fp)    
{
    INDEX ibbl;
    INDEX iins;
    INDEX iins_def;

    NUMBER ecount = 0;

    BOOL verbose;
    
    clock_t time = clock();

    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_RET ) continue;
        iins = ABBL_INS_LAST(ibbl);
        CASSERT( AINS_CODE(iins) == I_ret,("weird\n"));
        if( AINS_REG_B(iins) == REG_RA ) continue;
        if( ABBL_OUT(ibbl) & REG2BIT(AINS_REG_B(iins)) ) continue;
        if( ABBL_OUT(ibbl) & REG2BIT(REG_RA) ) continue;

        for(iins_def = AINS_PREV(iins); VALID(iins_def); iins_def = AINS_PREV(iins_def) )
        {
            if( AINS_REG_C(iins_def) == AINS_REG_B(iins) )
                break;
        }

        if( !VALID(iins_def) ) continue;
        ecount++;
#if 0        
        PrintBblSuper(stdout,ibbl);
#endif
        
        RenameRegister(iins_def,REG_RA);

#if 0        
        PrintBblSuper(stdout,ibbl);
#endif
    }
    
    verbose  = 1;
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    InfoStatus(fp,"\t[NormEpilog]: normalized epilogs %d   {%ldms}\n", ecount, time);
}

/*=======================================================
=======================================================*/

