#ifndef GLOBAL_H
#define GLOBAL_H


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "error.h"

/*=======================================================
=======================================================*/
#define CONFIG_NSEC          50
#define CONFIG_NFUN          (1024*16)
#define CONFIG_NBBL          (1024*256)
#define CONFIG_NEDG          (1024*512)
#define CONFIG_NINS          (1024*768) 
#define CONFIG_NSTK          (1024*192)
#define CONFIG_NTAB          (1024*30)   /* the last two combined should be a multiple of 2^14 */
#define CONFIG_NCON          (1024*2/2)

#define CALL_PROF_SIZE       (42*8)
#define COND_PROF_SIZE       (42*8)

#define CONSTRAINTS_ALLOCATED  (1024*1024)
#define SCHEDULEES_ALLOCATED   (1024*16)
#define SCHLOTS_ALLOCATED      CONFIG_NINS
#define CANDIDATES_ALLOCATED   (256)

/*=======================================================
=======================================================*/
#define CHANGE(a) (a ? '+' : '-')
/*=======================================================
=======================================================*/
typedef unsigned char  UINT8;
typedef   signed char  INT8;
typedef unsigned short UINT16;
typedef   signed short INT16;
typedef unsigned int   UINT32;
typedef   signed int   INT32;
typedef unsigned long  UINT64;
typedef   signed long  INT64;

typedef INT8  BOOL;
typedef INT32 INDEX;
typedef INT32 NUMBER;
typedef INT64 ADDRESS;
typedef char *STRING;
typedef UINT16 FLAGS;
typedef INT8  DATATYPE;
typedef INT8  EDGETYPE;
typedef UINT64 RESLIST;
typedef INT8  RES;
typedef INT32    LATENCY;
typedef UINT8   PIPELINES;
typedef UINT64 REGLIST;
#define REGSET REGLIST
typedef INT8  REG;

#define MAX_CLASS 64
#define ALTO_MAGIC 0xcaffe

/*=======================================================
=======================================================*/

typedef union
{
    INT64  quadword;

    struct
    {
        INT32 d1;
        INT32 d2;
    }doubleword;
    
    struct
    {
        INT16 w1;
        INT16 w2;
        INT16 w3;
        INT16 w4;
    }word;
    
    
    struct
    {
        INT8 b1;
        INT8 b2;
        INT8 b3;
        INT8 b4;
        INT8 b5;
        INT8 b6;
        INT8 b7;
        INT8 b8;
        
    }byte;
    
    struct
    {
        UINT8 b1;
        UINT8 b2;
        UINT8 b3;
        UINT8 b4;
        UINT8 b5;
        UINT8 b6;
        UINT8 b7;
        UINT8 b8;
            
        }ubyte;
}QUADWORD;
    

typedef union  
{
    INT64   Eval;
    double  FEval;
    INDEX   Def;
}INS_EVAL;

/*=======================================================
=======================================================*/

typedef struct
{
    INT32 allocated;
    INT32 used;
    INT32 size;
}ARRAY;

/*=======================================================
=======================================================*/
typedef struct
{
    INDEX Target;
}STRUCT_TAB;

typedef struct
{
    NUMBER    Offset;
    FLAGS     Flags;
    DATATYPE  ReadType;
    DATATYPE  WriteType;
    REG       ReadReg;
    REG       WriteReg;
    UINT8     ReadNum;
    UINT8     WriteNum;
}
STRUCT_STK;


typedef struct
{
    QUADWORD u;
    INT64 ArgPotential[64];
    FLAGS ArgFlags[64];
    ADDRESS   Gp;
    STRING    Name;
    REGLIST   RegsUsed;   /* may use */
    REGLIST   RegsThrough; /* may pass through */
    REGLIST   RegsSaved;  /* must save */
    REGLIST   RegsChanged; /* may change */
    INDEX     Next;
    INDEX     Prev;
    INDEX     Stk;
    INT32     StkMin;
    INT32     StkMax;
    INDEX     BblFirst;
    INDEX     BblLast;
    INDEX     Master;
    INT32     NumInss;
    /*
    INDEX     Address;*/  /* original virtual start address */
    NUMBER    Mod;        /* summary of the function's side effects */
    FLAGS     Flags;
    FLAGS     AuxFlags;
    REG       RetReg;
    REG       CallReg;
    REG       FpReg;
    BOOL      Recursive;
}STRUCT_FUN;

/*=======================================================
=======================================================*/




typedef struct
{
    FLAGS   Flags;
    FLAGS   AuxFlags;
    UINT8   Type;
    UINT8   Align;

    INT64   Count;
    REGLIST Out;

    INDEX   InsFirst;
    INDEX   InsLast;
    
    INDEX   Next;
    INDEX   Prev;

    INDEX   Pred;
    INDEX   Succ;
    
    INDEX   EdgSucc1;
    INDEX   EdgSucc2;

    INDEX   JumpTab;
    INDEX   Fun;      /* function the basic block is a member of */
    
    INDEX   PreDom;
    INDEX   PostDom;
    INDEX   LoopHeader;
    INDEX   LoopNextMemb;
    INDEX   LoopFirstMemb;

}STRUCT_BBL;

/*=======================================================
=======================================================*/

typedef struct
{
    INDEX   Head;
    INDEX   Tail;
    INDEX   Pred;
    INDEX   Succ;
    INT64   Count;
    FLAGS   Flags;
    EDGETYPE Type;
}STRUCT_EDG;

/*=======================================================
=======================================================*/

typedef struct
{
    INDEX   LoadDef;
    NUMBER  LoadOff;
    INDEX   Prev;
    INDEX   Next;
    INT32   Data;       /* must be signed */
    INDEX   Bblock;
    UINT16  Code;
    FLAGS   Flags;
    REG     RegA;
    REG     RegB;
    REG     RegC;
}STRUCT_INS;

/*=======================================================
=======================================================*/

typedef struct
{
    INDEX Pred;
    INDEX Succ;
    INDEX Head;
    INDEX Tail;
    UINT8 Type;
    UINT8 Regres;
    UINT16 Flags;
    INT16 LatencyAv;
    INT16 LatencyHi;
}CONSTRAINT;

/*=======================================================
=======================================================*/

typedef struct
{
    INDEX   Schedulee;
    LATENCY EarliestIssueCycle;
    NUMBER  BonusMalusHigh;
    LATENCY Latency;
    LATENCY TimeToEnd;
    NUMBER  BonusMalusLow;
}CANDIDATE;

/*=======================================================
=======================================================*/

typedef struct
{
    INDEX   Instr;
    LATENCY IssueCycle;
}SCHLOT;

/*=======================================================
=======================================================*/

typedef struct
{
    INDEX Pred;
    INDEX Succ;
    INDEX Instr;
    INDEX Pos;
    REGLIST LiveRegs;
    LATENCY TimeToEnd;
    LATENCY IssueCycle;
    UINT16 Flags;
    INT8  Type;
}SCHEDULEE;

/*=======================================================
=======================================================*/

typedef struct
{
    STRING    InlineMethod;
    STRING    LivenessMethod;
    
    STRING    PrintOptWhat;
    STRING    PrintOptWhen;
    STRING    PrintOptVerbose;
    INT64     PrintOptNum;
    STRING    PrintOptFun;

    STRING    DisabledOpt;
    
    NUMBER    FunAlign;
    NUMBER    LoopAlign;
    NUMBER    BblAlign;

    BOOL      Compress;
    BOOL      HaveBblProfile;
    STRING    ProfileWhat;
    NUMBER    InlineExpansion;
    NUMBER    Rounds;
    NUMBER    IcacheSize;
    INT64     Count;
    INT64     CallCount;
    

    STRING    InFile;
    STRING    OutFile;

    NUMBER    uSec;        
    NUMBER    uFun;        
    NUMBER    uBbl;        
    NUMBER    uEdg;
    NUMBER    uIns;
    NUMBER    uStk;
    NUMBER    uTab;
    NUMBER    uCon;
    NUMBER    uConstraints; 
    NUMBER    uCandidates; 
    NUMBER    uSchedulees; 
    NUMBER    uSchlots; 
    NUMBER    Nvalues;

    ADDRESS   SwitchSectionStart;
    ADDRESS   ConstSectionStart;
}GLOBAL;

extern GLOBAL Global;
extern STRUCT_FUN   aFun[];
extern STRUCT_BBL   aBbl[];
extern STRUCT_EDG   aEdg[];
extern STRUCT_INS   aIns[];
extern STRUCT_STK   aStk[];
extern STRUCT_TAB   aTab[];
extern SCHEDULEE    aSchedulee[];
extern SCHLOT       aSchlot[];
extern CONSTRAINT   aConstraint[];
extern CANDIDATE    aCandidate[];

extern QUADWORD     uBbl1[];
extern QUADWORD     uBbl2[];
extern QUADWORD     uBbl3[];

extern QUADWORD     uIns1[];

extern QUADWORD     uEdg1[];
extern QUADWORD     uEdg2[];

extern INS_EVAL     eInsA[];
extern INS_EVAL     eInsB[];
extern INS_EVAL     eInsC[];


/*=======================================================
=======================================================*/

typedef struct
{
    UINT16 OpCode;
    UINT16 FunCode;
    UINT16 Code;
    INT8   Name[16];
    INT8   Type;
    REGLIST ResUse;
    RESLIST ResDef;
    UINT8   Pipelines;
    INT16 LatencyAv;
    INT16 LatencyHi;
}OPCODE_TAB_ENTRY;

/*=======================================================
=======================================================*/
#define INLINE_EXPANSION   Global.InlineExpansion
#define INLINE_METHOD      Global.InlineMethod
#define LIVENESS_METHOD      Global.LivenessMethod


#define PRINT_OPT_WHAT     Global.PrintOptWhat
#define PRINT_OPT_WHEN     Global.PrintOptWhen
#define PRINT_OPT_VERBOSE  Global.PrintOptVerbose
#define PRINT_OPT_NUM      Global.PrintOptNum
#define PRINT_OPT_FUN      Global.PrintOptFun

#define DISABLED_OPT      Global.DisabledOpt

#define CONFIG_BBL_ALIGN Global.BblAlign
#define CONFIG_LOOP_ALIGN Global.LoopAlign
#define CONFIG_FUN_ALIGN Global.FunAlign

#define CONFIG_ROUNDS        	 Global.Rounds
#define CONFIG_ICACHE_SIZE       Global.IcacheSize
#define CONFIG_HAVE_BBL_PROFILE  Global.HaveBblProfile
#define CONFIG_MAKE_PROFILE      Global.MakeProfile
#define CONFIG_PROFILE_WHAT      Global.ProfileWhat
#define CONFIG_COUNT         	 Global.Count
#define CONFIG_CALLCOUNT         Global.CallCount
#define CONFIG_INFILE        	 Global.InFile
#define CONFIG_OUTFILE       	 Global.OutFile
#define CONFIG_NVALUES           Global.Nvalues
#define CONFIG_COMPRESS           Global.Compress

#define CONFIG_USEC          Global.uSec
#define CONFIG_UFUN          Global.uFun
#define CONFIG_UBBL          Global.uBbl
#define CONFIG_UEDG          Global.uEdg
#define CONFIG_UINS          Global.uIns
#define CONFIG_USTK          Global.uStk
#define CONFIG_UTAB          Global.uTab
#define CONFIG_UCON          Global.uCon
#define    CONSTRAINTS_USED  Global.uConstraints
#define    SCHEDULEES_USED   Global.uSchedulees
#define    SCHLOTS_USED      Global.uSchlots
#define    CANDIDATES_USED   Global.uCandidates

/*=======================================================
=======================================================*/

#define INVALID (-1)
#define VALID(i)  ((i)!=-1)

#define CONFIG_NINDIR       20

typedef struct {
    ADDRESS addr;
    ADDRESS count;
} indir;

typedef struct {
    INT64 iins;
    INT64 MagicNum;
    INDEX ins;
    REG reg1;
    REG reg2;
    INT64 totalcount;
    indir values[CONFIG_NINDIR];
    double benefit;
} Indir;


extern Indir *aIndir;
extern NUMBER aIndirFuns;

extern Indir *aValues;

typedef struct {
    ADDRESS counterStart;
    ADDRESS currentStart;
    NUMBER bblsProf;
    NUMBER edgsProf;
    NUMBER callsProf;
    NUMBER jmpsProf;
    NUMBER valsProf;
    INDEX valprofBbl;
} PROFILING;

extern PROFILING Profile;

/*=======================================================
=======================================================*/

#define FOREACH_VPF(i)            for(i=0; i < CONFIG_NVALUES; i++)
#define AVPF(i)                   aValues[i]
#define AVPF_MAGIC(i)             aValues[i].MagicNum
#define AVPF_COUNT(i)             aValues[i].totalcount
#define AVPF_FIRST_VALUE(i)       aValues[i].values[0].addr /* most frequent value */
#define AVPF_FIRST_COUNT(i)       aValues[i].values[0].count
#define AVPF_REG1(i)              aValues[i].reg1
#define AVPF_REG2(i)              aValues[i].reg2
#define AVPF_BENEFIT(i)           aValues[i].benefit


/*=======================================================
=======================================================*/

#define ATAB(i)             aTab[i]
#define ATAB_TARGET(i)      ATAB(i).Target

#define ASTK(i)             aStk[i]

#define ASTK_OFFSET(i)      ASTK(i).Offset
#define ASTK_FLAGS(i)       ASTK(i).Flags
#define ASTK_READ_TYPE(i)   ASTK(i).ReadType
#define ASTK_WRITE_TYPE(i)  ASTK(i).WriteType
#define ASTK_READ_REG(i)    ASTK(i).ReadReg
#define ASTK_WRITE_REG(i)   ASTK(i).WriteReg
#define ASTK_READ_NUM(i)    ASTK(i).ReadNum
#define ASTK_WRITE_NUM(i)   ASTK(i).WriteNum

#define FOREACH_FUN_STK(if,is) for(is=AFUN_STK(if); is<AFUN_STK_SIZE(if)+AFUN_STK(if); is++)

/*=======================================================
=======================================================*/
#define AINS(i)              aIns[i]
#define AINS_LOAD_DEF(i)     AINS(i).LoadDef
#define AINS_LOAD_OFF(i)     AINS(i).LoadOff
#define AINS_PREV(i)         AINS(i).Prev
#define AINS_NEXT(i)         AINS(i).Next
#define AINS_CODE(i)         AINS(i).Code
#define AINS_TYPE(i)         InstructionTable[AINS_CODE(i)].Type
#define AINS_NAME(i)         InstructionTable[AINS_CODE(i)].Name
#define AINS_RES_USE(i)      InstructionTable[AINS_CODE(i)].ResUse
#define AINS_RES_DEF(i)      InstructionTable[AINS_CODE(i)].ResDef
#define AINS_LATENCY_AV(i)         InstructionTable[AINS_CODE(i)].LatencyAv
#define AINS_LATENCY_HI(i)         InstructionTable[AINS_CODE(i)].LatencyHi
#define AINS_PIPELINES(i)       InstructionTable[AINS_CODE(i)].Pipelines

#define AINS_FLAGS(i)        AINS(i).Flags
#define AINS_DATA(i)         AINS(i).Data
#define AINS_REG_A(i)        AINS(i).RegA
#define AINS_REG_B(i)        AINS(i).RegB
#define AINS_REG_C(i)        AINS(i).RegC
#define AINS_BBL(i)          AINS(i).Bblock

#define AINS_C_EVAL(i)       eInsC[i].Eval
#define AINS_C_FEVAL(i)      eInsC[i].FEval
#define AINS_C_DEF(i)        eInsC[i].Def

#define AINS_B_EVAL(i)       eInsB[i].Eval
#define AINS_B_FEVAL(i)      eInsB[i].FEval
#define AINS_B_DEF(i)        eInsB[i].Def

#define AINS_A_EVAL(i)       eInsA[i].Eval
#define AINS_A_FEVAL(i)      eInsA[i].FEval
#define AINS_A_DEF(i)        eInsA[i].Def

#define AINS_IS_KILLED(i)    (AINS_FLAGS(i) & F_KILLED)

#define AINS_IS_RELOC(i)     (AINS_FLAGS(i) & F_RELOC)

#define AINS_U1(i) uIns1[i]
/*=======================================================
=======================================================*/

#define ABBL(ib)             aBbl[ib]

#define ABBL_FLAGS(i)        ABBL(i).Flags
#define ABBL_AUXFLAGS(i)     ABBL(i).AuxFlags
/*
#define ABBL_OLD_ADDR(i)     ABBL(i).OldAddr
*/
#define ABBL_SUCC(i)         ABBL(i).Succ
#define ABBL_PRED(i)         ABBL(i).Pred
#define ABBL_NEXT(i)         ABBL(i).Next
#define ABBL_PREV(i)         ABBL(i).Prev
#define ABBL_FUN(i)          ABBL(i).Fun
#define ABBL_INS_FIRST(i)    ABBL(i).InsFirst
#define ABBL_INS_LAST(i)     ABBL(i).InsLast
#define ABBL_PRE_DOM(i)      ABBL(i).PreDom
#define ABBL_POST_DOM(i)     ABBL(i).PostDom
#define ABBL_FALL_THRU(i)    ABBL(i).FallThru
#define ABBL_BR_TARGET(i)    ABBL(i).BrTarget
#define ABBL_COUNT(i)        ABBL(i).Count
#define ABBL_OUT(i)          ABBL(i).Out
#define ABBL_AVAIL(i)        ABBL(i).Avail
#define ABBL_ALIGN(i)        ABBL(i).Align
#define ABBL_TYPE(i)         ABBL(i).Type
#define ABBL_EDG_TRUE(i)     ABBL(i).EdgSucc1
#define ABBL_EDG_FALSE(i)    ABBL(i).EdgSucc2
#define ABBL_EDG_CALL(i)     ABBL(i).EdgSucc1
#define ABBL_EDG_LINK(i)     ABBL(i).EdgSucc2
#define ABBL_EDG_NORM(i)     ABBL(i).EdgSucc2
#define ABBL_EDG_UNCOND(i)   ABBL(i).EdgSucc1
#define ABBL_EDG_JUMP(i)     ABBL(i).EdgSucc1
#define ABBL_EDG_EXTRA(i)    ABBL(i).EdgSucc2
#define ABBL_JUMP_TAB(i)     ABBL(i).JumpTab

#define ABBL_LOOP_HEADER(i)       ABBL(i).LoopHeader
#define ABBL_LOOP_NEXT_MEMBER(i)  ABBL(i).LoopNextMemb
#define ABBL_LOOP_FIRST_MEMBER(i) ABBL(i).LoopFirstMemb

#define ABBL_IS_LOOP_HEADER(i)    ( ABBL_FLAGS(i) & FB_LOOP_HEADER )
#define ABBL_IS_IN_ANY_LOOP(bbl) \
           (VALID(ABBL_LOOP_HEADER(bbl)) || ABBL_IS_LOOP_HEADER(bbl))
#define ABBL_FOREACH_LOOP_MEMBER(loop_header,memb)  \
     for( memb=ABBL_LOOP_FIRST_MEMBER(loop_header); \
	  VALID(memb);                              \
	  memb=ABBL_LOOP_NEXT_MEMBER(memb) )

     
#define ABBL_U1(i) uBbl1[i]
#define ABBL_U2(i) uBbl2[i]
#define ABBL_U3(i) uBbl3[i]

#define FOREACH_FUN_TOPLOOP(f,loop) \
   FOREACH_FUN_BBL(f, loop)          \
     if( (! ABBL_IS_LOOP_HEADER(loop) || VALID(ABBL_LOOP_HEADER(loop))) ) \
        continue; else

#define FOREACH_FUN_TOPBBL(f,loop) \
   FOREACH_FUN_BBL(f, loop)          \
     if( ABBL_IS_IN_ANY_LOOP(loop) ) continue; else


#define ABBL_MARK(ibbl) ABBL_FLAGS(ibbl) |= F_MARKED
#define ABBL_UNMARK(ibbl) ABBL_FLAGS(ibbl) &= ~F_MARKED
     
#define ABBL_IS_KILLED(i)    (ABBL_FLAGS(i) & F_KILLED)
#define ABBL_IS_RELOC(i)     (ABBL_FLAGS(i) & F_RELOC)
#define ABBL_HAS_PREDS(i)    VALID( ABBL_PRED(i))
#define ABBL_HAS_SUCCS(i)    VALID( ABBL_SUCC(i))
#define ABBL_HAS_INSS(i)     VALID( ABBL_INS_FIRST(i))
#define ABBL_IS_FIRST(i)     (i == AFUN_BBL_FIRST(ABBL_FUN(i)))
#define ABBL_IS_LAST(i)      (i == AFUN_BBL_LAST(ABBL_FUN(i)))
#define ABBL_IS_BUSY(i)      (ABBL_FLAGS(i) & FB_BUSY)
#define ABBL_IS_CRITICAL(i)  (ABBL_FLAGS(i) & FB_CRITICAL)
#define ABBL_IS_CRIT_PATH(i) (ABBL_FLAGS(i) & FB_CRIT_PATH)
#define ABBL_IS_CRIT_ESC(i)  (ABBL_FLAGS(i) & FB_CRIT_ESC)
     
#define ABBL_IS_MARKED(i)    (ABBL_FLAGS(i) & F_MARKED)
#define ABBL_IS_ALIGNED(i)   (ABBL_FLAGS(i) & FB_ALIGNED)
#define ABBL_IS_HOT(i)       (ABBL_COUNT(i) >= CONFIG_COUNT)
#define ABBL_IS_ZERO(i)      (ABBL_COUNT(i) <= 0)

     /* #define ABBL_CALL_IS_HOT(i)  (ABBL_COUNT(i) >= CONFIG_CALLCOUNT) */



#define FOREACH_BBL_TAB(ib,it) for(it=ABBL_JUMP_TAB(ib); VALID(ATAB_TARGET(it)); it++)
/*=======================================================
=======================================================*/
#define AEDG(ie)             aEdg[ie]
     
#define AEDG_FLAGS(i)        AEDG(i).Flags
#define AEDG_HEAD(i)         AEDG(i).Head
#define AEDG_TAIL(i)         AEDG(i).Tail
#define AEDG_FROM(i)         AEDG(i).Head
#define AEDG_TO(i)           AEDG(i).Tail
#define AEDG_SUCC(i)         AEDG(i).Succ
#define AEDG_PRED(i)         AEDG(i).Pred
#define AEDG_TYPE(i)         AEDG(i).Type
#define AEDG_COUNT(i)        AEDG(i).Count
#define AEDG_IS_KILLED(i)    (AEDG_FLAGS(i) & F_KILLED)
#define AEDG_IS_INTER(i)     (AEDG_TYPE(i) == ET_FUN_CALL || \
                              AEDG_TYPE(i) == ET_FUN_RETURN || \
                              ABBL_FUN(AEDG_HEAD(i)) != ABBL_FUN(AEDG_TAIL(i)))
#define AEDG_IS_FALL_THRU(i) (AEDG_TYPE(i)==ET_NORM  || AEDG_TYPE(i)==ET_EXTRA ||\
                              AEDG_TYPE(i)==ET_FALSE || AEDG_TYPE(i)==ET_FUN_LINK)
#define AEDG_IS_RECURSIVE(i) (AEDG_FLAGS(i) & FE_RECURSIVE)


#define AEDG_U1(i) uEdg1[i]
#define AEDG_U2(i) uEdg2[i]
/*=======================================================
=======================================================*/
#define AFUN(if)             aFun[if]


#define AFUN_PSEUDO_HELL            0            
#define AFUN_PSEUDO_PAL             1
#define AFUN_PSEUDO_CALL            2
#define AFUN_PSEUDO_HELL_LITE       3
     
#define AFUN_PREV(i)         AFUN(i).Prev
#define AFUN_NEXT(i)         AFUN(i).Next
#define AFUN_FLAGS(i)        AFUN(i).Flags
#define AFUN_ADDRESS(i)      AFUN(i).Address
#define AFUN_NAME(i)         AFUN(i).Name
#define AFUN_BBL_FIRST(i)    AFUN(i).BblFirst
#define AFUN_BBL_LAST(i)     AFUN(i).BblLast
#define AFUN_STK_MAX(i)      AFUN(i).StkMax
#define AFUN_STK_MIN(i)      AFUN(i).StkMin
#define AFUN_GP(i)           AFUN(i).Gp
#define AFUN_RET_REG(i)      AFUN(i).RetReg
#define AFUN_NUM_INSS(i)     AFUN(i).NumInss
#define AFUN_CALL_REG(i)     AFUN(i).CallReg
#define AFUN_FP_REG(i)       AFUN(i).FpReg
#define AFUN_REGS_USED(i)    AFUN(i).RegsUsed
#define AFUN_REGS_THROUGH(i) AFUN(i).RegsThrough
#define AFUN_REGS_SAVED(i)   AFUN(i).RegsSaved
#define AFUN_REGS_CHANGED(i) AFUN(i).RegsChanged
#define AFUN_STK(i)          AFUN(i).Stk
#define AFUN_MOD(i)          AFUN(i).Mod
#define AFUN_MASTER(i)       AFUN(i).Master   /* set to the master function if this is just an entry fun */
#define AFUN_IS_RECURSIVE(i) AFUN(i).Recursive

#define AFUN_AUXFLAGS(i)     (AFUN(i).AuxFlags)
     
#define AFUN_IS_KILLED(i)    (AFUN_FLAGS(i) & F_KILLED)
#define AFUN_IS_PSEUDO(i)    (AFUN_FLAGS(i) & FF_PSEUDO)

#define AFUN_USE_DEF_ARG_POTENTIAL(f,r) AFUN(f).ArgPotential[r]
#define AFUN_ARG_FLAGS(f,r) AFUN(f).ArgFlags[r]
     
#define AFUN_HAS_BAD_STK(i)  (!VALID(AFUN_STK_SIZE(i)))
#define AFUN_IS_CRITICAL(i)  (AFUN_FLAGS(i) & FF_CRITICAL)
#define AFUN_HAS_STK_ARG(i) (AFUN_FLAGS(i) & FF_STK_ARG)
#define AFUN_HAS_STK_INVALID(i) (AFUN_FLAGS(i) & FF_STK_INVALID)
#define AFUN_HAS_STK_ALIAS(i) (AFUN_FLAGS(i) & FF_STK_ALIAS)
     
#define AFUN_IS_SETJMP(i)    (AFUN_FLAGS(i) & FF_SETJMP)
#define AFUN_IS_LONGJMP(i)    (AFUN_FLAGS(i) & FF_LONGJMP)
     
#define FOREACH_LIVE_FUN(i) for(i=0; i < CONFIG_UFUN; i++)\
     if( AFUN_IS_KILLED(i) ) continue; else
     
#define FOREACH_LIVE_FUN_R(i) for(i=CONFIG_UFUN-1; i>=0; i--)\
     if( AFUN_IS_KILLED(i) ) continue; else

/*=======================================================
=======================================================*/

#define FOREACH_INS(i) for(i=0; i < CONFIG_UINS; i++)
#define FOREACH_BBL(i) for(i=0; i < CONFIG_UBBL; i++)
#define FOREACH_FUN(i) for(i=0; i < CONFIG_UFUN; i++)
#define FOREACH_SEC(i) for(i=0; i < CONFIG_USEC; i++ )
#define FOREACH_EDG(i) for(i=0; i < CONFIG_UEDG; i++ )
#define FOREACH_TAB(i) for(i=0; i < CONFIG_UTAB; i++ )
#define FOREACH_CON(i) for(i=0; i < CONFIG_UCON; i++ )
     
#define FOREACH_EDG_R(i) for(i=CONFIG_UEDG-1; i>=0; i-- )

#define FOREACH_LIVE_INS(i) for(i=0; i < CONFIG_UINS; i++)\
     if( AINS_IS_KILLED(i) ) continue; else

#define FOREACH_LIVE_INS_R(i) for(i= CONFIG_UINS -1; i >=0; i--)\
     if( AINS_IS_KILLED(i) ) continue; else

#define FOREACH_LIVE_BBL(i) for(i=0; i < CONFIG_UBBL; i++)\
     if( ABBL_IS_KILLED(i) ) continue; else
     
#define FOREACH_LIVE_BBL_R(i) for(i=CONFIG_UBBL-1; i >=0; i--)\
     if( ABBL_IS_KILLED(i) ) continue; else
     
#define FOREACH_LIVE_EDG(i) for(i=0; i < CONFIG_UEDG; i++ )\
     if( AEDG_IS_KILLED(i) ) continue; else
     
#define FOREACH_LIVE_EDG_R(i) for(i=CONFIG_UEDG-1; i>=0; i-- )\
     if( AEDG_IS_KILLED(i) ) continue; else
     
#define FOREACH_PDR(i) for(i=0; i < BIN_SYMBOL_NPDT; i++ )
#define FOREACH_FDR(i) for(i=0; i < BIN_SYMBOL_NFDT; i++ ) 

#define FOREACH_SEC_REL(is,ir)    for(ir=0; ir < ASEC_NRELOCS(is); ir++ )
#define FOREACH_BBL_SUCC(ib, is)  for(is=ABBL_SUCC(ib); VALID(is); is=AEDG_SUCC(is))
#define FOREACH_BBL_PRED(ib, ip)  for(ip=ABBL_PRED(ib); VALID(ip); ip=AEDG_PRED(ip))
#define FOREACH_BBL_INS(ib, ii)   for(ii=ABBL_INS_FIRST(ib); VALID(ii); ii=AINS_NEXT(ii))
#define FOREACH_BBL_INS_R(ib, ii) for(ii=ABBL_INS_LAST(ib); VALID(ii); ii=AINS_PREV(ii))
#define FOREACH_FUN_BBL(if,ib)    for(ib=AFUN_BBL_FIRST(if); VALID(ib); ib=ABBL_NEXT(ib))
#define FOREACH_FUN_BBL_SAFE(if,ib,ic) \
     for(ib=AFUN_BBL_FIRST(if); VALID(ib) && (ic=ABBL_NEXT(ib),1); ib=ic)
#define FOREACH_FUN_BBL_R(if,ib)  for(ib=AFUN_BBL_LAST(if); VALID(ib); ib=ABBL_PREV(ib))
#define FOREACH_DOMT_CHILD(parent,child) \
               for( child = ABBL_DOMT_FIRST_CHILD(parent); \
		    VALID(child); \
		    child = ABBL_DOMT_NEXT_SIBL(child) )

/* An abstract datatype for sets of registers */

/*=======================================================
=======================================================*/

#define  PrIns(ins) PrintIns(stdout, ins)
#define  PrBblSuper(bbl) PrintBblSuper(stdout, bbl);

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 1
#endif

/*=======================================================
=======================================================*/


#define AinsUse(iins) \
     (((REG2BIT(AINS_REG_A(iins)) | \
     REG2BIT(AINS_REG_B(iins)) | \
     ((AINS_TYPE(iins) == IT_ICM || AINS_TYPE(iins) == IT_FCM) ? REG2BIT(AINS_REG_C(iins)) : REGS_NONE))) & REGS_ALL_BUT_ZERO) 

#define  AinsDef(iins)  (REG2BIT(AINS_REG_C(iins)) & REGS_ALL_BUT_ZERO)


/*=======================================================
=======================================================*/


     
#endif /* GLOBAL_H */
