/*=======================================================
  Author:   R. Muth
  Version:  $Id: peephole.c,v 1.35 1999/06/16 22:23:41 muth Exp $
  Comment:  This file contains the code for various peephole optimizations
=======================================================*/

#include "global.h"
#include "functions.h"
#include "opcode.h"
#include "regs.h"
#include "flags.h"
#include "fltconst.h"
#include <string.h>

/*=======================================================
=======================================================*/
#undef __PROC__ 
#define __PROC__ "PeepHoleULD"

static BOOL PeepHoleULD(INDEX iins)
{
    INT64 address;
    INDEX ibbl = AINS_BBL(iins);
    if( !(AINS_FLAGS(iins) & FI_EVAL_B) ) return 0;

    address = AINS_B_EVAL(iins) + AINS_DATA(iins);
    address &= ~7L;
    address -=  AINS_B_EVAL(iins);
    HURRAH(("LDQ_U -> LDQ  offset %04x -> %04x in bbl %d (%ld)\n",AINS_DATA(iins),address,ibbl,ABBL_COUNT(ibbl)));
    AINS_DATA(iins) = address;
    AINS_CODE(iins) = I_ldq;
    return 1;
}


/*=======================================================
=======================================================*/
#undef __PROC__ 
#define __PROC__ "PeepHoleUST"

static BOOL PeepHoleUST(INDEX iins)
{
    INT64 address;
    INDEX ibbl = AINS_BBL(iins);
    if( !(AINS_FLAGS(iins) & FI_EVAL_B) ) return 0;

    address = AINS_B_EVAL(iins) + AINS_DATA(iins);
    address &= ~7L;
    address -=  AINS_B_EVAL(iins);
    HURRAH(("STQ_U -> STQ  offset %04x -> %04x in bbl %d (%ld)\n",AINS_DATA(iins),address,ibbl,ABBL_COUNT(ibbl)));
    AINS_DATA(iins) = address;
    AINS_CODE(iins) = I_stq;
    return 1;
}

/*=======================================================
=======================================================*/
#undef __PROC__ 
#define __PROC__ "PeepHoleFMUL"

static BOOL PeepHoleFMUL(INDEX iins)
{
    INDEX ibbl = AINS_BBL(iins);
    REG   src;
    REG   dst;
    INT64 val;
    
    
    if( AINS_FLAGS(iins) & FI_EVAL_B )
    {
        src = AINS_REG_A(iins);
        val = AINS_B_EVAL(iins);
        dst = AINS_REG_C(iins);
    }
    else if( AINS_FLAGS(iins) & FI_EVAL_A )
    {
        src = AINS_REG_B(iins);
        val = AINS_A_EVAL(iins);
        dst = AINS_REG_C(iins);
    }
    else
    {
        return 0;
    }

    if( val == DOUBLE_NZERO || val == DOUBLE_PZERO )
    {
        HURRAH(("FMUL multiply with %016lx found in iins %d  ibbl %d (%ld) \n",val,iins,ibbl,ABBL_COUNT(ibbl)));
        /* We make a tiny mistake here by ignoring the sign of zero which under some circumstances
           might cause wrong results */
        AinsMakeCpys(iins,REG_FZERO,REG_FZERO,dst);
        return 1;
    }

    if( val == DOUBLE_PONE )
    {
        HURRAH(("FMUL multiply with %016lx found in iins %d  [%s] ibbl %d (%ld) \n",val,iins,AINS_NAME(iins),ibbl,ABBL_COUNT(ibbl)));
        AinsMakeCpys(iins,src,src,dst);
        return 1;
    }

    if(  val == DOUBLE_NONE )
    {
        HURRAH(("FMUL multiply with %016lx found in iins %d  [%s] ibbl %d (%ld) \n",val,iins,AINS_NAME(iins),ibbl,ABBL_COUNT(ibbl)));
        AinsMakeCpysn(iins,src,src,dst);
        return 1;
    }
    
    if( val == DOUBLE_PTWO )
    {
        HURRAH(("FMUL multiply with %016lx found in iins %d  [%s] ibbl %d (%ld) \n",val,iins,AINS_NAME(iins),ibbl,ABBL_COUNT(ibbl)));
        AinsMakeAddt(iins,src,src,dst);
        return 0;
    }
    
    return 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__ 
#define __PROC__ "PeepHoleFADD"

static BOOL PeepHoleFADD(INDEX iins)     
{
    if( AINS_FLAGS(iins) & FI_EVAL_A )
    {
        if( AINS_A_EVAL(iins) == DOUBLE_NZERO )
        {
            HURRAH(("FADD ins %d [%s] %016lx %016lx \n", iins, AINS_NAME(iins),AINS_A_EVAL(iins),  AINS_B_EVAL(iins)));
            AinsMakeCpys(iins,AINS_REG_B(iins),AINS_REG_B(iins),AINS_REG_C(iins));
        }
        else if( AINS_A_EVAL(iins) == DOUBLE_PZERO )
        {
            HURRAH(("FADD ins %d [%s] %016lx %016lx \n", iins, AINS_NAME(iins),AINS_A_EVAL(iins),  AINS_B_EVAL(iins)));
            AinsMakeCpys(iins,AINS_REG_B(iins),AINS_REG_B(iins),AINS_REG_C(iins));
        }
        else
        {
            return 0;
        }
        
    }
    else if( AINS_FLAGS(iins) & FI_EVAL_B )
    {
        if( AINS_B_EVAL(iins) == DOUBLE_NZERO )
        {
            HURRAH(("FADD ins %d [%s] %016lx %016lx \n", iins, AINS_NAME(iins),AINS_A_EVAL(iins),  AINS_B_EVAL(iins)));
            AinsMakeCpys(iins,AINS_REG_A(iins),AINS_REG_A(iins),AINS_REG_C(iins));
        }
        else if( AINS_B_EVAL(iins) == DOUBLE_PZERO )
        {
            HURRAH(("FADD ins %d [%s] %016lx %016lx \n", iins, AINS_NAME(iins),AINS_A_EVAL(iins),  AINS_B_EVAL(iins)));
            AinsMakeCpys(iins,AINS_REG_A(iins),AINS_REG_A(iins),AINS_REG_C(iins));
        }
        else
        {
            return 0;
        }
        
    }
    else
    {
        return 0;
    }

    return 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__ 
#define __PROC__ "PeepHoleFSUB"

static BOOL PeepHoleFSUB(INDEX iins)     
{
    if( AINS_REG_A(iins) == AINS_REG_B(iins)  )
    {
        POTENTIAL(("FSUB Peephole sub r r\n"));
    }
    else if( AINS_FLAGS(iins) & FI_EVAL_A )
    {
        if( AINS_A_EVAL(iins) == DOUBLE_NZERO )
        {
            HURRAH(("FSUB ins %d [%s]  %016lx %016lx\n", iins, AINS_NAME(iins),AINS_A_EVAL(iins),  AINS_B_EVAL(iins)));
            AinsMakeCpysn(iins,AINS_REG_B(iins),AINS_REG_B(iins),AINS_REG_C(iins));

        }
        else if( AINS_A_EVAL(iins) == DOUBLE_PZERO )
        {
            HURRAH(("FSUB ins %d [%s] %016lx %016lx \n", iins, AINS_NAME(iins),AINS_A_EVAL(iins),  AINS_B_EVAL(iins)));
            AinsMakeCpysn(iins,AINS_REG_B(iins),AINS_REG_B(iins),AINS_REG_C(iins));
        }
        else
        {
            return 0;
        }
        
    }
    else if( AINS_FLAGS(iins) & FI_EVAL_B )
    {
        if( AINS_B_EVAL(iins) == DOUBLE_NZERO )
        {
            HURRAH(("FSUB ins %d [%s] %016lx %016lx \n", iins, AINS_NAME(iins),AINS_A_EVAL(iins),  AINS_B_EVAL(iins)));
            AinsMakeCpys(iins,AINS_REG_A(iins),AINS_REG_A(iins),AINS_REG_C(iins));
        }
        else if( AINS_B_EVAL(iins) == DOUBLE_PZERO )
        {
            HURRAH(("FSUB ins %d [%s] %016lx %016lx \n", iins, AINS_NAME(iins),AINS_A_EVAL(iins),  AINS_B_EVAL(iins)));
            AinsMakeCpys(iins,AINS_REG_A(iins),AINS_REG_A(iins),AINS_REG_C(iins));
        }
        else
        {
            return 0;
        }
        
    }
    else
    {
        return 0;
    }

    return 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__ 
#define __PROC__ "PeepHoleFDIV"

static BOOL PeepHoleFDIV(INDEX iins)
{
    INDEX ibbl = AINS_BBL(iins);
    INDEX ifun = ABBL_FUN(ibbl);
    INT64 val;
    
    if( AINS_REG_A(iins) == AINS_REG_B(iins)  )
    {
        if( ABBL_IS_HOT(ibbl) )
            POTENTIAL(("FDIV Peephole div r r\n"));
    }    
    else if( AINS_FLAGS(iins) & FI_EVAL_A )
    {
        val = AINS_A_EVAL(iins);

        if( val == DOUBLE_PZERO  || val == DOUBLE_NZERO )
        {
            if( ABBL_IS_HOT(ibbl) )
                POTENTIAL(("FDIV with %016lx in fun %d [%s]  bbl %d  (%ld)  ins %d %8.3f\n",
                           val,ifun,AFUN_NAME(ifun),ibbl,ABBL_COUNT(ibbl),iins,*(double*)&val));
        }
        
        return 0;
    }
    else if( AINS_FLAGS(iins) & FI_EVAL_B )
    {
        val = AINS_B_EVAL(iins);

        if( ABBL_IS_HOT(ibbl) )
            POTENTIAL(("FDIV with const in fun %d [%s]  bbl %d  (%ld)  ins %d %8.3f\n",
                       ifun,AFUN_NAME(ifun),ibbl,ABBL_COUNT(ibbl),iins,*(double*)&val));
        return 0;
    }
    else
    {
        return 0;
    }

    return 0;
}

    
/*=======================================================
=======================================================*/
#undef __PROC__ 
#define __PROC__ "AinsMakeQuadMultiply"

static void AinsMakeQuadMultiply(INDEX iins,REG src, REG dst, INT64 val)
{
        if( val == 1 )
        {
            AinsMakeAddq(iins,src,REG_ZERO,dst);
        }
        else if( val == 2 )
        {
            AinsMakeAddq(iins,src,src,dst);
        }
        else if( val == 3 )
        {
            AinsMakeS4subq(iins,src,src,dst);
        }
        else if( val == 4 )
        {
            AinsMakeS4addq(iins,src,REG_ZERO,dst);
        }
        else if( val == 5 )
        {
            AinsMakeS4addq(iins,src,src,dst);
        }
        else if( val == 7 )
        {
            AinsMakeS8subq(iins,src,src,dst);
        }
        else if( val == 8 )
        {
            AinsMakeS8addq(iins,src,REG_ZERO,dst);
        }
        else if( val == 9 )
        {
            AinsMakeS8addq(iins,src,src,dst);
        }
        else if( (val & (val - 1)) == 0 ) /* only one bit is set */
        {
            AinsMakeSll(iins,src,REG_LIT,dst,BinaryLog(val));
        }
        else
        {
            return;
        }
}

/*=======================================================
=======================================================*/
#undef __PROC__ 
#define __PROC__ "AinsMakeQuadMultiply"

static void AinsMakeLongMultiply(INDEX iins,REG src, REG dst, INT64 val)
{
        if( val == 1 )
        {
            AinsMakeAddl(iins,src,REG_ZERO,dst);
        }
        else if( val == 2 )
        {
            AinsMakeAddl(iins,src,src,dst);
        }
        else if( val == 3 )
        {
            AinsMakeS4subl(iins,src,src,dst);
        }
        else if( val == 4 )
        {
            AinsMakeS4addl(iins,src,REG_ZERO,dst);
        }
        else if( val == 5 )
        {
            AinsMakeS4addl(iins,src,src,dst);
        }
        else if( val == 7 )
        {
            AinsMakeS8subl(iins,src,src,dst);
        }
        else if( val == 8 )
        {
            AinsMakeS8addl(iins,src,REG_ZERO,dst);
        }
        else if( val == 9 )
        {
            AinsMakeS8addl(iins,src,src,dst);
        }
        else
        {
            return;
        }
}


/*=======================================================
=======================================================*/
#undef __PROC__ 
#define __PROC__ "PeepHoleMUL"

static BOOL PeepHoleMUL(INDEX iins)
{
    INDEX ibbl = AINS_BBL(iins);
    INDEX ifun = ABBL_FUN(ibbl);
    REG   src;
    REG   dst;
    INT64 val;
    
    
    if( AINS_FLAGS(iins) & FI_EVAL_B )
    {
        src = AINS_REG_A(iins);
        val = AINS_B_EVAL(iins);
        dst = AINS_REG_C(iins);
    }
    else if( AINS_FLAGS(iins) & FI_EVAL_A )
    {
        src = AINS_REG_B(iins);
        val = AINS_A_EVAL(iins);
        dst = AINS_REG_C(iins);
    }
    else
    {
        return 0;
    }

    if( val == 0 )
    {
        HURRAH(("peephole evaluated %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
        AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
        return 1;
    }
    

    switch( AINS_CODE(iins) )
    {
    case I_mulq:
        AinsMakeQuadMultiply(iins,src,dst,val);
        if( AINS_CODE(iins) != I_mulq )
        {
            HURRAH(("MULQ (%ld) strength reduced to %s in bbl %d (%ld)\n",val,AINS_NAME(iins),ibbl,ABBL_COUNT(ibbl)));
            return 1;
        }
        else
        {
            if( BinaryBits(val) <= 4  || BinaryBits(val-1) <= 4  || BinaryBits(val+1) <= 4 )
                POTENTIAL(("MULQ (%d)  in bbl %d (%ld)\n",val,ibbl,ABBL_COUNT(ibbl)));
            return 0;
        }
        break;

    case I_mull:
        AinsMakeLongMultiply(iins,src,dst,val);
        if( AINS_CODE(iins) != I_mull )
        {
            HURRAH(("MULL (%ld) strength reduced to %s in bbl %d (%ld)\n",val,AINS_NAME(iins),ibbl,ABBL_COUNT(ibbl)));
            return 1;
        }
        else
        {
            if( BinaryBits(val) <= 4  || BinaryBits(val-1) <= 4  || BinaryBits(val+1) <= 4 )
                POTENTIAL(("MULL (%d)  in bbl %d (%ld)\n",val,ibbl,ABBL_COUNT(ibbl)));
            return 0;
        }
        break;
    }

    
    return 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PeepHoleIOP"

static BOOL PeepHoleICM(INDEX iins)
{
    INDEX jins;
    
    if( AINS_CODE(iins) != I_cmoveq ) return 0;
    if( AINS_REG_A(iins) != AINS_REG_B(iins) ) return 0;
    
    for( jins = AINS_PREV(iins); VALID(jins); jins = AINS_PREV(jins) )
    {
        if( AINS_REG_C(jins) == AINS_REG_C(iins) )
            break;
    }

    if( !VALID(jins) ) return 0;

    if( !(AINS_FLAGS(jins) & FI_EVAL_C) ) return 0;

    if( AINS_C_EVAL(jins) != 1 ) return 0;

    AinsMakeCmpult(iins,REG_ZERO,AINS_REG_A(iins),AINS_REG_C(iins),0);

#if 0
    HURRAH(("peephole changed lda/cmoveq combo\n")); 
    PrintBblSuper(stdout,ibbl);
#endif


    return 1;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PeepHoleIOP"

static BOOL PeepHoleIOP(INDEX iins)
{
    INDEX ibbl = AINS_BBL(iins);
    INDEX ifun = ABBL_FUN(ibbl);

    if( AINS_REG_A(iins) == AINS_REG_B(iins) )
    {
        if( AINS_CODE(iins) == I_subl || AINS_CODE(iins) == I_subq )
        {
            HURRAH(("peephole evaluated %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
            return 1;
        }
    }

    return 0;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PeepHoleLOG"

static BOOL PeepHoleLOG(INDEX iins)
{
    /*
    INDEX ibbl = AINS_BBL(iins);
    INDEX ifun = ABBL_FUN(ibbl);
    */

    if( AINS_REG_A(iins) == AINS_REG_B(iins) )
    {
        switch( AINS_CODE(iins) )
        {
        case I_and:
            if( AINS_REG_B(iins) == REG_LIT )
                AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), AINS_DATA(iins));
            else
                AinsMakeLda(iins,AINS_REG_B(iins), AINS_REG_C(iins), 0);
            return 1;
        case I_bis:
            if( AINS_REG_B(iins) == REG_LIT )
                AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), AINS_DATA(iins));
            else
                AinsMakeLda(iins,AINS_REG_B(iins), AINS_REG_C(iins), 0);
            return 1;
        case I_bic:
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
            return 1;            
        case I_eqv:
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), -1);
            return 1;
        case I_ornot:
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), -1);
            return 1;
        case I_xor:
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
            return 1;
        }
    }
    
    if( AINS_REG_A(iins) == REG_ZERO || ((AINS_FLAGS(iins) & FI_EVAL_A) && AINS_A_EVAL(iins) == 0) )
    {
        switch( AINS_CODE(iins) )
        {
        case I_and:
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
            return 1;
        case I_bis:
            if( AINS_REG_B(iins) == REG_LIT )
                AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), AINS_DATA(iins));
            else
                AinsMakeLda(iins,AINS_REG_B(iins), AINS_REG_C(iins), 0);
            return 1;
        case I_bic:
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
            return 1;            
        case I_eqv:
            AinsMakeNot(iins,AINS_REG_B(iins),AINS_REG_C(iins),AINS_DATA(iins));
            break;
        case I_ornot:
            /* this is the normal form for NOT !!! */
            break;
        case I_xor:
            if( AINS_REG_B(iins) == REG_LIT )
                AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), AINS_DATA(iins));
            else
                AinsMakeLda(iins,AINS_REG_B(iins), AINS_REG_C(iins), 0);
            return 1;
        }
    }

    if( AINS_REG_B(iins) == REG_ZERO || ((AINS_FLAGS(iins) & FI_EVAL_B) && AINS_B_EVAL(iins) == 0) )
    {
        switch( AINS_CODE(iins) )
        {
        case I_and:
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
            return 1;
        case I_bis:
            AinsMakeLda(iins,AINS_REG_A(iins), AINS_REG_C(iins), 0);
            return 1;
        case I_bic:
            AinsMakeLda(iins,AINS_REG_A(iins), AINS_REG_C(iins), 0);
            return 1;            
        case I_eqv:
            AinsMakeNot(iins,AINS_REG_A(iins),AINS_REG_C(iins),AINS_DATA(iins));
            return 1;
        case I_ornot:
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), -1);
            return 1;
        case I_xor:
            AinsMakeLda(iins,AINS_REG_A(iins), AINS_REG_C(iins), 0);
            return 1;
        }
    }


    if( (AINS_FLAGS(iins) & FI_EVAL_A) && AINS_A_EVAL(iins) == -1 )
    {
        switch( AINS_CODE(iins) )
        {
        case I_and:
            if( AINS_REG_B(iins) == REG_LIT )
                AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), AINS_DATA(iins));
            else
                AinsMakeLda(iins,AINS_REG_B(iins), AINS_REG_C(iins), 0);
            return 1;
        case I_bis:
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), -1);
            return 1;
        case I_bic:
            AinsMakeNot(iins,AINS_REG_B(iins), AINS_REG_C(iins), AINS_DATA(iins));
            return 1;
        case I_eqv:
            if( AINS_REG_B(iins) == REG_LIT )
                AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), AINS_DATA(iins));
            else
                AinsMakeLda(iins,AINS_REG_B(iins), AINS_REG_C(iins), 0);
            return 1;
        case I_ornot:
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), -1);
            return 1;
        case I_xor:
            AinsMakeNot(iins,AINS_REG_B(iins), AINS_REG_C(iins), AINS_DATA(iins));
            return 1;
        }
    }
    
    
    if( (AINS_FLAGS(iins) & FI_EVAL_B) && AINS_B_EVAL(iins) == -1 )
    {
        switch( AINS_CODE(iins) )
        {
        case I_and:
            AinsMakeLda(iins,AINS_REG_A(iins), AINS_REG_C(iins), 0);
            return 1;
        case I_bis:
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), -1);
            return 1;
        case I_bic:
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
            return 1;
        case I_eqv:
            AinsMakeLda(iins,AINS_REG_A(iins), AINS_REG_C(iins), 0);
            return 1;
        case I_ornot:
            AinsMakeLda(iins,AINS_REG_A(iins), AINS_REG_C(iins), 0);
            return 1;
        case I_xor:
            AinsMakeNot(iins,AINS_REG_A(iins), AINS_REG_C(iins), AINS_DATA(iins));
            return 1;
        }
    }
    
    return 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PeepHoleSFT"

static BOOL PeepHoleSFT(INDEX iins)
{
    if( AINS_REG_A(iins) == REG_ZERO )
    {
        AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
        return 1;
    }

    if( AINS_REG_B(iins) == REG_ZERO && AINS_CODE(iins) == I_zap )
    {
        AinsMakeLda(iins,AINS_REG_A(iins), AINS_REG_C(iins), 0);
        return 1;
    }

    if( AINS_REG_B(iins) == REG_ZERO && AINS_CODE(iins) == I_zapnot )
    {
        AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
        return 1;
    }

    if( AINS_REG_B(iins) == REG_LIT && AINS_DATA(iins) == 0xff && AINS_CODE(iins) == I_zap )
    {
        AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
        return 1;
    }

    if( AINS_REG_B(iins) == REG_LIT && AINS_DATA(iins) == 0xff && AINS_CODE(iins) == I_zapnot )
    {
        AinsMakeLda(iins,AINS_REG_A(iins), AINS_REG_C(iins), 0);
        return 1;
    }

    return 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PeepHoleICP"

static BOOL PeepHoleICP(INDEX iins)
{
    
    INDEX ibbl = AINS_BBL(iins);
    INDEX ifun = ABBL_FUN(ibbl);
    
    if( AINS_REG_A(iins) == AINS_REG_B(iins) )
    {
        /*HURRAH(("peephole evaluated %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));*/
            
        switch( AINS_CODE(iins) )
        {
        case I_cmpbge:
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0xff);
            return 1;
        case I_cmple:
        case I_cmpeq:
        case I_cmpule:
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 1);
            return 1;
        case I_cmpult:
        case I_cmplt:
            AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
            return 1;
        default:
            ERROR(("unexpected instruction %s\n",AINS_NAME(iins)));
        }
    }

    if( AINS_CODE(iins) == I_cmpult && AINS_REG_B(iins) == REG_ZERO )
    {
        /*HURRAH(("peephole evaluated %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));*/
        AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0);
        return 1;
    }

    if( AINS_CODE(iins) == I_cmpbge && AINS_REG_B(iins) == REG_ZERO )
    {
        
        /*HURRAH(("peephole evaluated %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));*/
        AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 0xff);
        return 1;
    }

    if( AINS_CODE(iins) == I_cmpule && AINS_REG_A(iins) == REG_ZERO )
    {
        HURRAH(("peephole evaluated %s in %d [%s] %ld\n",AINS_NAME(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
        AinsMakeLda(iins,REG_ZERO, AINS_REG_C(iins), 1);
        return 1;
    }

    
    if( AINS_CODE(iins) == I_cmplt && (AINS_FLAGS(iins) & FI_EVAL_A)  && AINS_A_EVAL(iins) == -1 )
    {
        HURRAH(("peephole simplified %s in bbl %d in fun %d [%s] %ld\n",AINS_NAME(iins),ibbl,ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
        AINS_CODE(iins) = I_cmple;
        AINS_REG_A(iins) = REG_ZERO;
        AINS_A_EVAL(iins) = 0;
        return 1;
    }
    
    return 0;
}


/*=======================================================
@OPT
Peephole optimizations and strength reductions
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptPeepHole"

extern BOOL OptPeepHole(FILE *fp)
{
#if 0
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Peep");
#endif
    BOOL change;
    NUMBER ecount=0;
    NUMBER ecounth=0;
    NUMBER kcount=0;
    NUMBER kcounth=0;
    NUMBER mcount=0;
    NUMBER mcounth=0;
    INDEX iins;
    INDEX ibbl;
    INDEX ifun;

    clock_t time = clock();

    FOREACH_LIVE_INS(iins)
    {
        if( AINS_FLAGS(iins) & FI_EVAL_C ) continue;

        ibbl = AINS_BBL(iins);
        ifun = ABBL_FUN(ibbl);

        switch( AINS_TYPE(iins) )
        {
        case IT_ULD:
            if( PeepHoleULD(iins) )
            {
                mcount++;
                if( ABBL_IS_HOT(ibbl) )
                    mcounth++;
            }
            break;

        case IT_UST:
            if( PeepHoleUST(iins) )
            {
                mcount++;
                if( ABBL_IS_HOT(ibbl) )
                    mcounth++;
            }
            break;
        case IT_ICM:
            if( PeepHoleICM(iins) )
            {
                mcount++;
                if( ABBL_IS_HOT(ibbl) )
                    mcounth++;
            }
            break;
            
        case IT_ICP:
            if( PeepHoleICP(iins) )
            {
                ecount++;
                if( ABBL_IS_HOT(ibbl) )
                    ecounth++;
            }
            break;

        case IT_IOP:
            if( PeepHoleIOP(iins) )
            {
                ecount++;
                if( ABBL_IS_HOT(ibbl) )
                    ecounth++;
            }
            break;
        case IT_LOG:
            if( PeepHoleLOG(iins) )
            {
                ecount++;
                if( ABBL_IS_HOT(ibbl) )
                    ecounth++;
            }
            break;
        case IT_SFT:
            if( PeepHoleSFT(iins) )
            {
                ecount++;
                if( ABBL_IS_HOT(ibbl) )
                    ecounth++;
            }
            break;

        case IT_MLL:
        case IT_MLQ:
        case IT_MLH:
            if( PeepHoleMUL(iins) )
            {
                mcount++;
                if( ABBL_IS_HOT(ibbl) )
                    mcounth++;
            }
            break;

        case IT_FM4:
        case IT_FM8:
            PeepHoleFMUL(iins);
            break;

        case IT_FD4:
        case IT_FD8:
            PeepHoleFDIV(iins);
            break;

        case IT_FA4:
        case IT_FA8:
            PeepHoleFADD(iins);
            break;

        case IT_FS4:
        case IT_FS8:
            PeepHoleFSUB(iins);
            break;

        default:
                break;
        }
    }
    

    change = (ecount+kcount)>0;
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;

    InfoStatus(fp,"\t[OptPeep%c] Eval Inss: %d#%d   Mod Ins: %d#%d  Killed Inss: %d#%d {%d ms}\n",
               CHANGE(change),ecount,ecounth,mcount,mcounth,kcount,kcounth,time);

        
    return change;
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "AbblFindSuitablePredecessor"

static INDEX AbblFindSuitablePredecessor(INDEX ibbl,INDEX succ)
{
    INDEX iedg;
    INDEX pred;
    

    FOREACH_BBL_PRED(ibbl,iedg)
    {
        pred = AEDG_HEAD(iedg);
        if( ABBL_TYPE(pred) != BT_BRANCH ) continue;
        if( AEDG_TAIL(ABBL_EDG_TRUE(pred)) != succ && AEDG_TAIL(ABBL_EDG_TRUE(pred)) != ibbl ) continue;
        if( AEDG_TAIL(ABBL_EDG_FALSE(pred)) != succ && AEDG_TAIL(ABBL_EDG_FALSE(pred)) != ibbl ) continue;

        return pred;
    }

    return INVALID;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblHasBadIns"

static BOOL AbblHasBadIns(INDEX ibbl)
{
    INDEX iins;
    
    FOREACH_BBL_INS(ibbl,iins)
    {
        switch( AINS_TYPE(iins) )
        {
        case IT_ILD:
        case IT_IST:
        case IT_ULD:
        case IT_UST:
        case IT_FLD:
        case IT_FST:
            return 1;
        case IT_LDA:
        case IT_ICP: 
        case IT_LOG:
        case IT_SFT:
        case IT_IOP:
        default:
            /* using live register we could do some tricks here like
               computing a value into an unused regsister and only use
               it, ie. conditionally move it, if the condition is true */
            break;
            

        }
    }
    
    return 0;
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "AbblsIsNotInteresting"

static BOOL AbblIsNotInteresting(INDEX ibbl,REG reg)
{
    INDEX iins;
    NUMBER new_instructions = 0;
    
    if( REG_IS_INT(reg) )
    {
        FOREACH_BBL_INS(ibbl,iins)
        {
            if( AINS_REG_C(iins) == reg && VALID(AINS_NEXT(iins)) ) return 1;
            
            switch( AINS_TYPE(iins) )
            {
            case IT_LDA:
                if( AINS_CODE(iins) != I_lda ) return 1;
                    if( AINS_DATA(iins) == 0 ) 
                    {
                        new_instructions++;
                    }
                    else if( AINS_DATA(iins) > 0 && AINS_DATA(iins) < 256 && AINS_REG_B(iins) == REG_ZERO )
                    {
                        new_instructions++;
                    }
                    else
                    {
                        return 1;
                    }
                    break;

            case IT_ICP: 
            case IT_LOG:
            case IT_SFT:
            case IT_IOP:
                /* using live register we could do some tricks here like
                   computing a value into an unused regsister and only use
                   it, ie. conditionally move it, if the condition is true */
                return 1;

            default:
                return 1;
            }
        }
        
        if( new_instructions > 2 ) return 1;
    }
    else
    {
        FOREACH_BBL_INS(ibbl,iins)
        {
            if( AINS_REG_C(iins) == reg && VALID(AINS_NEXT(iins)) ) return 1;
            
            switch( AINS_TYPE(iins) )
            {
            case IT_FCS:
                CASSERT( AINS_CODE(iins) == I_cpys,("weird\n"));
                if( AINS_REG_A(iins) != AINS_REG_B(iins) ) return 1;
                new_instructions++;
                break;

            default:
                return 1;
            }
        }
        
        if( new_instructions > 2 ) return 1;
        
    }
    
    return 0;
}

/*=======================================================
=======================================================*/



static struct
{
    NUMBER branch;
    NUMBER true;
    NUMBER false;
}
OpcodeTranslator[] =
{
    {I_blbc, I_cmovlbc, I_cmovlbs},
    {I_blbs, I_cmovlbs, I_cmovlbc},

    {I_beq,  I_cmoveq,  I_cmovne },
    {I_bne,  I_cmovne,  I_cmoveq },
    {I_ble,  I_cmovle,  I_cmovgt }, 
    {I_bgt,  I_cmovgt,  I_cmovle },
    {I_bge,  I_cmovge,  I_cmovlt },
    {I_blt,  I_cmovlt,  I_cmovge },

    {I_fbeq,  I_fcmoveq,  I_fcmovne },
    {I_fbne,  I_fcmovne,  I_fcmoveq },
    {I_fble,  I_fcmovle,  I_fcmovgt }, 
    {I_fbgt,  I_fcmovgt,  I_fcmovle },
    {I_fbge,  I_fcmovge,  I_fcmovlt },
    {I_fblt,  I_fcmovlt,  I_fcmovge },
    {0, 0, 0}
    
};

/*=======================================================
@OPT
Conditional move generation
Conditional moves are introduced for the basic block on the left hand side

@example
      O
      |\
      | O
      |/
      O
@end example

@END
=======================================================*/

#undef __PROC__
#define __PROC__ "OptCondMoveOne"

extern BOOL OptCondMoveOne(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Condmoveone");
    BOOL change;
    NUMBER count=0;
    NUMBER counth=0;
    NUMBER fcount=0;
    NUMBER fcounth=0;
    NUMBER i;
    UINT16 opcode;
    
    INDEX ibbl;
    INDEX pred;
    INDEX succ;
    INDEX iins;
    REG   reg_condition;
    
    clock_t time = clock();

    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_NORM ) continue;
        succ = AEDG_TAIL(ABBL_EDG_NORM(ibbl));
        pred = AbblFindSuitablePredecessor(ibbl,succ);
        if( !VALID(pred) ) continue;
        
        iins = ABBL_INS_LAST(pred);
        CASSERT( AINS_TYPE(iins) == IT_IBR ||
                 AINS_TYPE(iins) == IT_FBR,("unexpected instruction\n"));

        reg_condition = AINS_REG_A(iins);

        opcode = 0;
        
        for(i=0; OpcodeTranslator[i].branch!=0;i++)
        {
            if( OpcodeTranslator[i].branch == AINS_CODE(iins) )
            {
                if( AEDG_TAIL(ABBL_EDG_TRUE(pred)) == ibbl )
                    opcode = OpcodeTranslator[i].true;
                else
                    opcode = OpcodeTranslator[i].false;    
                break;
            }
            
        }
        CASSERT( opcode!=0,("weird\n"));

#if 0
        if( verbose )
        {
            fprintf(fp,"**Before:\n");
            
            PrintBblSuper(fp, pred);
            PrintBblSuper(fp, ibbl);
        }
#endif

#if 0
        if( ABBL_IS_HOT(ibbl) && !AbblHasBadIns(ibbl) )
        {
            HURRAH(("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"));
            PrintBblSuper(stdout,pred);
            PrintBblSuper(stdout,ibbl);
            PrintBblSuper(stdout,succ);
        }
#endif
        
        if( AbblIsNotInteresting(ibbl,reg_condition) ) continue;


        if( verbose )
        {
            fprintf(fp,"Before:\n");
            
            PrintBblSuper(fp, pred);
            PrintBblSuper(fp, ibbl);
        }

        
        AedgKillFancy(ABBL_EDG_TRUE(pred));
        AedgKillFancy(ABBL_EDG_FALSE(pred));
        AinsKill(iins);
        ABBL_TYPE(pred) = BT_NORM;
        AedgLinkFancy( AedgNewFancy(pred,succ,ET_NORM) );        


        FOREACH_BBL_INS(ibbl,iins)
        {
            INDEX jins = AinsNew();

            AINS_BBL(jins) = pred;
            AINS_CODE(jins) = opcode;

            switch( AINS_TYPE(iins) )
            {
            case IT_LDA:
                if( AINS_DATA(iins) == 0 )
                {
                    
                    AINS_REG_A(jins) = reg_condition;
                    AINS_REG_B(jins) = AINS_REG_B(iins);
                    AINS_REG_C(jins) = AINS_REG_C(iins);
                }
                else if( AINS_DATA(iins) > 0 && AINS_DATA(iins) < 256 && AINS_REG_B(iins) == REG_ZERO )
                {
                    AINS_REG_A(jins) = reg_condition;
                    AINS_REG_B(jins) = REG_LIT;
                    AINS_DATA(jins) = AINS_DATA(iins);
                    AINS_REG_C(jins) = AINS_REG_C(iins);                
                }
                else
                {
                    ERROR(("unexpected case\n"));
                }
                break;

            case IT_FCS:
                AINS_REG_A(jins) = reg_condition;
                AINS_REG_B(jins) = AINS_REG_B(iins);
                AINS_REG_C(jins) = AINS_REG_C(iins);
                break;

            default:
                ERROR(("unexpected case\n"));
            }
            
           
            AinsAppend(jins);
        }
        
        if( verbose )
        {
            fprintf(fp,"After:\n");
            PrintBblSuper(fp, pred);
        }

        if( REG_IS_INT(reg_condition) )
        {
            count++;
            if( ABBL_IS_HOT(pred) ) counth++;
        }
        else
        {
            fcount++;
            if( ABBL_IS_HOT(pred) ) fcounth++;
        }
        
    }

    change = (count)>0;
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;

    InfoStatus(fp,"\t[OptCondmoveone%c] Added ICMs: %d#%d  Added FCM: %d#%d  {%d ms}\n",
            CHANGE(change),count,counth,fcount,fcounth,time);

    return change;
}


/*=======================================================
@OPT

Conditional move generation.
Conditional moves are introduced for the basic blocks in the middle

@example
      O
     / \
    O   O
     \ /
      O
@end example
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptCondMoveTwo"

extern BOOL OptCondMoveTwo(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Condmovetwo");
    BOOL change;
    NUMBER count=0;
    NUMBER counth=0;
    NUMBER fcount=0;
    NUMBER fcounth=0;
    NUMBER opcode;
    
    INDEX iins;
    INDEX jins;
    INDEX iins_t;
    INDEX iins_f;

    INDEX ibbl;
    INDEX ibbl_t;
    INDEX ibbl_f;
    INDEX ibbl_join;

    REGLIST avail;
    REG   reg;
    REG   reg_condition;

    clock_t time = clock();

    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_BRANCH ) continue;
        iins = ABBL_INS_LAST(ibbl);
        CASSERT( AINS_TYPE(iins) == IT_IBR ||
                 AINS_TYPE(iins) == IT_FBR,("unexpected instruction\n"));
        reg_condition = AINS_REG_A(iins);

        ibbl_t = AEDG_TAIL(ABBL_EDG_TRUE(ibbl));
        ibbl_f = AEDG_TAIL(ABBL_EDG_FALSE(ibbl));
        
        if( ABBL_TYPE(ibbl_f) != BT_NORM )  continue;
        if( ABBL_TYPE(ibbl_t) != BT_NORM )  continue;

        ibbl_join = AEDG_TAIL(ABBL_EDG_NORM(ibbl_t));
        if( AEDG_TAIL(ABBL_EDG_NORM(ibbl_f)) != ibbl_join ) continue;

#if 0
        if( ABBL_IS_HOT(ibbl) && !AbblHasBadIns(ibbl_f) && !AbblHasBadIns(ibbl_t) )
        {
            HURRAH(("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"));
            PrintBblSuper(stdout,ibbl);
            PrintBblSuper(stdout,ibbl_t);
            PrintBblSuper(stdout,ibbl_f);
            PrintBblSuper(stdout,ibbl_join);
        }
#endif
        
        if( AbblNInss(ibbl_t) > 1 ) continue;
        if( AbblNInss(ibbl_f) > 1 ) continue;

#if 0
        if( AbblNpreds(ibbl_f) > 1 ) continue;
        if( AbblNpreds(ibbl_t) > 1 ) continue;
#endif
        
        iins_f = ABBL_INS_FIRST(ibbl_f);
        iins_t = ABBL_INS_FIRST(ibbl_t);
#if 0        
        if( AbblNpreds(ibbl_join) > 2 ) continue;
#endif

            
        for(opcode=0; OpcodeTranslator[opcode].branch!=0;opcode++)
        {
            if( OpcodeTranslator[opcode].branch == AINS_CODE(iins) )
                  break;
         }
 
        /* check whether instructions iins_t and iins_f are suitable */
        
        if( REG_IS_INT(reg_condition) )
        {
            if( AINS_CODE(iins_f) != I_lda ) continue;
            if( AINS_DATA(iins_f) !=0  &&
                (AINS_DATA(iins_f) < 0 || AINS_DATA(iins_f) >= 256 || AINS_REG_B(iins_f) != REG_ZERO) ) continue;
            if( AINS_CODE(iins_t) != I_lda ) continue;
            if( AINS_DATA(iins_t) !=0  &&
                (AINS_DATA(iins_t) < 0 || AINS_DATA(iins_t) >= 256 || AINS_REG_B(iins_t) != REG_ZERO) ) continue;

#if 0
            /* we can handle this now gracefully */
            CASSERT( AINS_REG_C(iins_t) != reg_condition,("weird\n"));
            CASSERT( AINS_REG_C(iins_f) != reg_condition,("weird\n"));
#endif
        }
        else
        {
            if( AINS_CODE(iins_f) != I_cpys ) continue;
            if( AINS_REG_A(iins_f) != AINS_REG_B(iins_f) ) continue;
            if( AINS_CODE(iins_t) != I_cpys ) continue;
            if( AINS_REG_A(iins_t) != AINS_REG_B(iins_t) ) continue;
#if 0
            /* we can handle this now gracefully */
            CASSERT( AINS_REG_C(iins_t) != reg_condition,("weird\n"));
            CASSERT( AINS_REG_C(iins_f) != reg_condition,("weird\n"));
#endif
        }
        
        /*
          check whether the condition register might get overwritten 
          we need live register analysis here
         */

        if( AINS_REG_C(iins_f) == reg_condition || AINS_REG_C(iins_t) == reg_condition )
        {
            if( REG_IS_INT(reg_condition) )
            {
                avail = REGS_INT_BUT_ZERO & ~ABBL_OUT(ibbl);
                avail &= ~REG2BIT(reg_condition);
                avail &= ~REG2BIT(AINS_REG_C(iins_t));
                avail &= ~REG2BIT(AINS_REG_C(iins_f));
                if( avail == REGS_NONE )
                {
                    WARNING(("failed to find a free register \n"));
                    continue;
                }
                                    
                reg = RegsGetFirst(avail);
                jins = AinsNew();
                AINS_BBL(jins) = ibbl;
                AinsMakeLda(jins,reg_condition,reg,0);
                AinsAppend(jins);
            }
            else
            {
                avail = REGS_FLT_BUT_ZERO & ~ABBL_OUT(ibbl);
                if( avail == REGS_NONE )
                {
                    WARNING(("failed\n"));
                    continue;
                }
                WARNING(("succeeded\n"));
                
                reg = RegsGetFirst(avail);
                jins = AinsNew();
                AINS_BBL(jins) = ibbl;
                AinsMakeCpys(jins,reg_condition,reg_condition,reg);
                AinsAppend(jins);
            }

            reg_condition = reg;
        }
        
                
            
        if( verbose )
        {
            fprintf(fp,"Before\n");
            PrintBblSuper(fp, ibbl);
            PrintBblSuper(fp, ibbl_t);
            PrintBblSuper(fp, ibbl_f);
            PrintBblSuper(fp, ibbl_join);
        }
            


        if( REG_IS_INT(reg_condition) )
        {
            jins = AinsNew();
            AINS_BBL(jins) = ibbl;

            if( AINS_REG_C(iins_t) == AINS_REG_C(iins_f) )
            {
                AinsMakeLda(jins,AINS_REG_B(iins_t),AINS_REG_C(iins_t),AINS_DATA(iins_t));
            }
            else
            {
                AINS_REG_A(jins) = reg_condition;
                AINS_CODE(jins) = OpcodeTranslator[opcode].true;
                if( AINS_DATA(iins_t) == 0 )
                {
                    AINS_REG_B(jins) = AINS_REG_B(iins_t);
                    AINS_REG_C(jins) = AINS_REG_C(iins_t);
                }
                else 
                {
                    AINS_REG_B(jins) = REG_LIT;
                    AINS_DATA(jins) = AINS_DATA(iins_t);
                    AINS_REG_C(jins) = AINS_REG_C(iins_t);                
                }
            }
            
            AinsAppend(jins);
            
            jins = AinsNew();
            AINS_BBL(jins) = ibbl;
            AINS_REG_A(jins) = reg_condition;
            AINS_CODE(jins) = OpcodeTranslator[opcode].false;
                        
            if( AINS_DATA(iins_f) == 0 )
            {
                AINS_REG_B(jins) = AINS_REG_B(iins_f);
                AINS_REG_C(jins) = AINS_REG_C(iins_f);
            }
            else 
            {
                AINS_REG_B(jins) = REG_LIT;
                AINS_DATA(jins) = AINS_DATA(iins_f);
                AINS_REG_C(jins) = AINS_REG_C(iins_f);                
            }
            AinsAppend(jins);

        }
        else
        {
            jins = AinsNew();
            AINS_BBL(jins) = ibbl;

            if( AINS_REG_C(iins_t) == AINS_REG_C(iins_f) )
            {
                AinsMakeCpys(jins,AINS_REG_A(iins_t),AINS_REG_B(iins_t),AINS_REG_C(iins_t));
            }
            else
            {
                AINS_REG_A(jins) = reg_condition;
                AINS_CODE(jins) = OpcodeTranslator[opcode].true;
                AINS_REG_B(jins) = AINS_REG_B(iins_t);
                AINS_REG_C(jins) = AINS_REG_C(iins_t);
            }
            AinsAppend(jins);
                        
            jins = AinsNew();
            AINS_REG_A(jins) = reg_condition;
            AINS_CODE(jins) = OpcodeTranslator[opcode].false;
            AINS_BBL(jins) = ibbl;
            AINS_REG_B(jins) = AINS_REG_B(iins_f);
            AINS_REG_C(jins) = AINS_REG_C(iins_f);
            AinsAppend(jins);
        }
        
        AedgKillFancy(ABBL_EDG_TRUE(ibbl));
        AedgKillFancy(ABBL_EDG_FALSE(ibbl));
        AinsKill(iins);
        ABBL_TYPE(ibbl) = BT_NORM;
        AedgLinkFancy( AedgNewFancy(ibbl,ibbl_join,ET_NORM) );
        
        
        if( verbose )
        {
            fprintf(fp,"After\n");
            PrintBblSuper(fp, ibbl);
            PrintBblSuper(fp, ibbl_t);
            PrintBblSuper(fp, ibbl_f);
            PrintBblSuper(fp, ibbl_join);
        }


        if( REG_IS_INT(reg_condition) )
        {
            count++;
            if( ABBL_IS_HOT(ibbl) ) counth++;
        }
        else
        {
            fcount++;
            if( ABBL_IS_HOT(ibbl) ) fcounth++;
        }
    }
    
    change = (count)>0;
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;

    InfoStatus(fp,"\t[OptCondmovetwo%c] Added ICMs: %d#%d  Added FCM: %d#%d  {%d ms}\n",
            CHANGE(change),count,counth,fcount,fcounth,time);

    return change;
}

/*=======================================================
*=======================================================*/

