/*=======================================================
  Author:   R. Muth
  Version:  $Id: potential.c,v 1.13 1999/08/05 19:37:17 muth Exp $
  Comment:  This file contains the optimization code
=======================================================*/

#include "global.h"
#include "flags.h"
#include "opcode.h"
#include "functions.h"
#include "regs.h"

extern INT64 potential[];

INT64 potential[ CONFIG_NINS + (64*CONFIG_NBBL)];




#define AINS_USE_DEF_POTENTIAL(i) potential[i]

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsUseDefConstAncestor"

extern INDEX AinsUseDefConstAncestor(INDEX iins)
{
    if( AINS_TYPE(iins) == IT_ILD ||
        AINS_TYPE(iins) == IT_FLD ||
        AINS_TYPE(iins) == IT_ULD ||
        AINS_TYPE(iins) == IT_LLD ||
        AINS_TYPE(iins) == IT_IST ||
        AINS_TYPE(iins) == IT_FST ||
        AINS_TYPE(iins) == IT_UST ||
        AINS_TYPE(iins) == IT_CST )
        return INVALID;

    if( !(AINS_FLAGS(iins) & FI_EVAL_A) && (AINS_FLAGS(iins) & FI_EVAL_B) )
    {
        return AINS_A_EVAL(iins);
    }
    else if( (AINS_FLAGS(iins) & FI_EVAL_A) && !(AINS_FLAGS(iins) & FI_EVAL_B) )
    {
        return AINS_B_EVAL(iins);
    }
    else if( !(AINS_FLAGS(iins) & FI_EVAL_A) && !(AINS_FLAGS(iins) & FI_EVAL_B) && AINS_REG_A(iins) == AINS_REG_B(iins) )
    {
        return AINS_B_EVAL(iins);
    }
    else
    {
        return INVALID;
    }
}

/*=======================================================  
=======================================================*/

extern BOOL AinsUseDefIsAncestor(INDEX ancestor, INDEX iins)
{
    if( !VALID(ancestor) ) return 0;
    if( !VALID(iins) ) return 0;
    if( ancestor == iins ) return 1;
    if( iins >=  CONFIG_NINS ) return 0;
    
    return AinsUseDefIsAncestor(ancestor,AinsUseDefConstAncestor(iins));
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsUseDefPotential"

static INT64 AinsUseDefPotentialRegC(INDEX iins)
{
    switch( AINS_TYPE(iins) )
    {
    case IT_ILD:   
    case IT_FLD:
    case IT_ULD:
    case IT_LLD:          
        return 4*2;
    case IT_IST:
    case IT_FST:
    case IT_UST:
    case IT_CST:
        return 4*0;
    
    case IT_ICM:    /* the potential actually depends on which reg is involved */
    case IT_FCM:
        
    case IT_ICP:
    case IT_FCP:
            
    case IT_LDA:

    case IT_IOP:
        
    case IT_F2I:
    case IT_I2F:
    case IT_LOG:
    case IT_SFT:
    case IT_EXT:
        return 4*1;

    case IT_MSC:
    case IT_BRA:
    case IT_PAL:
    case IT_F2R:
    case IT_RX:
    case IT_MEM:
        return 4*0;
        
    case IT_IBR:
    case IT_FBR:
    case IT_JMP:
        return 4*4;
        
    case IT_FM4: 
    case IT_FM8:
    case IT_FA4:
    case IT_FA8:
    case IT_FS4:
    case IT_FS8:
    case IT_FCV:
    case IT_FCS:
    case IT_FCN:
    case IT_FCE:
        return 4*4-2*4;
        
    case IT_FD4:
        return 4*30-2*4;
                
    case IT_FD8:
        return 4*60-2*4;
        
    case IT_MLL:
    case IT_MLQ:
    case IT_MLH:
        return 4*16;
        
    default:
        ERROR(("unexpected instruction type\n"));
        PrintIns(stdout,iins);
        return 0;
    }
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsUseDefPotentialRegA"

static INT64 AinsUseDefPotentialRegA(INDEX iins)
{
    switch( AINS_TYPE(iins) )
    {
    case IT_ILD:   
    case IT_FLD:
    case IT_ULD:
    case IT_LLD:          
        
    case IT_IST:
    case IT_FST:
    case IT_UST:
    case IT_CST:
        return 0;
    
    case IT_ICM:    
    case IT_FCM:
        return 2;
            
    case IT_ICP:
    case IT_FCP:
            
    case IT_LDA:

    case IT_IOP:
        
    case IT_F2I:
    case IT_I2F:
    case IT_LOG:
    case IT_SFT:
    case IT_EXT:
        return 0;

    case IT_MSC:
    case IT_BRA:
    case IT_PAL:
    case IT_F2R:
    case IT_RX:
    case IT_MEM:
        return 0;
        
    case IT_IBR:
    case IT_FBR:
        return 4*4;  
    case IT_JMP:
        return 0;
        
    case IT_FA4:
    case IT_FA8:
    case IT_FS4:
    case IT_FS8:
        return 1;

    case IT_FCS:
    case IT_FCN:
    case IT_FCE:
    case IT_FCV:
        return 0;

        /* speculate on strength reductions */
        
    case IT_FM4: 
    case IT_FM8:
        return 2;
        
    case IT_FD4:
        return 4;
                
    case IT_FD8:
        return 8;
        
    case IT_MLL:
    case IT_MLQ:
    case IT_MLH:
        return 2;
        
    default:
        ERROR(("unexpected instruction type\n"));
        PrintIns(stdout,iins);
        return 0;
    }
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsUseDefPotentialRegB"

static INT64 AinsUseDefPotentialRegB(INDEX iins)
{
    switch( AINS_TYPE(iins) )
    {
    case IT_ILD:   
    case IT_FLD:
    case IT_ULD:
    case IT_LLD:          
        
    case IT_IST:
    case IT_FST:
    case IT_UST:
    case IT_CST:
        return 1;
    
    case IT_ICM:    /* the potential actually depends on which reg is involved */
    case IT_FCM:
        return 1;

    case IT_LDA:
        return 4;
        
    case IT_ICP:
    case IT_FCP:
    case IT_IOP:        
    case IT_F2I:
    case IT_I2F:
    case IT_LOG:
    case IT_SFT:
    case IT_EXT:
        return 0;

    case IT_MSC:
    case IT_BRA:
    case IT_PAL:
    case IT_F2R:
    case IT_RX:
    case IT_MEM:
        return 0;
        
    case IT_IBR:
    case IT_FBR:
        return 0;
    case IT_JMP:
        return 4*4;  
        
    case IT_FA4: 
    case IT_FA8:
    case IT_FS4: 
    case IT_FS8:
        return 1;

    case IT_FCV:
    case IT_FCS:
    case IT_FCN:
    case IT_FCE:
        return 0;

        /* speculate on strength reductions */

    case IT_FM4: 
    case IT_FM8:
        return 2;

    case IT_FD4:
        return 15;   /* can be converted to muliplications */
        
    case IT_FD8:
        return 22;   /* can be converted to muliplications */
        
    case IT_MLL:
    case IT_MLQ:
    case IT_MLH:
        return 2;
        
    default:
        ERROR(("unexpected instruction type\n"));
        PrintIns(stdout,iins);
        return 0;
    }
}
/*=======================================================
=======================================================*/

#define FI_HAS_BEEN_PROCESSED F_LOCAL_1
#define FI_CANNOT_BE_PROCESSED F_LOCAL_2

#define AEDG_VALUE(i) AEDG_U1(i).quadword
#define AEDG_REGS_CONST(i) AEDG_U2(i).quadword


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ComUseDefConstPot"

extern void ComUseDefConstPot(FILE *fp)
{
    INDEX iins;
    INDEX ibbl;
    INDEX def;
    INDEX i;
    NUMBER  count_unprocessed;
    NUMBER  count_processed;

    /* zero initialized !!! */
    
    clock_t time = clock();

    PrintProfile(fp);

        
    for(i=0;i<CONFIG_NINS + 64 * CONFIG_NBBL;i++ )
        AINS_USE_DEF_POTENTIAL(i) = 0;
    
    FOREACH_LIVE_INS(iins)
    {
        if( !(AINS_FLAGS(iins) & FI_EVAL_A) )
        {
            def = AINS_A_EVAL(iins);
            if( def >  CONFIG_NINS + 64 * CONFIG_NINS + 63 )
            {
                
                ERROR(("weird a-def %d for iins: %d %d\n",def,iins,(def-CONFIG_NINS)/64));
                PrintBblSuper(stdout,AINS_BBL(iins));
            }
        }
        
        if( !(AINS_FLAGS(iins) & FI_EVAL_B) )
        {
            def = AINS_B_EVAL(iins);
            if( def >  CONFIG_NINS + 64 * CONFIG_NINS + 63 )
            {
                
                ERROR(("weird b-def %d for iins: %d %d\n",def,iins,(def-CONFIG_NINS)/64));
                PrintBblSuper(stdout,AINS_BBL(iins));
            }
        }
    }
        
#if 0
    WARNING(("INIT\n"));
#endif

    FOREACH_LIVE_BBL(ibbl)
    {
        FOREACH_BBL_INS(ibbl,iins)
        {
            AINS_FLAGS(iins) &= ~FI_HAS_BEEN_PROCESSED;
        }
    }
    
#if 0
    WARNING(("PROPAGATE\n"));
#endif

    /*
      processing an instruction means to compute its potential
      the potential can be computed if the the potential of
      all its dependees has been computed.
      If this is not the case we set FI_CANNOT_BE_PROCESSED.
      Once the potential is computed we set FI_HAS_BEEN_PROCESSED
      */
        

    do
    {
        FOREACH_LIVE_INS(iins)
        {
            if( AINS_FLAGS(iins) & FI_HAS_BEEN_PROCESSED )
            {
                AINS_FLAGS(iins) |= FI_CANNOT_BE_PROCESSED;
            }
            else
            {
                AINS_FLAGS(iins) &= ~FI_CANNOT_BE_PROCESSED;
            }
        }

#if 0
        WARNING(("PROPAGATE CANDIDATES\n"));
#endif


        /* mark unprocessable ones */
        
        FOREACH_LIVE_INS(iins)
        {
            if( AINS_FLAGS(iins) & FI_HAS_BEEN_PROCESSED ) continue;

            if( !(AINS_FLAGS(iins) & FI_EVAL_A) )
            {
                def = AINS_A_EVAL(iins);
                if( def < CONFIG_NINS  && VALID(def) && !(AINS_FLAGS(def) & FI_HAS_BEEN_PROCESSED) )
                    AINS_FLAGS(def) |= FI_CANNOT_BE_PROCESSED;
            }

            if( !(AINS_FLAGS(iins) & FI_EVAL_B) )
            {
                def = AINS_B_EVAL(iins);
                if( def < CONFIG_NINS  && VALID(def) && !(AINS_FLAGS(def) & FI_HAS_BEEN_PROCESSED) )
                    AINS_FLAGS(def) |= FI_CANNOT_BE_PROCESSED;
            }
        }
        
#if 0
        WARNING(("PROPAGATE PROPAGATE\n"));
#endif
        /*
          propagate potential to processable instructions
          if an instruction has not been processed its value has not stabalized
          and we can therfore not propagate that value
          */
            

        FOREACH_LIVE_INS_R(iins)
        {
            if( !(AINS_FLAGS(iins) & FI_HAS_BEEN_PROCESSED) ) continue;

            /* check evaluability */
            
            def = AinsUseDefConstAncestor(iins);
            if( VALID(def) && def < CONFIG_NINS )
            {
                if( !(AINS_FLAGS(def) & FI_HAS_BEEN_PROCESSED) && !(AINS_FLAGS(def) & FI_CANNOT_BE_PROCESSED) )
                    AINS_USE_DEF_POTENTIAL(def) += AINS_USE_DEF_POTENTIAL(iins) +
                                                   AinsUseDefPotentialRegC(iins) * ABBL_COUNT(AINS_BBL(iins));
                continue;
            }

            if( !(AINS_FLAGS(iins) & FI_EVAL_A) )
            {
                def = AINS_A_EVAL(iins);
                if( def < CONFIG_NINS  && VALID(def) && !(AINS_FLAGS(def) & FI_CANNOT_BE_PROCESSED) )
                    AINS_USE_DEF_POTENTIAL(def) += AinsUseDefPotentialRegA(iins) * ABBL_COUNT(AINS_BBL(iins));
            }

            if( !(AINS_FLAGS(iins) & FI_EVAL_B) )
            {
                def = AINS_B_EVAL(iins);
                if( def < CONFIG_NINS  && VALID(def) && !(AINS_FLAGS(def) & FI_CANNOT_BE_PROCESSED) )
                    AINS_USE_DEF_POTENTIAL(def) += AinsUseDefPotentialRegB(iins) * ABBL_COUNT(AINS_BBL(iins));
            }
        }

#if 0
        WARNING(("PROPAGATE CLEANUP\n"));
#endif

        count_processed = 0 ;
        count_unprocessed = 0;
        
        FOREACH_LIVE_INS_R(iins)
        {
            if( AINS_FLAGS(iins) & FI_HAS_BEEN_PROCESSED ) continue;
            if( AINS_FLAGS(iins) & FI_CANNOT_BE_PROCESSED )
            {
                count_unprocessed++;

                continue;
            }
            
            AINS_FLAGS(iins) |= FI_HAS_BEEN_PROCESSED;
            count_processed++;
        }

#if 0
        printf("unpr %d   prc %d\n",count_unprocessed,count_processed);
#endif
        
        if( count_unprocessed > 0 && count_processed == 0 )
        {
            printf("circle detected\n");
	    break;
        }
        
    }
    while( count_unprocessed > 0 );

    
#if 0
    WARNING(("PROPAGATE PSEUDO INS\n"));
#endif
    
    /* now handle confluence points separately, we do not have processed flags for these */
        
    FOREACH_LIVE_INS(iins)
    {
        def = AinsUseDefConstAncestor(iins);
        if( VALID(def) && def >= CONFIG_NINS )
        {
            AINS_USE_DEF_POTENTIAL(def) += AINS_USE_DEF_POTENTIAL(iins) +
                                           AinsUseDefPotentialRegC(iins) * ABBL_COUNT(AINS_BBL(iins));
            continue;
        }
        
        if( !(AINS_FLAGS(iins) & FI_EVAL_A) )
        {
            def = AINS_A_EVAL(iins);
            if( VALID(def) && def >= CONFIG_NINS )
            AINS_USE_DEF_POTENTIAL(def) += AinsUseDefPotentialRegA(iins) * ABBL_COUNT(AINS_BBL(iins));
        }

        if( !(AINS_FLAGS(iins) & FI_EVAL_B) )
        {
            def = AINS_B_EVAL(iins);
            if( VALID(def) && def >= CONFIG_NINS )
                AINS_USE_DEF_POTENTIAL(def) += AinsUseDefPotentialRegB(iins) * ABBL_COUNT(AINS_BBL(iins));
        }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;

    InfoStatus(fp,"\t[ComUseDefConstPot]  {%d ms}\n",time);
    

}


/*=======================================================
=======================================================*/
#if 0

#undef __PROC__
#define __PROC__ "ValueProfInstrument"

extern void ValProfInstrument(FILE *fp)
{
    INDEX iins;
    INDEX ibbl;
    INDEX ifun;
    INDEX iedg;
    INDEX def;
    INT64 hcount;
    INT64 hpotential;
    NUMBER  count_unprocessed;
    NUMBER  count_processed;
    REG reg;

    /* zero initialized !!! */
    
    clock_t time = clock();


#if 0
    WARNING(("PRINT\n"));
#endif

    FOREACH_LIVE_EDG(iedg)
    {
        AEDG_REGS_CONST(iedg) = REGS_NONE;
    }
    
    for(reg=0;reg<63;reg++)
    {
        if( reg == REG_ZERO || reg == REG_FZERO ) continue;
        EvalConstPropagationEdges(reg, TheLatticeBot, TheLatticeTop);
        FOREACH_LIVE_BBL(ibbl)
        {
#if 0
            if( !ABBL_IS_HOT(ibbl) ) continue;
#else
            if( ABBL_COUNT(ibbl) <= 0 ) continue;
#endif
            if( ABBL_FLAGS(ibbl) & FB_LOOP_HEADER ) continue;
            iins = CONFIG_NINS + 64 * ibbl + reg;
            if( AINS_USE_DEF_POTENTIAL(iins) / (2+ABBL_COUNT(ibbl)) < 4 ) continue;
            FOREACH_BBL_PRED(ibbl,iedg)
            {
                INDEX pred = AEDG_HEAD(iedg);

                if( !ABBL_IS_HOT(pred) ) continue;
                if( AEDG_VALUE(iedg) == TheLatticeTop ) continue;
                if( AEDG_VALUE(iedg) == TheLatticeBot ) continue;
                if( AEDG_TYPE(iedg) == ET_FUN_CALL ) continue;

                AEDG_REGS_CONST(iedg) |= REG2BIT(reg);
                ifun = ABBL_FUN(ibbl);
                printf("\t[*]fun: %d [%s] hd: %d tl: %d reg: %d [%016lx]\n",
                       ifun,AFUN_NAME(ifun),pred,ibbl,reg,AEDG_VALUE(iedg));
            }
            
        }
        
    }
    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
}

#endif

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "ComUseDefConstPotValueProfile"

extern void ComUseDefConstPotValueProfile(FILE *fp)
{
    INDEX iins;
    INDEX ibbl;
    INDEX ifun;
    INDEX icount = 0;
    INDEX bcount = 0;
    INDEX pcount = 0;
    
    REG reg;

    /* zero initialized !!! */
    
    clock_t time = clock();

    /* ====================================================================== */

    FOREACH_LIVE_FUN(ifun)
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        REGLIST reglist = REGS_NONE;

        FOREACH_BBL_INS(ibbl,iins)
        {
            if( AINS_USE_DEF_POTENTIAL(iins) < CONFIG_COUNT ) continue;
            if( AINS_USE_DEF_POTENTIAL(iins) <  TestCost(0) * ABBL_COUNT(ibbl) ) continue;
            reg = AINS_REG_C(iins);
            reglist |= REG2BIT(reg);

            if( reg == REG_ZERO || reg == REG_FZERO || reg == REG_SP || reg == REG_GP ) continue;

            printf("\t[*] fun %5d  bbl %6d  ins %8d  reg %2d  cnt %9ld  pot %11ld  [%s]\n",
                   ifun,ibbl,iins,reg,ABBL_COUNT(ibbl),AINS_USE_DEF_POTENTIAL(iins),AFUN_NAME(ifun));
            AfunMarkUsesTransitive(ifun, ibbl, iins);
            icount++;
        }

        for(reg=0;reg<63;reg++)
        {
            if( reg == REG_ZERO || reg == REG_FZERO || reg == REG_SP || reg == REG_GP ) continue;
                
            iins = CONFIG_NINS + 64 * ibbl + reg;

            if( AINS_USE_DEF_POTENTIAL(iins) < CONFIG_COUNT ) continue;
            if( AINS_USE_DEF_POTENTIAL(iins) < TestCost(0) * ABBL_COUNT(ibbl) ) continue;
            reglist |= REG2BIT(reg);

            printf("\t[*] fun %5d  bbl %6d  ins %8d  reg %2d  cnt %9ld  pot %11ld  [%s]\n",
                   ifun,ibbl,iins,reg,ABBL_COUNT(ibbl),AINS_USE_DEF_POTENTIAL(iins),AFUN_NAME(ifun));
            AfunMarkUsesTransitive(ifun, ibbl, iins);
            pcount++;
        }

        if( reglist )
        {
            bcount++;
#if 0
            PrintBblSuper(stdout,ibbl);
#endif   
        }

    }
        
    /* ====================================================================== */
    /* Beware: This is hairy! Since we are adding code while iterating over it */
    /* ====================================================================== */

    ValProfInstrumentInit();

    FOREACH_LIVE_FUN(ifun)
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        FOREACH_BBL_INS_R(ibbl,iins)
        {
            if( AINS_USE_DEF_POTENTIAL(iins) < CONFIG_COUNT ) continue;
            if( AINS_USE_DEF_POTENTIAL(iins) <  TestCost(0) * ABBL_COUNT(ibbl) ) continue;
            reg = AINS_REG_C(iins);

            if( reg == REG_ZERO || reg == REG_FZERO || reg == REG_SP || reg == REG_GP ) continue;

            ValProfInstrumentProgramPoint(iins, reg, iins, -1);
            AINS_USE_DEF_POTENTIAL(iins) = 0;
        }

        for(reg=0;reg<63;reg++)
        {
            if( reg == REG_ZERO || reg == REG_FZERO || reg == REG_SP || reg == REG_GP ) continue;
            
            iins = CONFIG_NINS + 64 * ibbl + reg;

            if( AINS_USE_DEF_POTENTIAL(iins) < CONFIG_COUNT ) continue;
            if( AINS_USE_DEF_POTENTIAL(iins) <  TestCost(0) * ABBL_COUNT(ibbl) ) continue;

            ValProfInstrumentProgramPoint(iins, reg, -1, ibbl);
            AINS_USE_DEF_POTENTIAL(iins) = 0;
        }
    }
    
    ValProfInstrumentExit();

    /* ====================================================================== */
    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;

    InfoStatus(fp,"\t[ComUseDefConstPotValueProfile] bbl: %d  ins: %d  pins: %d  {%d ms}\n",
               bcount,icount,pcount,time);
}


/*=======================================================
@OPT
@END
=======================================================*/

#if 0

#undef __PROC__
#define __PROC__ "OptDupRegionOne"

extern BOOL OptDupRegionOne(FILE *fp, BOOL no_code_growth)
{
    BOOL change = 0;
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Dupregionone");
    INDEX hd;
    INDEX tl;
    INDEX iedg;
    INDEX jedg;
    INDEX ibbl;
    INDEX ifun;
    INDEX succ;
    double flowfrac;
    NUMBER bcount = 0;
    NUMBER bcounth = 0;
    NUMBER icount = 0;
    NUMBER icounth = 0;
    NUMBER n;
    REG reg;
    
    clock_t time = clock();

    FOREACH_LIVE_EDG_R(iedg)
    {
        if( no_code_growth ) continue;
        if( AEDG_REGS_CONST(iedg) == REGS_NONE ) continue;

        hd  = AEDG_HEAD(iedg);
        tl  = AEDG_TAIL(iedg);
        
        ifun = ABBL_FUN(tl);
        
        FOREACH_FUN_BBL(ifun,ibbl)
        {
            ABBL_FLAGS(ibbl) &= ~F_MARKED;
        }

        flowfrac = AEDG_COUNT(iedg) / ABBL_COUNT(tl);

        printf("\t[*]fun: %d [%s] hd: %d tl: %d flowfrac: %4.2g regs: ",
               ifun,AFUN_NAME(ifun),hd,tl,flowfrac);

        for( reg = 0; reg <= 63; reg++ )
        {
            if( REG2BIT(reg) &  AEDG_REGS_CONST(iedg) )
            {
                printf("%d ",reg);
                AfunMarkUsesTransitive(ifun,tl,CONFIG_NINS + 64 * tl + reg);
            }
            
        }

        printf("\n");
        
        PrintBblSuper(stdout,hd);
        FOREACH_FUN_BBL(ifun,ibbl)
        {
            if( ABBL_FLAGS(ibbl) & F_MARKED )
                PrintBblSuper(stdout,ibbl);
        }

        WARNING(("Cloning %d %s %d\n",ifun, AFUN_NAME(ifun),ibbl));
        
        ibbl = AfunCloneMarkedBbls(ifun,tl,flowfrac);

        WARNING(("Edging %d %s %d\n",ifun, AFUN_NAME(ifun),ibbl));
        
        jedg = AedgNewFancy(hd,ibbl, AEDG_TYPE(iedg));
        AedgKillFancy(iedg);
        AedgLinkFancy(jedg);
    }
    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = (bcount>0);

    InfoStatus(fp,"\t[OptDupbblone%c] Dup Bbls: %d#%d  Inss: %d#%d {%d ms}\n",
            CHANGE(change),bcount,bcounth,icount,icounth,time);
    
    return change;
}

#endif

/*=======================================================
=======================================================*/









