/*=======================================================
  Author:   R. Muth, A. Kagedal
  Version:  $Id: print.c,v 1.47 1999/08/05 19:37:41 muth Exp $
  Comment:  This file contains the low level printing code
=======================================================*/


#include "global.h"
#include "functions.h"
#include "flags.h"

/*=======================================================
=======================================================*/

extern void PrintSep0(FILE *fp)
{
    fprintf(fp,"########################################################################\n");
}

/*=======================================================
=======================================================*/

extern void PrintSep1(FILE *fp)
{
    fprintf(fp,"========================================================================\n");
}

/*=======================================================
=======================================================*/

extern void PrintSep2(FILE *fp)
{
    fprintf(fp,"------------------------------------------------------------------------\n");
}


/*=======================================================
=======================================================*/

extern void PrintSep3(FILE *fp)
{
    fprintf(fp,"************************************************************************\n");
}

/*=======================================================
=======================================================*/

extern void PrintNL(FILE *fp)
{
    fprintf(fp,"\n");
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "PrintHexDump"

#include <ctype.h>

static char toprint(signed char c)
{
    if( c == 0) return '.';
    return (c>0 && isprint(c)) ? c : '_';
}

extern void PrintHexDump(FILE *fp, UINT64 address, UINT8 *s, UINT32 length)
{
    CWARNING( (length & 0xf)!=0,("len should be a multiple of 16"));
    
    while( length-=16 )
    {
        fprintf(fp,
                "%16lx:  "
                "%02x %02x %02x %02x  %02x %02x %02x %02x  "
                "%02x %02x %02x %02x  %02x %02x %02x %02x  "
                "%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c\n",
                address,
                s[0],s[1],s[2],s[3],s[4],s[5],s[6],s[7],
                s[8],s[9],s[10],s[11],s[12],s[13],s[14],s[15],
                toprint(s[0]),toprint(s[1]),toprint(s[2]),toprint(s[3]),
                toprint(s[4]),toprint(s[5]),toprint(s[6]),toprint(s[7]),
                toprint(s[8]),toprint(s[9]),toprint(s[10]),toprint(s[11]),
                toprint(s[12]),toprint(s[13]),toprint(s[14]),toprint(s[15]) );
        s+=16;
        address+=16;
    }
}

/*=======================================================
=======================================================*/
#if 0
#undef __PROC__
#define __PROC__ "PrintFdr"

extern void PrintFdr(FILE *fp,INDEX i)
{
    FDR *f = &AFDR(i);
                   
    fprintf(fp,
            "adr:        %016lx\n"
            "cbLineOff.: %016lx\n"
            "cbLine:     %016lx\n"
            "cbSs:       %016lx\n"
            "rss:        %08x\n"
            "issBase:    %08x\n"
            "isymBase:   %08x\n"    
            "csym:       %08x\n"
            "ilineBase:  %08x\n"
            "cline:      %08x\n"
            "ioptBase:   %08x\n"
            "copt:       %08x\n"
            "ipdFirst:   %08x\n"
            "cpd:        %08x\n"
            "iauxBase:   %08x\n"
            "caux:       %08x\n"
            "rfdBase:    %08x\n"
            "crfd:       %08x\n"
            "lang:       %1x\n"
            "fMerge:     %1x\n"
            "fReadin:    %1x\n"
            "fBigendian: %1x\n"
            "glevel:     %1x\n",
            f->adr,
            f->cbLineOffset,
            f->cbLine,
            f->cbSs,
            f->rss,
            f->issBase,
            f->isymBase,    
            f->csym,
            f->ilineBase,
            f->cline,
            f->ioptBase,
            f->copt,
            f->ipdFirst,
            f->cpd,
            f->iauxBase,
            f->caux,
            f->rfdBase,
            f->crfd,
            f->lang,
            f->fMerge,
            f->fReadin,
            f->fBigendian,
            f->glevel);
}
#endif 
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintInsNoNL"

void PrintInsNoNL(FILE *fp,INDEX ii)
{
    if( AINS_FLAGS(ii) & FI_EVAL_C)
        fprintf(fp,"%016lx ",AINS_C_EVAL(ii));
    else
        fprintf(fp,"%16s ","");
    
    fprintf(fp,"%p: ",(void *)Ins2Addr(ii));
    
    InstructionPrint(fp,ii,Ins2Addr(ii));
    fprintf(fp,"\t[%c%c%c%c%c] %6d",
            AINS_FLAGS(ii) & F_KILLED  ? 'K' : '_',
            AINS_FLAGS(ii) & FI_CHANGED? 'M' : '_',
            AINS_FLAGS(ii) & FI_EVAL_A ? 'A' : '_',
            AINS_FLAGS(ii) & FI_EVAL_B ? 'B' : '_',
            AINS_FLAGS(ii) & FI_EVAL_C ? 'C' : '_',
            ii);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintIns"

extern void PrintIns(FILE *fp,INDEX ii)
{
   PrintInsNoNL(fp, ii);
   fprintf(fp, "\n");
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintEdg"

extern void PrintEdg(FILE *fp,INDEX iedg)
{
    fprintf(fp,"EDG: %d [%c%c%c] HD: %d  TL: %d  PR: %d  SU: %d\n",
            iedg,
            AEDG_FLAGS(iedg) & F_KILLED     ? 'K' : '_',
            AEDG_IS_INTER(iedg)             ? 'I' : '_',
            AEDG_TYPE(iedg),
            AEDG_HEAD(iedg),AEDG_TAIL(iedg),AEDG_PRED(iedg),AEDG_SUCC(iedg));
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintEdgSuper"

extern void PrintEdgSuper(FILE *fp,INDEX iedg)
{
    PrintEdg(fp,iedg);
    PrintBbl(fp,AEDG_HEAD(iedg));
    PrintBbl(fp,AEDG_TAIL(iedg));
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintBblSuper"

extern void PrintBblSuper(FILE *fp,INDEX ibbl)
{
    INDEX iins;
    
    PrintBbl(stdout,ibbl);
    FOREACH_BBL_INS(ibbl,iins)
    {
        PrintIns(fp,iins);
    }
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintBbl"

extern void PrintBbl(FILE *fp,INDEX ibbl)
{
    INDEX edge;

    fprintf(fp,"BBL: %d [%c%c%c%c%c%c%c] RO: %016lx  RI: %016lx  FUN: %d [%s]\n",
            ibbl,
            ABBL_TYPE(ibbl),
            ABBL_FLAGS(ibbl) & F_KILLED       ? 'K' : '_',
            ABBL_FLAGS(ibbl) & FB_LOOP_HEADER ? 'L' : '_',
            ABBL_FLAGS(ibbl) & F_REFQUAD      ? 'Q' : '_',
            ABBL_FLAGS(ibbl) & F_REFQUAD_HARD ? 'H' : '_',
            ABBL_FLAGS(ibbl) & F_GPREL32      ? 'G' : '_',
            ABBL_FLAGS(ibbl) & FB_PROG_ENTRY  ? 'E' : '_',
            ABBL_OUT(ibbl),
            AbblRegsLiveIn(ibbl),
            ABBL_FUN(ibbl),
            AFUN_NAME(ABBL_FUN(ibbl)));

    fprintf(fp,"IF: %d  IL: %d  BP: %d  BN: %d  ET: %d  EF: %d  JT: %d  CNT: %ld\n",
            ABBL_INS_FIRST(ibbl),
            ABBL_INS_LAST(ibbl),
            ABBL_PREV(ibbl),
            ABBL_NEXT(ibbl),
            VALID(ABBL_EDG_TRUE(ibbl)) ? AEDG_TAIL(ABBL_EDG_TRUE(ibbl)) : -1,
            VALID(ABBL_EDG_FALSE(ibbl)) ? AEDG_TAIL(ABBL_EDG_FALSE(ibbl)) : -1,
            ABBL_JUMP_TAB(ibbl),
            ABBL_COUNT(ibbl));

    fprintf(fp,"DM: %5d  LOOP: %5d  SB: %5d  CD: %5d\n",
            ABBL_PRE_DOM(ibbl),
	    ABBL_LOOP_HEADER(ibbl),
	    ABBL_LOOP_NEXT_MEMBER(ibbl),
	    ABBL_LOOP_FIRST_MEMBER(ibbl));

    fprintf(fp,"PRED:");
    FOREACH_BBL_PRED(ibbl,edge)
    {
        fprintf(fp," %d",AEDG_HEAD(edge));
        if( AEDG_IS_INTER(edge) )
            printf("*");
    }
    fprintf(fp,"\n");

    fprintf(fp,"SUCC:");
    
    FOREACH_BBL_SUCC(ibbl,edge)
    {
        fprintf(fp," %d",AEDG_TAIL(edge));
        if( AEDG_IS_INTER(edge) )
            printf("*");
    }
    fprintf(fp,"\n");
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintStk"

extern void PrintStk(FILE *fp,INDEX istk)
{
    fprintf(fp,"\tOFF: %4d [%c%c%c%c] R(RNT): (%2d,%2d,%c)  W(RNT): (%2d,%2d,%c)\n",
            ASTK_OFFSET(istk),
            ASTK_FLAGS(istk) & FS_SAV_FIRST     ? 'S' : '_',
            ASTK_FLAGS(istk) & FS_SAV_NON_FIRST ? 'N' : '_',
            ASTK_FLAGS(istk) & FS_RES_LAST      ? 'R' : '_',
            ASTK_FLAGS(istk) & FS_RES_NON_LAST  ? 'N' : '_',

            ASTK_READ_REG(istk),
            ASTK_READ_NUM(istk),
            ASTK_READ_TYPE(istk),

            ASTK_WRITE_REG(istk),
            ASTK_WRITE_NUM(istk),
            ASTK_WRITE_TYPE(istk));
    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintStk"

extern void PrintFunStk(FILE *fp,INDEX ifun)
{
#if 0

    INDEX istk;

    FOREACH_FUN_STK(ifun,istk)
    {
        if( ASTK_READ_NUM(istk) == 0 &&
            ASTK_WRITE_NUM(istk)== 0 &&
            ASTK_FLAGS(istk) == 0)
            continue;
        
        PrintStk(fp,istk);
    }
#endif
        
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintFun"

extern void PrintFun(FILE *fp,INDEX ifun)
{
    INDEX ibbl;
    
    fprintf(fp,"FUN: %d [%c%c%c%c] SW: %2d  JM: %2d  STK: [%c%c%c%c] SM: %3d   S3m: %d NM: %s\n",
            ifun,
            AFUN_IS_KILLED(ifun)  ? 'K' : '_',
            AFUN_IS_PSEUDO(ifun)  ? 'P' : '_',
            AfunHasEsc2(ifun)     ? 'E' : '_',
            AfunHas2Esc(ifun)     ? 'I' : '_',
            AfunNJumps(ifun),
            AfunNSwitches(ifun),
            AFUN_HAS_STK_INVALID(ifun) ? 'I' : '_',
            AFUN_HAS_STK_ARG(ifun)     ? 'A' : '_',
            AFUN_HAS_STK_ALIAS(ifun)   ? 'L' : '_',
            AfunCallsStackArgFuns(ifun) ? 'C' : '_',
            AFUN_STK_MAX(ifun),
            AFUN_STK_MIN(ifun),
            AFUN_NAME(ifun));

    fprintf(fp,"SAV: %016lx USE: %016lx CHG: %016lx KLL: %016lx \n",
            AFUN_REGS_SAVED(ifun),AFUN_REGS_USED(ifun),AFUN_REGS_CHANGED(ifun),AFUN_REGS_THROUGH(ifun));
    
    fprintf(fp,"BLOCKS:");
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        fprintf(fp," %d",ibbl);
    }
    
    fprintf(fp,"\n");
}

/*=======================================================
=======================================================*/
/* Printing Loops                                      */
/*=======================================================
=======================================================*/

#define INDENT(fp,i) fprintf(fp,"%*s",i,"")

#define IND_INCR 4

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintLoopBBL"

static void PrintLoopBBL(FILE *fp, INDEX bbl, int ind)
{
   INDEX iins;
   int ind2;
   char *s;

   s = ABBL_IS_HOT(bbl) ? "*" : "";

   INDENT(fp,ind); fprintf(fp, "%%BBL{%d}%s{\n", bbl, s);
   
   ind2 = ind+IND_INCR;

   if (VALID(ABBL_COUNT(bbl)))
      {
	 INDENT(fp,ind2); fprintf(fp, "%%COUNT{%ld}\n", ABBL_COUNT(bbl));
      }

   FOREACH_BBL_INS(bbl,iins)
   {
      INDENT(fp,ind2);
      fprintf(fp,"%%I{");
      PrintInsNoNL(fp,iins);
      fprintf(fp,"}%%I\n");
   }
   INDENT(fp,ind); fprintf(fp, "} # End of BBL %d\n", bbl);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintFunBBLsNotInLoops"

static void PrintFunBBLsNotInLoops(FILE *fp, INDEX f, int ind)
{
   INDEX bbl;
   
   FOREACH_FUN_TOPBBL(f, bbl)
   {
      PrintLoopBBL(fp, bbl, ind);
   }
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintBBLSuccEdges"

static void PrintBBLSuccEdges(FILE *fp, INDEX bbl)
{
   INDEX succ_edge;

   FOREACH_BBL_SUCC(bbl, succ_edge)
   {
      if ( ! AEDG_IS_INTER(succ_edge) )
	if (AEDG_COUNT(succ_edge) > 0) {
	 fprintf(fp, "{%d -> %d : %ld} ", 
		 AEDG_FROM(succ_edge), 
		 AEDG_TO(succ_edge),
		 AEDG_COUNT(succ_edge));
	}
	else {
	 fprintf(fp, "{%d -> %d} ", AEDG_FROM(succ_edge), AEDG_TO(succ_edge) );
	}
   }
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintLoopEdges2"

static void PrintLoopEdges2(FILE *fp, INDEX loop_header)
{
   INDEX bbl;

   PrintBBLSuccEdges(fp, loop_header);

   ABBL_FOREACH_LOOP_MEMBER(loop_header, bbl)
   {
      PrintBBLSuccEdges(fp, bbl);

      if( ABBL_IS_LOOP_HEADER( bbl ) )
	 PrintLoopEdges2(fp, bbl);	 
   }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintLoopEdges"

static void PrintLoopEdges(FILE *fp, INDEX loop_header, int ind)
{
   INDENT(fp,ind); fprintf(fp, "%%EDGES{");
   
   PrintLoopEdges2(fp, loop_header);

   fprintf(fp, "}\n");
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintFunEdges"

static void PrintFunEdges(FILE *fp, INDEX f, int ind)
{
   INDEX bbl;

   INDENT(fp,ind); fprintf(fp, "%%EDGES{");
   
   FOREACH_FUN_BBL(f, bbl)
      PrintBBLSuccEdges(fp, bbl);

   fprintf(fp, "}\n");
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintLoopHeader"

static void PrintLoopHeader(FILE *fp, INDEX loop_header, int ind)
{
   INDENT(fp,ind); fprintf(fp, "%%LOOPHEADER{\n");

   PrintLoopBBL(fp, loop_header, ind+IND_INCR);

   INDENT(fp,ind); fprintf(fp, "}\n");
}
   

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintLoop"

extern void PrintLoop(FILE *fp, INDEX loop_header, int ind)
{
   INDEX memb;
   INDEX ind2;

   INDENT(fp,ind);
   fprintf(fp, "%%LOOP{%d}{\n", loop_header);

   ind2 = ind+IND_INCR;
   
   PrintLoopHeader(fp, loop_header, ind2);
    
   ABBL_FOREACH_LOOP_MEMBER(loop_header,memb)
   {
      if (ABBL_IS_LOOP_HEADER(memb))
	 PrintLoop(fp, memb, ind2);
      else
	 PrintLoopBBL(fp, memb, ind2);
   }

   PrintLoopEdges(fp, loop_header, ind2);  

   INDENT(fp,ind);
   fprintf(fp, "} # End of loop\n");
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintFunLoops2"


static void PrintFunLoops2(FILE *fp, INDEX fun, int ind)
{
   INDEX toploop;

   FOREACH_FUN_TOPLOOP(fun, toploop)
   {
      PrintLoop(fp, toploop, ind);
   }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintFunLoops"

extern void PrintFunLoops(FILE *fp, INDEX f, int ind)
{
   INDEX ibbl;
   INDEX iedg;
#if 0   
   if (AFUN_IS_PSEUDO(f)) return;
#endif
   CASSERT( (!AFUN_IS_KILLED(f)), ("Expected a live function.") );

   INDENT(fp,ind);
   fprintf(fp, "%%FUNCTION{%s}{%d}{\n",AFUN_NAME(f), f);


   /* printf callsite information */

   INDENT(fp,ind+IND_INCR);
   fprintf(fp, "%%CALLSITES{\n");
   ibbl = AFUN_BBL_FIRST(f);
   FOREACH_BBL_PRED(ibbl,iedg)
   {
       if( AEDG_TYPE(iedg) != ET_FUN_CALL && AEDG_TYPE(iedg) != ET_HELL ) continue;
       INDENT(fp,ind+2*IND_INCR);
       fprintf(fp,"%%CALLER{%s{%d},%d,%d,%ld}\n",
	       AFUN_NAME(ABBL_FUN(AEDG_HEAD(iedg))),
	       ABBL_FUN(AEDG_HEAD(iedg)),
	       AEDG_HEAD(iedg),
	       AEDG_TAIL(iedg),
	       ABBL_COUNT(AEDG_HEAD(iedg)));
   }
   INDENT(fp,ind+IND_INCR);
   fprintf(fp, "} # End of callsites\n");

   
   if( AfunIsLoopAnalyzable(f) )
   {
       INDENT(fp,ind+IND_INCR);
       fprintf(fp, "%%CHECKED{yes}\n");
       
       PrintFunBBLsNotInLoops(fp, f, ind+IND_INCR);
       
       PrintFunLoops2(fp, f, ind+IND_INCR);
	 
       PrintFunEdges(fp, f, ind+IND_INCR);
   }
   else
   {
       INDENT(fp,ind+IND_INCR);
       fprintf(fp, "%%CHECKED{no}\n");
       
       FOREACH_FUN_BBL(f, ibbl)
       {
           PrintLoopBBL(fp, ibbl, ind+IND_INCR);
       }

       PrintFunEdges(fp, f, ind+IND_INCR);
   }
   INDENT(fp,ind);
   fprintf(fp, "} # End of function %s\n\n", AFUN_NAME(f));
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintFunSuper"

extern void PrintFunSuper(FILE *fp,INDEX ifun)
{
    INDEX ibbl;

    PrintSep2(fp);
    PrintFun(fp,ifun);
    PrintSep2(fp);

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        PrintBblSuper(fp,ibbl);
        PrintSep2(fp);
    }
}
/*=======================================================
=======================================================*/

