/*=======================================================
=======================================================*/

#include <stdio.h>
#include <stdlib.h>

#define FILE_PDF 0x0a213e4644503c21L
#define FILE_ADR 0x0f0e0010L
#define FILE_CNT 0x0f0e0a11L
#define FILE_CNT2 0x0f0e0a33L



/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ComputeReadProfilingInformation"

#define USE_NEW_ATOM_PIXIE 1
typedef unsigned char  UINT8;
typedef   signed char  INT8;
typedef unsigned short UINT16;
typedef   signed short INT16;
typedef unsigned int   UINT32;
typedef   signed int   INT32;
typedef unsigned long  UINT64;
typedef   signed long  INT64;

typedef INT8  BOOL;
typedef INT32 INDEX;
typedef INT32 NUMBER;
typedef INT64 ADDRESS;
typedef char *STRING;
typedef UINT16 FLAGS;
typedef INT8  DATATYPE;
typedef INT8  EDGETYPE;
typedef UINT64 RESLIST;
typedef INT8  RES;
typedef INT32    LATENCY;
typedef UINT8   PIPELINES;
typedef UINT64 REGLIST;
#define REGSET REGLIST
typedef INT8  REG;



#define CONFIG_NINDIR       20
#define ALTO_MAGIC 0xcaffe

typedef struct {
    ADDRESS addr;
    ADDRESS count;
} indir;


typedef struct {
    INT64 iins;
    INDEX ins;
    REG reg1;
    REG reg2;
    INT64 totalcount;
    indir values[CONFIG_NINDIR];
} Indir;

typedef struct {
    indir values[2*CONFIG_NINDIR];
} doubleIndir;


Indir *aIndir;
NUMBER aIndirFuns;
Indir *aValues;
NUMBER aValueCount;

/*=======================================================
=======================================================*/

BOOL flag_important_only =0;
BOOL flag_floats = 0;
double cut_off_percentile = 10.0;
INT64 cut_off_count = 0;
/*=======================================================
=======================================================*/

static int cmp(const void *a, const void *b)
{
    indir *v1 = (indir *)a;
    indir *v2 = (indir *)b;

    return (v2->count - v1->count);
}

static int cmp2(const void *a, const void *b)
{
    indir *v1 = (indir *)a;
    indir *v2 = (indir *)b;

    return (v2->addr - v1->addr);
}


extern void PrintIndirs(indir *p, int num) 
{
    int i;

    printf("========================\n");
    for (i = 0; i< num; i++) {
	printf("val  %16lx  %16ld\n", p[i].addr, p[i].count);
    }
    printf("========================\n");

}


extern void mergeprofs(Indir p1, Indir p2, Indir p3) 
{
    doubleIndir dp;
    int i;
    int j;
    int k;

    qsort(p1.values, CONFIG_NINDIR-1, sizeof(indir), &cmp2);
    qsort(p2.values, CONFIG_NINDIR-1, sizeof(indir), &cmp2);



    /* now merge sort the two lists together */
    i = 0;
    j = 0;
    k = 0;

    while ( (j < CONFIG_NINDIR-1) && (i < CONFIG_NINDIR-1) ) {
	if (p1.values[i].addr < p2.values[j].addr) {
	    dp.values[k].addr = p1.values[i].addr;
	    dp.values[k].count = p1.values[i].count;
	    i++;
	    k++;
	}
	else if (p1.values[i].addr > p2.values[j].addr) {
	    dp.values[k].addr = p2.values[j].addr;
	    dp.values[k].count = p2.values[j].count;
	    j++;
	    k++;
	}
	else if (p1.values[i].addr == p2.values[j].addr) {
	    dp.values[k].addr = p1.values[i].addr;
	    dp.values[k].count = p1.values[i].count;
	    dp.values[k].count += p2.values[j].count;
	    j++;
	    k++;
	    i++;
	}
    }

    while (i < CONFIG_NINDIR-1) {
	dp.values[k].addr = p1.values[i].addr;
	dp.values[k].count = p1.values[i].count;
	k++;
	i++;
    }

    while (j < CONFIG_NINDIR-1) {
	dp.values[k].addr = p2.values[j].addr;
	dp.values[k].count = p2.values[j].count;
	k++;
	j++;
    }


    PrintIndirs(p1.values, CONFIG_NINDIR);
    PrintIndirs(p2.values, CONFIG_NINDIR);
    PrintIndirs(dp.values, k);



    /* then sort by counts */
    qsort(dp.values, k, sizeof(indir), &cmp);

    PrintIndirs(dp.values, k);

    
    /* sweep bottom half of the table */
    for (i = 1; i< CONFIG_NINDIR; i++) {
	dp.values[CONFIG_NINDIR].count += dp.values[CONFIG_NINDIR + i].count;
    }

    /* fill in the final array */

    for (i = 0; i < CONFIG_NINDIR; i++) {
	p3.values[i].addr = dp.values[i].addr;
	p3.values[i].count = dp.values[i].count;
    }

    p3.values[CONFIG_NINDIR-1].count += p1.values[CONFIG_NINDIR-1]. count;
    p3.values[CONFIG_NINDIR-1].count += p2.values[CONFIG_NINDIR-1]. count;

}


/*=======================================================
 we want to have the value profile create an ascii file
 so we can look at the values used 
=======================================================*/
#undef __PROC__
#define __PROC__ "ComReadValueProfilingInformation"


extern void ComReadValueProfilingInformation(FILE *fp, char *CONFIG_INFILE1, char *CONFIG_INFILE2, char *CONFIG_OUTFILE)
{
    FILE *fp_adr1;
    FILE *fp_adr2;
    FILE *fp_out;
    FILE *valprof;
    char fn_adr[256];
    INT64 cnt;
    INT64 bcount1 = 0, bcount2 = 0;
    INT64 ecount1 = 0, ecount2 = 0;
    INT64 fcount1 = 0, fcount2 = 0;
    INT64 mapcount1 = 0, mapcount2 = 0;
    INT64 totalcount;
    int i;
    int j;
    int temp;
    INT64 counter;
    INT64 counter1;
    INT64 counter2;
    INT64 iedg1;
    INT64 iedg2;
    Indir val1;
    Indir val2;
    Indir val3;

    /*
      open address file and advance file pointer
      to beginning of address data
      */
    
    fp_adr1 = fopen(CONFIG_INFILE1,"r");
    if( fp_adr1 == 0 )
    {
        
        fprintf(stderr,"Cannot open adr file: %s, no value profiles\n", CONFIG_INFILE1);
        return;
    }
    
    fread(&cnt,sizeof(cnt),1,fp_adr1);

    if ( cnt != ALTO_MAGIC ) {
	fprintf(stderr, "Not an alto profile [%s]\n", CONFIG_INFILE1);
	return;
    }


    fp_adr2 = fopen(CONFIG_INFILE2,"r");
    if( fp_adr2 == 0 )
    {
        
        fprintf(stderr,"Cannot open adr file: %s, no value profiles\n", CONFIG_INFILE2);
        return;
    }
    
    fread(&cnt,sizeof(cnt),1,fp_adr2);

    if ( cnt != ALTO_MAGIC ) {
	fprintf(stderr, "Not an alto profile [%s]\n", CONFIG_INFILE2);
	return;
    }


    /* read headers for first file */
    fread(&bcount1,sizeof(bcount1),1,fp_adr1);
    fread(&ecount1,sizeof(ecount1),1,fp_adr1);
    fread(&fcount1,sizeof(fcount1),1,fp_adr1);
    fread(&mapcount1,sizeof(mapcount1),1,fp_adr1);

    /* read headers for second file */ 
    fread(&bcount2,sizeof(bcount2),1,fp_adr2);
    fread(&ecount2,sizeof(ecount2),1,fp_adr2);
    fread(&fcount2,sizeof(fcount2),1,fp_adr2);
    fread(&mapcount2,sizeof(mapcount2),1,fp_adr2);

    if (bcount1 != bcount2) {
	fprintf(stderr, "Number of bbl counts doesn't match\n");
    }

    if (ecount1 != ecount2) {
	fprintf(stderr, "Number of edge counts doesn't match\n");
    }

    if (fcount1 != fcount2) {
	fprintf(stderr, "Number of value counts doesn't match\n");
    }

    if (mapcount1 != mapcount2) {
	fprintf(stderr, "Number of map entries doesn't match\n");
    }

    /* now we believe we have two matching profiles...*/

    fp_out = fopen(CONFIG_OUTFILE,"w");
    if( fp_out == 0 )
    {
        
        fprintf(stderr,"Cannot open adr file: %s, no value profiles\n", CONFIG_OUTFILE);
        return;
    }

    fwrite(&cnt, sizeof(cnt), 1, fp_out);
    fwrite(&bcount1, sizeof(bcount1), 1, fp_out);
    fwrite(&ecount1, sizeof(ecount1), 1, fp_out);
    fwrite(&fcount1, sizeof(fcount1), 1, fp_out);
    fwrite(&mapcount1, sizeof(mapcount1), 1, fp_out);


    for (i = 0; i < bcount1; i++ ) {
	fread(&counter1, sizeof(counter1), 1, fp_adr1);
	fread(&counter2, sizeof(counter2), 1, fp_adr2);

	counter1 += counter2;
	fwrite(&counter1, sizeof(counter1), 1, fp_out);
    }
	

    for (i = 0; i < ecount1; i++) {
	fread(&iedg1, sizeof(counter1), 1, fp_adr1);
	fread(&iedg2, sizeof(counter1), 1, fp_adr2);
	fread(&counter1, sizeof(counter2), 1, fp_adr1);
	fread(&counter2, sizeof(counter1), 1, fp_adr2);

	if (iedg1 != iedg2) {
	    fprintf(stderr,"ERROR, Edg labels off %ld %ld\n", iedg1, iedg2);
	    return;
	}

	fwrite(&iedg1, sizeof(counter1), 1, fp_out);
	counter1 += counter2;
	fwrite(&counter1, sizeof(counter1), 1, fp_out);
    }

    for (i = 0; i < fcount1; i++) {
	fread(&counter1, sizeof(counter1), 1, fp_adr1);
	fread(&counter2, sizeof(counter2), 1, fp_adr2);

	if (counter1 != counter2) {
	    fprintf(stderr, "ERROR: magic numbers don't match %ld %ld\n", counter1, counter2);
	    return;
	}
	printf("merging profiles for magic number %ld\n", counter1);

	fwrite(&counter1, sizeof(counter1), 1, fp_out);

	/* counter of time until sweep */
	fread(&counter1, sizeof(counter1), 1, fp_adr1);
	fread(&counter2, sizeof(counter2), 1, fp_adr2);
	fwrite(&counter1, sizeof(counter1), 1, fp_out);


	/* now read in the two profiles, and merge them */
	for (j=0; j < CONFIG_NINDIR; j++) {
	    fread(&counter, sizeof(counter), 1, fp_adr1);
	    val1.values[j].addr = counter;
	    fread(&counter, sizeof(counter), 1, fp_adr1);
	    val1.values[j].count = counter;
	}


	for (j=0; j < CONFIG_NINDIR; j++) {
	    fread(&counter, sizeof(counter), 1, fp_adr2);
	    val2.values[j].addr = counter;
	    fread(&counter, sizeof(counter), 1, fp_adr2);
	    val2.values[j].count = counter;
	}
	
	/* merge two value profiles */
	mergeprofs(val1, val2, val3);

	for (j=0; j < CONFIG_NINDIR; j++) {
	    counter = val3.values[j].addr;
	    fwrite(&counter, sizeof(counter), 1, fp_out);
	    counter = val3.values[j].count;
	    fwrite(&counter, sizeof(counter), 1, fp_out);

	}
    }
}

/*=======================================================
=======================================================*/

static void Usage()
{
    printf("usage: profadder src1 src2 dst\n");
}

/*=======================================================
=======================================================*/

extern int main(int argc, char *argv[])
{

    int i;
    
    while( EOF != (i = getopt(argc,argv,"ihc:")) )
    {
        switch(i)
        {
        case 'c':
            cut_off_count = atoi(optarg);
            break;
        case 'i':
            flag_important_only = 1;
            break;
        case 'f':
            flag_floats = 1;
            break;
        case 'h':
        case '?':
        default:
            Usage();
            break;
        }
    }

    argv += optind;     
    argc -=optind;

    ComReadValueProfilingInformation(stdout,argv[0], argv[1], argv[2]);
    return 0;
}


/*=======================================================
=======================================================*/





