/*=======================================================
  Author:   R. Muth
  Version:  $Id: stat.c,v 1.37 1999/07/22 21:19:40 muth Exp $
  Comment:  This file contains printing functions that
            require some computation
=======================================================*/

#include "global.h"
#include "functions.h"
#include "flags.h"
#include "regs.h"
#include "opcode.h"

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "BinaryBits"

extern NUMBER BinaryBits(INT64 n)
{
    NUMBER count = 0;
    for( ; n != 0; n &= n-1 )
        count++;
    
    return count;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "BinaryCondenseOctets"

extern NUMBER BinaryCondenseOctets(INT64 n)
{
    NUMBER result = 0;
    NUMBER i;

    for( i=0; i<8; i++ )
    {
        if( (n & 0xff) == 0 )
        {
            /* do nothing */
        }
        else if( (n & 0xff) == 0xff )
        {
            result |= (1<<i);
        }
        else
        {
            return -1;
        }
        n >>= 8;
    }

    return result;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "BinaryCondenseOctets"

extern NUMBER BinaryShiftedByte(UINT64 n)
{
    NUMBER i;

    for( i=0; i<63; i++ )
    {
        if( (n & (0xffL <<i)) == n )
        {
            return i;
        }
    }

    return INVALID;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "BinaryLog"

extern INT64 BinaryLog(INT64 n)
{
    INT64 t=0;
    
    if( n <= 0 ) return t;

    if( n >> 32 ) { n >>= 32; t += 32;}
    if( n >> 16 ) { n >>= 16; t += 16;}
    if( n >>  8 ) { n >>=  8; t +=  8;}
    if( n >>  4 ) { n >>=  4; t +=  4;}
    if( n >>  2 ) { n >>=  2; t +=  2;}
    if( n >>  1 ) { n >>=  1; t +=  1;}

    return t;
}

/*=======================================================
=======================================================*/
static int cmp(const void *x1,const void *x2)
{
    INT64 c1 = ABBL_COUNT(*(INDEX *)x1);
    INT64 c2 = ABBL_COUNT(*(INDEX *)x2);

    if( c1 < c2 )
        return 1;
    else if( c1 > c2 )
        return -1;
    else
        return 0;    
}



extern void ComputeCountCutOff(FILE *fp,INT64 MinCount,double MinWeightFrac, NUMBER MaxIns)
{
    BOOL verbose = 0;
    INDEX ibbl;

    INT64 sumw;
    INT64 sumi;
    INT64 sumb;
    INT64 totw;
    INT64 toti;
    INT64 totb;
    
    INT64 n;
    INDEX *array;
    

    if( !CONFIG_HAVE_BBL_PROFILE )
    {
        InfoStatus(fp,"\t[ComCutoff] no profile information\n");
        CONFIG_COUNT = 0;
        return;
    }


    array = malloc( sizeof(INDEX) * CONFIG_UBBL);

    totw = 0;
    toti = 0;
    totb = 0;
    
    
    FOREACH_BBL(ibbl)
    {
        array[ibbl] = ibbl;
        
        if( ABBL_IS_KILLED(ibbl) ) continue;
        totw += AbblWeight(ibbl);
        toti += AbblNInss(ibbl);
        totb ++;
        
    }
    
    qsort(array,CONFIG_UBBL,sizeof(INDEX),cmp);

    sumw = 1;
    sumi = 1;
    sumb = 0;
    
    CONFIG_COUNT = INVALID;

    n = 128;
    
    FOREACH_BBL(ibbl)
    {
        if( ABBL_IS_KILLED(array[ibbl]) ) continue;
        
        if( ABBL_COUNT(array[ibbl]) <= 0 ) break;

        sumw += AbblWeight(array[ibbl]);
        sumi += AbblNInss(array[ibbl]);
        sumb ++;

        if( sumi > MaxIns && (double) sumw / (double) totw > MinWeightFrac && CONFIG_COUNT == INVALID )
        {
            CONFIG_COUNT = ABBL_COUNT(array[ibbl]);
            if( CONFIG_COUNT < MinCount )
                CONFIG_COUNT = MinCount;
        }
        
        if( sumi > n )
        {
            n *= 2;
            if( verbose )
                fprintf(fp,"\t[&] %-7ld   bbl: %ld/%ld  ins: %ld/%ld (%.4f)   weight: %ldM/%ldM (%.4f)\n",
                        ABBL_COUNT(array[ibbl]),sumb,totb,sumi,toti,(double)sumi/(double)toti,
                        sumw>>20,totw>>20,(double)sumw/(double)totw);
        }
    }

    if( !VALID(CONFIG_COUNT) )
        CONFIG_COUNT = MinCount;
                     
    free(array);

    InfoStatus(fp,"\t[ComCutoff] cutoff %-7ld   min cnt: %ld  min wgt frc %f  max ins: %d \n",
               CONFIG_COUNT,MinCount,MinWeightFrac,MaxIns);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintStackStat"

extern void PrintStatStack(FILE *fp)
{
    INDEX ifun;


    NUMBER StackZero = 0;
    NUMBER StackBad = 0;
    NUMBER StackSave = 0;
    NUMBER StackReg = 0;
    INT64  CallsZero = 0;
    INT64  CallsBad = 0;
    INT64  CallsSave = 0;
    INT64  CallsReg = 0;
    INT64  SavesZero = 0;
    INT64  SavesBad = 0;
    INT64  SavesSave = 0;
    INT64  SavesReg = 0;
    
    
        
    FOREACH_LIVE_FUN(ifun)
    {
        INT64 calls=0;
        INT64 saves=0;
        INDEX ibbl = AFUN_BBL_FIRST(ifun);
        
        if( VALID(ibbl) )
        {
            calls = ABBL_COUNT(ibbl);
            saves = calls * BinaryBits(AFUN_REGS_SAVED(ifun));
        }
        
        if( AFUN_HAS_STK_INVALID(ifun) )
        {

            StackBad++;
            CallsBad += calls;
            SavesBad += saves;
            continue;
        }
        else if( AFUN_STK_MIN(ifun) == 0 )
        {
            StackZero++;
            CallsZero += calls;
            SavesZero += saves;
            continue;
        }

        if( BinaryBits(AFUN_REGS_SAVED(ifun)) * 8 == -AFUN_STK_MIN(ifun) )
        {
            StackReg++;
            CallsReg += calls;
            SavesReg += saves;
        }
            

        StackSave++;
        CallsSave += calls;
        SavesSave += saves;
            
    }

    
    fprintf(fp,
            "\tbad  n: %5d c: %6ld s: %8ld\n"
            "\tzero n: %5d c: %6ld s: %8ld\n"
            "\treg  n: %5d c: %6ld s: %8ld\n"
            "\tsave n: %5d c: %6ld s: %8ld\n",
            StackBad,CallsBad,SavesBad,
            StackZero,CallsZero,SavesZero,
            StackReg,CallsReg,SavesReg,
            StackSave,CallsSave,SavesSave);
}

/*=======================================================
  =======================================================*/
#undef __PROC__
#define __PROC__ "PrintProfile"

extern void PrintProfile(FILE *fp)
{
    NUMBER i;
#if 0
    INDEX ifun;
#endif
    INDEX ibbl;

    INT64 sum;
    INT64 weights[64];
    INT64 instrs[64];
    INT64 counts[64];
    INT64 totweights;
    INT64 totinstrs;
    INT64 totcounts;
    INT64 highcount;
        
    for( i=0; i < 64; i++)
    {
        counts[i] = 0;
        weights[i] = 0;
        instrs[i] = 0;
    }
    
    highcount = 0;
    totcounts = 1;
    totweights = 1;
    totinstrs = 1;
    
    FOREACH_LIVE_BBL(ibbl)
    {
        INT64 c = ABBL_COUNT(ibbl) < 0 ? 0 : ABBL_COUNT(ibbl);
        INT64 l = BinaryLog(c);
        INT64 i = AbblNInss(ibbl);
        INT64 w = c * i;

        weights[l] += w;
        totweights += w;
        instrs[l]  += i;
        totinstrs  += i;
        counts[l]  += 1;
        totcounts  += 1;
        if( ABBL_COUNT(ibbl) > highcount ) highcount = ABBL_COUNT(ibbl);
    }
    

    PrintSep3(stdout);
    fprintf(fp,"lb(count)  count  #bbl       sum(ins)     sum(count*ins)\n");
    PrintSep3(fp);
    
    for( i=0; i < 64; i++)
    {
        if( counts[i] == 0) continue;
        
        fprintf(fp,"%2d  %8d  %5ld (%2ld)  %6ld (%2ld)  %11ld  (%2ld)\n",
                i,(1<<i),
               counts[i],(counts[i]*100) /totcounts,
               instrs[i],(instrs[i]*100) /totinstrs,
               weights[i],(weights[i]*100) /totweights);
    }
    


    PrintSep3(stdout);
    printf("highest count: %ld\n",highcount);
    PrintSep3(stdout);

    sum = 0;

    for( i=63; i >=0; i--)
    {
        sum += weights[i];
        if( (sum * 100) / totweights > 90 )
            break;
    }
#if 0
    printf("Threshold: %d\n",i);
    
    FOREACH_LIVE_FUN(ifun)
    {
        BOOL flag = 0;
        FOREACH_FUN_BBL(ifun,ibbl)
        {
            if( BinaryLog(ABBL_COUNT(ibbl)) < i ) continue;

            if( flag == 0 )
            {
                PrintFun(stdout,ifun);
                flag = 1;
            }
            
            fprintf(fp,"\tbbl: %5d  count: %5ld  ins: %5d  weight: %11ld\n",
               ibbl,ABBL_COUNT(ibbl),AbblNInss(ibbl),ABBL_COUNT(ibbl)*AbblNInss(ibbl));
        }
        
    }
#endif
}

/*=======================================================
=======================================================*/



#define IT_ICM  15    /* cmovlbc,cmoveq,cmovlt,cmovle,cmovlbs,cmovne,cmovge,cmovgt */
#define IT_FCM  16    /* fcmoveq,fcmovlt,fcmovle,fcmovne,fcmovge,fcmovgt */

#define IT_ICP  17    /* cmpbge,cmpult,cmpeq,cmpule,cmplt,cmple */
#define IT_FCP  18    /* cmpt?? cmpg?? */

#define IT_FM4  19    /* muls?,mulf? */
#define IT_FM8  20    /* mult?,mulg? */
#define IT_FD4  21    /* divs?,divf? */
#define IT_FD8  22    /* divt?,divg? */

#define IT_F2I  23    /* ftois,ftoit */ 
#define IT_I2F  24    /* itofs,itoff,itoft */

         
#define IT_LOG  25    /* and,bic,bis,ornot,xor,eqv */
#define IT_SFT  26    /* msk??,ins??,ext??,zap??,s?? */
                        
#define IT_MLL  27    /* mull,mull/v */
#define IT_MLQ  28    /* mulq,mulq/v */
#define IT_MLH  29    /* umulh */

#define IT_IOP  30    /* rest */
#define IT_FOP  31    /* rest */

#define IT_RX   32    /* rc,rs */

#define IT_FCY  33    /* cpys */

#define IT_MEM  34    /* mb,wmb,fetch,fetch_m */

#define IT_EXT  35    /* sextb,sextw */

#define IT_PAL  36    /* call_pal */

#define IT_F2R  37    /* mt_fpcr */

#define IT_R2F  38    /* mf_fpcr */

#define IT_ERR  64    /* error condition */


#define IT_IBR  11    /* blbc,beq,blt,ble,blbs,bne,bge,bgt */
#define IT_FBR  12    /* fbeq,fblt,fble,fbne,fbge,fbgt */
#define IT_BRA  13    /* bra,bsr */
#define IT_JMP  14    /* jmp,jsr,ret,jsr_co */

static char TableFormat[][64] =
{
    "???",    /* flag error */
    "MSC",    /* misc instructions */
    "LDA",    /* load address */
    "ILD",    /* integer load */
    "FLD",    /* float load */
    "ULD",    /* integer load unaligned */
    "LLD",    /* integer load locked */

    "IST",    /* integer store */
    "FST",    /* float store */
    "UST",    /* integer store unaligned */
    "CLD",    /* integer store locked */

    "IBR",    /* integer cond branch */
    "FBR",    /* float cond branch */
    "BRA",    /* uncoditional branch: br bsr */
    "JMP",    /* jump */

    "ICM",    /* cmovlbc,cmoveq,cmovlt,cmovle,cmovlbs,cmovne,cmovge,cmovgt */
    "FCM",    /* fcmoveq,fcmovlt,fcmovle,fcmovne,fcmovge,fcmovgt */

    "ICP",    /* cmpbge,cmpult,cmpeq,cmpule,cmplt,cmple */
    "FCP",    /* cmpt?? cmpg?? */

    "FM4",    /* muls?,mulf? */
    "FM8",    /* mult?,mulg? */
    "FD4",    /* divs?,divf? */
    "FD8",    /* divt?,divg? */

    "F2I",    /* ftois,ftoit */ 
    "I2F",    /* itofs,itoff,itoft */

         
    "LOG",    /* and,bic,bis,ornot,xor,eqv */
    "SFT",    /* msk??,ins??,ext??,zap??,s?? */
                        
    "MLL",    /* mull,mull/v */
    "MLQ",    /* mulq,mulq/v */
    "MLH",    /* umulh */

    "IOP",    /* rest */
    "FOP",    /* rest */

    "RX",    /* rc,rs */

    "FCY",    /* cpys */

    "MEM",    /* mb,wmb,fetch,fetch_m */

    "EXT",    /* sextb,sextw */

    "PAL",    /* call_pal */

    "F2R",    /* mt_fpcr */

    "R2F",    /* mf_fpcr */

    "XXX",
    "XXX",
    "XXX",
    "XXX"
};

/*=======================================================
=======================================================*/
#undef __PROC__
 #define __PROC__ "PrintStatBinary"

#define THE_MAX 4096

extern void PrintStatBinary(FILE *fp)
{
    INT64 count[THE_MAX];
    INT64 weight[THE_MAX];
    
    NUMBER i;
    INDEX  index;
    

    /* count funs */

    memset( count,0,sizeof(count));
    memset( weight,0,sizeof(weight));
    
    i = 0;
    FOREACH_LIVE_FUN(index)
    {
        NUMBER n = AfunNBbls(index);
        if( n >= THE_MAX ) n = THE_MAX -1;
        count[n]++;
        i++;
    }

    printf("\tFuns: %d\n",i);
    
    for(i=0; i<THE_MAX; i++)
    {
        if( count[i] > 0 )
            fprintf(fp,"\t\t%6d bbls: %6ld\n",i,count[i]);
    }
    
    /* count bbls */
    
    memset( count,0,sizeof(count));
    memset( weight,0,sizeof(weight));
    
    FOREACH_LIVE_BBL(index)
    {
        NUMBER n = AbblNInss(index);
        if( n >= THE_MAX ) n = THE_MAX -1;
        count[n]++;
        i++;
    }
    
    fprintf(fp,"\tBbls: %d\n",i);

    for(i=0; i<THE_MAX; i++)
    {
        if( count[i] > 0 )
            fprintf(fp,"\t\t%6d inss:  %6ld\n",i,count[i]);
    }
    
    /* count edgs */

    memset( count,0,sizeof(count));
    memset( weight,0,sizeof(weight));
    
    i = 0;
    FOREACH_LIVE_EDG(index)
    {
        i++;
    }
    
    fprintf(fp,"\tEdgs: %d\n",i);

    
    /* count inss */

    memset( count,0,sizeof(count));
    memset( weight,0,sizeof(weight));

    i=0;
    FOREACH_LIVE_INS(index)
    {
        if( AINS_TYPE(index) == IT_MSC )
            PrintIns(stdout,index);
        
        count[ AINS_TYPE(index) ]++;
        weight[ AINS_TYPE(index) ]+= ABBL_COUNT(AINS_BBL(index));
        i++;
    }

    fprintf(fp,"\tInss: %d\n",i);

    for(i=0; i<THE_MAX; i++)
    {
        if( count[i] > 0 )
            fprintf(fp,"\t\t%s:  %6ld  %10ld\n",TableFormat[i],count[i],weight[i]);
    }
    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "InfStatResourceAllCount"

static void InfStatResourceAllCount(FILE *fp)
{
    INDEX  index;
    NUMBER icount = 0;
    NUMBER icounth = 0;
    NUMBER bcount = 0;
    NUMBER bcounth = 0;
    NUMBER ecount = 0;
    NUMBER fcount = 0;
    NUMBER fcounth = 0;

    INT64 weight = 0;
    INT64 weighth = 0;
    
    FOREACH_LIVE_FUN(index)
    {
        AFUN_FLAGS(index) &= ~F_MARKED;
    }
        
    FOREACH_LIVE_BBL(index)
    {
        bcount++;
        weight += AbblNInss(index)*ABBL_COUNT(index);
        if( ABBL_IS_HOT(index) )
        {
            AFUN_FLAGS(ABBL_FUN(index)) |= F_MARKED;
            weighth += AbblNInss(index)*ABBL_COUNT(index);
            bcounth++;
        }
    }
    
    FOREACH_LIVE_FUN(index)
    {
        fcount++;        
        if( AFUN_FLAGS(index) & F_MARKED)
            fcounth++;
    }
    
    FOREACH_LIVE_EDG(index)
    {
        ecount++;
    }
    
    FOREACH_LIVE_INS(index)
    {
        icount++;
        if( ABBL_IS_HOT(AINS_BBL(index)) )
            icounth++;
    }

    InfoStatus(fp,"\t[InfResource] w %ldM/%ldM (%.3f)  i %d:%d/%dk/%dk  b %d:%d/%dk/%dk  e %d/%dk/%dk  f %d/%d/%d/%d\n",
            weighth>>20,weight>>20,(double) weighth/(double)weight,
            icounth,icount,CONFIG_UINS>>10,CONFIG_NINS>>10,
            bcounth,bcount,CONFIG_UBBL>>10,CONFIG_NBBL>>10,
            ecount,CONFIG_UEDG>>10,CONFIG_NEDG>>10,
            fcounth,fcount,CONFIG_UFUN,CONFIG_NFUN);

}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "InfStatResourceFunStack"

static void InfStatResourceFunStack(FILE *fp)
{
    NUMBER StackZero = 0;
    NUMBER StackArg = 0;
    NUMBER StackAlias = 0;
    NUMBER StackBad = 0;
    NUMBER CallsStackArgFun = 0;
    NUMBER HasEsc2 = 0;
    
    NUMBER StackZeroH = 0;
    NUMBER StackArgH = 0;
    NUMBER StackAliasH = 0;
    NUMBER StackBadH = 0;
    NUMBER CallsStackArgFunH = 0;
    NUMBER HasEsc2H = 0;
   
    INDEX  index;

    FOREACH_LIVE_FUN(index)
    {
        AFUN_FLAGS(index) &= ~F_MARKED;
    }
        
    FOREACH_LIVE_BBL(index)
    {
        if( ABBL_IS_HOT(index) )  AFUN_FLAGS(ABBL_FUN(index)) |= F_MARKED;
    }
    

    FOREACH_LIVE_FUN(index)
    {
        if( AFUN_HAS_STK_INVALID(index) )
            StackBad++;
        if( AFUN_HAS_STK_ARG(index) )
            StackArg++;
        if( AFUN_HAS_STK_ALIAS(index) )
            StackAlias++;
        if( AFUN_STK_MIN(index) == 0 )
            StackZero++;
        if( AfunHasEsc2(index) )
            HasEsc2++;
        if( AfunCallsStackArgFuns(index) )
            CallsStackArgFun++;
        
        
        if( AFUN_FLAGS(index) & F_MARKED)
        {
            if( AFUN_HAS_STK_INVALID(index) )
                StackBadH++;
            if( AFUN_HAS_STK_ARG(index) )
                StackArgH++;
            if( AFUN_HAS_STK_ALIAS(index) )
                StackAliasH++;
            if( AFUN_STK_MIN(index) == 0 )
                StackZeroH++;
            if( AfunHasEsc2(index) )
                HasEsc2H++;
            if( AfunCallsStackArgFuns(index) )
                CallsStackArgFunH++;
        }
    }
    
    InfoStatus(fp,"\t[InfStack] bad %d#%d  arg %d#%d  alias %d#%d  zero %d#%d  esc2 %d#%d  cstf %d#%d\n",
            StackBad,StackBadH,
            StackArg,StackArgH,
            StackAlias,StackAliasH,
            StackZero,StackZeroH,
            HasEsc2,HasEsc2H,
            CallsStackArgFun,CallsStackArgFunH);

}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsIsEvaluable"

static BOOL AinsIsEvaluable(INDEX iins)
{
    if( AINS_REG_C(iins) == REG_ZERO || AINS_REG_C(iins) == REG_FZERO )
        return 0;
    if( AINS_TYPE(iins) == IT_BRA || AINS_TYPE(iins) == IT_JMP )
        return 0;
    return 1;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "InfStatResourceInsEval"

extern void InfStatResourceInsEval(FILE *fp)
{
    NUMBER Int = 0;
    NUMBER Flt = 0;
    NUMBER Tot = 0;
    NUMBER IntEvaluable = 0;
    NUMBER FltEvaluable = 0;
    NUMBER TotEvaluable = 0;
    NUMBER IntEvaluated = 0;
    NUMBER FltEvaluated = 0;
    NUMBER TotEvaluated = 0;
    INDEX  index;

    FOREACH_LIVE_INS(index)
    {
        if( AINS_REG_C(index) >=0 && AINS_REG_C(index) <= 31 )
        {
            Int++;
            Tot++;

            if( AinsIsEvaluable(index) )
            {
                IntEvaluable++;
                TotEvaluable++;
                if( AINS_FLAGS(index) & FI_EVAL_C )
                {
                    IntEvaluated++;
                    TotEvaluated++;
                }
                
            }
        }
        else if( AINS_REG_C(index) >=32 && AINS_REG_C(index) <= 63 )
        {
            Flt++;
            Tot++;

            if( AinsIsEvaluable(index) )
            {
                FltEvaluable++;
                TotEvaluable++;
                if( AINS_FLAGS(index) & FI_EVAL_C )
                {
                    FltEvaluated++;
                    TotEvaluated++;
                }
            }
        }
        else
        {
            ERROR(("weird\n"));
        }
        
    }
        
    InfoStatus(fp,"\t[InfInsEval] Int: %d (%5.3f)/%d (%5.3f)/%d  Flt: %d (%5.3f)/%d (%5.3f)/%d  Tot: %d (%5.3f)/%d (%5.3f)/%d \n",
               IntEvaluated, (double)IntEvaluated/(double)IntEvaluable, IntEvaluable, (double)IntEvaluable/(double)Int, Int,
               FltEvaluated, (double)FltEvaluated/(double)FltEvaluable, FltEvaluable, (double)FltEvaluable/(double)Flt, Flt,
               TotEvaluated, (double)TotEvaluated/(double)TotEvaluable, TotEvaluable, (double)TotEvaluable/(double)Tot, Tot);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "InfStatMove"

extern void InfStatMove(FILE *fp)
{
    INDEX iins;

    INT64 nmove_lda = 0;
    INT64 nmove_addl = 0;
    INT64 nmove_zap = 0;
    INT64 nmove_zapnot = 0;
    INT64 nmove_cpys = 0;

    INT64 weight_move_lda = 0;
    INT64 weight_move_addl = 0;
    INT64 weight_move_zap = 0;
    INT64 weight_move_zapnot = 0;
    INT64 weight_move_cpys = 0;
    
    FOREACH_LIVE_INS(iins)
    {
        if( AINS_CODE(iins) == I_lda )
        {
            if( AINS_DATA(iins) == 0 )
            {
                nmove_lda++;
                weight_move_lda += ABBL_COUNT(AINS_BBL(iins));
            }
            
        }
        else if( AINS_CODE(iins) == I_addl )
        {
            if( AINS_REG_B(iins) == REG_ZERO || AINS_REG_A(iins) == REG_ZERO )
            {
                nmove_addl++;
                weight_move_addl += ABBL_COUNT(AINS_BBL(iins));
            }
        }
        else if( AINS_CODE(iins) == I_zap )
        {
            if( AINS_REG_B(iins) == REG_LIT && AINS_DATA(iins) == 0xf0 )
            {
                nmove_zap++;
                weight_move_zap += ABBL_COUNT(AINS_BBL(iins));
            }
        }
        else if( AINS_CODE(iins) == I_zapnot )
        {
            if( AINS_REG_B(iins) == REG_LIT && AINS_DATA(iins) == 0xf )
            {
                nmove_zapnot++;
                weight_move_zapnot += ABBL_COUNT(AINS_BBL(iins));
            }
        }
        else if( AINS_CODE(iins) == I_cpys )
        {
            if( AINS_REG_A(iins) == AINS_REG_B(iins) )
            {
                nmove_cpys++;
                weight_move_cpys += ABBL_COUNT(AINS_BBL(iins));
            }
        }
    }

    InfoStatus(fp,"\t[InfMove] lda: %ld (w: %ld) addl: %ld (w: %ld) zap: %ld (w: %ld) zapnot: %ld (w: %ld) cpys: %ld (w: %ld)\n",
               nmove_lda,weight_move_lda,
               nmove_addl,weight_move_addl,
               nmove_zap,weight_move_zap,
               nmove_zapnot,weight_move_zapnot,
               nmove_zapnot,weight_move_cpys);
    
}
/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "InfStatResource"

extern void InfStatResource(FILE *fp)
{
    InfStatResourceAllCount(fp);
    InfStatResourceFunStack(fp);
    InfStatResourceInsEval(fp);
    
}
/*=======================================================
=======================================================*/
   
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "InfStatEdges"

extern void InfStatEdges(FILE *fp)
{
    NUMBER i;
    INDEX  iedg;
    NUMBER Count[256];
    INT64  Weight[256];

    for(i=0;i<256;i++)
    {
        Count[i] = 0;
        Weight[i] = 0;
    }
    
    FOREACH_LIVE_EDG(iedg)
    {
        Count[ AEDG_TYPE(iedg) ]++;
        Weight[ AEDG_TYPE(iedg) ] += AEDG_COUNT(iedg);
    }
    
    InfoStatus(fp,"\t[InfEdges] call %d (%ldk)  true %d (%ldk)  false %d (%ldk)  norm %d (%ldk)  uncond %d (%ldk)\n",
               Count[ET_FUN_CALL],Weight[ET_FUN_CALL]/1024,
               Count[ET_TRUE],Weight[ET_TRUE]/1024,
               Count[ET_FALSE],Weight[ET_FALSE]/1024,
               Count[ET_NORM],Weight[ET_NORM]/1024,
               Count[ET_UNCOND],Weight[ET_UNCOND]/1024);
}

/*=======================================================
=======================================================*/
