#include <filehdr.h>
#include <syms.h>
#include <scnhdr.h>
#include <aouthdr.h>
#include <reloc.h>

/*=======================================================
=======================================================*/

#define STRUCT_REL struct reloc

/*=======================================================
=======================================================*/

typedef struct
{
    SCNHDR scnhdr;
    union
    {
        INT8 *p8;
        INT16 *p16;
        INT32 *p32;
        INT64 *p64;
    }Data;
    INT32 nRelocs;
    RELOC *Reloc;
}STRUCT_SEC;


/*=======================================================
=======================================================*/

typedef struct
{
    FILHDR  filhdr;
    AOUTHDR aouthdr;
    HDRR    hdrr;
    LINER      *lnt;
    DNR        *dnt;
    SYMR       *lst;
    char       *lss;
    EXTR       *est;
    char       *ess;
    FDR        *fdt;

#if 0
    PDR        *pdt;
    NUMBER     new_text_sec_size;
#endif    
   

    UINT8   SectionMap[16];
}STRUCT_BIN;

/*=======================================================
=======================================================*/

extern STRUCT_BIN    theBinary;
extern STRUCT_SEC    aSec[];

/*=======================================================
=======================================================*/

#define BIN_FILHDR            theBinary.filhdr
#define BIN_AOUTHDR           theBinary.aouthdr

/*
#define BIN_NEW_TEXT_SEC_SIZE theBinary.new_text_sec_size
*/

#define BIN_HDRR              theBinary.hdrr
#define BIN_NSECTIONS         theBinary.filhdr.f_nscns
#define BIN_ENTRY             theBinary.aouthdr.entry
#define BIN_GP                theBinary.aouthdr.gp_value
#define BIN_SECTION_MAP       theBinary.SectionMap


#define BIN_SYMBOL_NPDT       theBinary.hdrr.ipdMax
#define BIN_SYMBOL_NLST       theBinary.hdrr.isymMax
#define BIN_SYMBOL_NLSS       theBinary.hdrr.issMax
#define BIN_SYMBOL_NFDT       theBinary.hdrr.ifdMax
#define BIN_SYMBOL_NEST       theBinary.hdrr.iextMax
#define BIN_SYMBOL_NESS       theBinary.hdrr.issExtMax

#define BIN_SYMBOL_PDT        theBinary.pdt
#define BIN_SYMBOL_LST        theBinary.lst
#define BIN_SYMBOL_LSS        theBinary.lss
#define BIN_SYMBOL_FDT        theBinary.fdt
#define BIN_SYMBOL_ESS        theBinary.ess
#define BIN_SYMBOL_EST        theBinary.est

/*=======================================================
=======================================================*/

#define IS_STYP_REG(flags)      ((flags & 0xf) == STYP_REG)
#define IS_STYP_DSECT(flags)    ((flags & 0xf) == STYP_DSECT)
#define IS_STYP_NOLOAD(flags)   ((flags & 0xf) == STYP_NOLOAD)
#define IS_STYP_GROUP(flags)    ((flags & 0xf) == STYP_GROUP)
#define IS_STYP_PAD(flags)      ((flags & 0xf) == STYP_PAD)

#define IS_STYP_COPY(flags)     (flags & STYP_COPY)
#define IS_STYP_TEXT(flags)     (flags & STYP_TEXT)
#define IS_STYP_DATA(flags)     (flags & STYP_DATA)
#define IS_STYP_BSS(flags)      (flags & STYP_BSS)

#define IS_STYP_RDATA(flags)    (flags & STYP_RDATA)
#define IS_STYP_SDATA(flags)    (flags & STYP_SDATA)
#define IS_STYP_SBSS(flags)     (flags & STYP_SBSS)
#define IS_STYP_UCODE(flags)    (flags & STYP_UCODE)

#define IS_STYP_GOT(flags)      (flags & STYP_GOT)
#define IS_STYP_DYNAMIC(flags)  (flags & STYP_DYNAMIC)
#define IS_STYP_DYNSYM(flags)   (flags & STYP_DYNSYM)
#define IS_STYP_REL_DYN(flags)  (flags & STYP_REL_DYN)
#define IS_STYP_DYNSTR(flags)   (flags & STYP_DYNSTR)
#define IS_STYP_HASH(flags)     (flags & STYP_HASH)

#define IS_STYP_EXTENDESC(flags) (flags & STYP_EXTENDESC)

#define IS_STYP_CONFLICT(flags) (!IS_STYP_EXTENDESC(flags) && (flags & STYP_CONFLICT))
#define IS_STYP_REGINFO(flags)  (!IS_STYP_EXTENDESC(flags) && (flags & STYP_REGINFO))

#define IS_STYP_LITA(flags)     (flags & STYP_LITA)
#define IS_STYP_LIT8(flags)     (flags & STYP_LIT8)
#define IS_STYP_LIT4(flags)     (flags & STYP_LIT4)
#define IS_S_NRELOC_OVFL(flags) (flags & S_NRELOC_OVFL)     
#define IS_STYP_LIB(flags)      (flags & STYP_LIB)
#define IS_STYP_INIT(flags)     (flags & STYP_INIT)
#define IS_STYP_FINI(flags)     (flags & STYP_FINI)

#define IS_STYP_COMMENT(flags)  ((flags & STYP_EXTMASK) == STYP_COMMENT)
#define IS_STYP_RCONST(flags)   ((flags & STYP_EXTMASK) == STYP_RCONST)
#define IS_STYP_XDATA(flags)    ((flags & STYP_EXTMASK) == STYP_XDATA)
#define IS_STYP_PDATA(flags)    ((flags & STYP_EXTMASK) == STYP_PDATA)

/*=======================================================
=======================================================*/

#define APDR_ADDRESS(ip) Global.theBinary.pdt[ip].adr

/*=======================================================
=======================================================*/

#define ALSY(is)         BIN_SYMBOL_LST[is]
#define ALSY_VALUE(i)    ALSY(i).value
#define ALSY_NAME(i)     (BIN_SYMBOL_LSS + ALSY(i).iss)
#define ALSY_TYPE(i)     ALSY(i).st
#define ALSY_IS_PROC(i)  (ALSY_TYPE(i)==stProc || ALSY_TYPE(i)==stStaticProc)
#define FOREACH_LSY(i)  for(i=0; i < BIN_SYMBOL_NLST; i++ )

#define AESY(is)         BIN_SYMBOL_EST[is].asym
#define AESY_VALUE(i)    AESY(i).value
#define AESY_NAME(i)     (BIN_SYMBOL_ESS + AESY(i).iss)
#define AESY_TYPE(i)     AESY(i).st
#define AESY_IS_PROC(i)  (AESY_TYPE(i)==stProc || AESY_TYPE(i)==stStaticProc)
#define FOREACH_ESY(i)   for(i=0; i < BIN_SYMBOL_NEST; i++ )

/*=======================================================
=======================================================*/

#define APDR(ip)        theBinary.pdt[ip]
#define AFDR(if)        theBinary.fdt[if]

/*=======================================================
=======================================================*/
#define ASEC(i)         aSec[i]

#define ASEC_SCNHDR(i)  ASEC(i).scnhdr
#define ASEC_SIZE(i)    ASEC(i).scnhdr.s_size
#define ASEC_START(i)   ASEC(i).scnhdr.s_vaddr
#define ASEC_STARTP(i)   ASEC(i).scnhdr.s_paddr
#define ASEC_NRELOCS(i) ASEC(i).nRelocs
#define ASEC_RELOC(i)   ASEC(i).Reloc
#define ASEC_DATA8(i)   ASEC(i).Data.p8
#define ASEC_DATA16(i)  ASEC(i).Data.p16
#define ASEC_DATA32(i)  ASEC(i).Data.p32
#define ASEC_DATA64(i)  ASEC(i).Data.p64
#define ASEC_NAME(i)    ASEC(i).scnhdr.s_name 
#define ASEC_FLAGS(i)   ASEC(i).scnhdr.s_flags
     
#define ASEC_IS_TEXT(i)  IS_STYP_TEXT(ASEC_FLAGS(i))
#define ASEC_IS_RDATA(i) IS_STYP_RDATA(ASEC_FLAGS(i))
#define ASEC_IS_DATA(i)  IS_STYP_DATA(ASEC_FLAGS(i))
#define ASEC_IS_SDATA(i) IS_STYP_SDATA(ASEC_FLAGS(i))
#define ASEC_IS_BSS(i)   IS_STYP_BSS(ASEC_FLAGS(i))
#define ASEC_IS_SBSS(i)  IS_STYP_SBSS(ASEC_FLAGS(i))
#define ASEC_IS_INIT(i)  IS_STYP_INIT(ASEC_FLAGS(i))
#define ASEC_IS_LIT8(i)  IS_STYP_LIT8(ASEC_FLAGS(i))
#define ASEC_IS_LIT4(i)  IS_STYP_LIT4(ASEC_FLAGS(i))
#define ASEC_IS_XDATA(i) IS_STYP_XDATA(ASEC_FLAGS(i))
#define ASEC_IS_PDATA(i) IS_STYP_PDATA(ASEC_FLAGS(i))
#define ASEC_IS_FINI(i)  IS_STYP_FINI(ASEC_FLAGS(i))
#define ASEC_IS_LITA(i)  IS_STYP_LITA(ASEC_FLAGS(i))
#define ASEC_IS_RCONST(i) IS_STYP_RCONST(ASEC_FLAGS(i))
#define ASEC_HAS_RELOC_OVFL(i) IS_S_NRELOC_OVFL(ASEC_FLAGS(i))

#define ASEC_AREL(is,ir)         ASEC(is).Reloc[ir]
#define ASEC_AREL_EXTERN(is,ir)  ASEC_AREL(is,ir).r_extern
#define ASEC_AREL_SECTION(is,ir) (ASEC_AREL(is,ir).r_symndx & 0xf)
#define ASEC_AREL_TYPE(is,ir)    ASEC_AREL(is,ir).r_type
#define ASEC_AREL_ADDRESS(is,ir) ASEC_AREL(is,ir).r_vaddr
#define ASEC_AREL_SYMBOL(is,ir)  ASEC_AREL(is,ir).r_symndx
#define ASEC_AREL_OFFSET(is,ir)  ASEC_AREL(is,ir).r_offset
#define ASEC_AREL_SIZE(is,ir)    ASEC_AREL(is,ir).r_size


