/*=======================================================
  Author:   R. Muth
  Version:  $Id: system.elf.c,v 1.2 1998/11/17 23:33:39 saw Exp $
  Comment:  This file contains platform specific functions
=======================================================*/

#include "global.h"
#include "functions.h"
#include "system.elf.h"
#include "flags.h"
#include "opcode.h"
#include "regs.h"
#include "error.h"


/*=======================================================
=======================================================*/

extern INT64 SysGetGlobalPointer()
{

    return BIN_GP;
}

extern void SysSetGlobalPointer(ADDRESS addr)
{
    BIN_GP = addr;
}

/* Explicit assumption:  3 gp add instructions to set GP,
   and they happen right after the 1st instruction */

#define __PROC__ "SysComputeGlobalPointer"

extern void SysComputeGlobalPointer()
{
    INDEX iins;
    ADDRESS addr;

    printf("computing global Pointer\n");
    iins = Addr2Ins(SysGetProgramEntry());
    printf("entry instruction is %d\n", iins);
    iins++;

    CFATAL( strcmp((char *)AINS_NAME(iins),"br") , ("Second instruction not a br, aborting\n"));
    CFATAL( AINS_REG_C(iins) != 29, ("Second instrucion doesn't store into GP\n"));
    CFATAL( AINS_DATA(iins) != 0, ("second instruction doesn't branch to 3rd\n"));
    
    iins++;

    addr = Ins2Addr(iins);
    CFATAL( strcmp((char *)AINS_NAME(iins),"ldah") , ("Third instruction not a ldah, aborting\n"));
    CFATAL( AINS_REG_B(iins) != 29, ("Third instrucion doesn't add from GP\n"));
    CFATAL( AINS_REG_C(iins) != 29, ("Third instrucion doesn't store into GP\n"));
    addr += AINS_DATA(iins) << 16;

    iins++;
    CFATAL( strcmp((char *)AINS_NAME(iins),"lda") , ("Fourth instruction not a lda, aborting\n"));
    CFATAL( AINS_REG_B(iins) != 29, ("Fourth instrucion doesn't add from GP\n"));
    CFATAL( AINS_REG_C(iins) != 29, ("Fourth instrucion doesn't store into GP\n"));
    addr += AINS_DATA(iins);


    printf("computed address for gp of %lx\n", addr);
    SysSetGlobalPointer(addr);
}


/*=======================================================
=======================================================*/
extern ADDRESS SysGetProgramEntry()
{

    return BIN_ENTRY;
}

/*=======================================================
=======================================================*/
extern INDEX SysGetTextSection()
{
    return BIN_SECTION_MAP[R_SN_TEXT];
}
/*=======================================================
=======================================================*/
extern INDEX SysGetSwitchSection()
{
    return BIN_SECTION_MAP[R_SN_DATA];
}
/*=======================================================
=======================================================*/
extern void SysUpdateTextSegmentSize()
{
#ifdef COFF
    INDEX isec = SysGetTextSection();
#endif
    /*
      we assume that the text section is the last section within the text segment
      without this assumption we could not even enlarge the text section
     */

    /* I don't think this is a problem in ELF */
#if 0
    printf("Old: TextSegemnt size is: %016lx\n",BIN_AOUTHDR.tsize);
#endif
#ifdef COFF

    BIN_AOUTHDR.tsize = ASEC_START(isec) + ASEC_SIZE(isec) - BIN_AOUTHDR.text_start;
    BIN_AOUTHDR.tsize = (BIN_AOUTHDR.tsize + 0x2000) & ~(0x2000-1);
#endif
#if 0
    printf("New: TextSegemnt size is: %016lx\n",BIN_AOUTHDR.tsize);
#endif    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "Addr2Ins"

extern INDEX Addr2Ins(ADDRESS address)
{
    INDEX isec_text = BIN_SECTION_MAP[R_SN_TEXT];
    return (address-ASEC_START(isec_text))>>2;    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "Ins2Addr"

extern ADDRESS Ins2Addr(INDEX iins)
{
    INDEX isec_text = BIN_SECTION_MAP[R_SN_TEXT];
    if( !VALID(iins) ) {
	printf("returning invalid, iins is %d\n", iins);
        return INVALID;
    }
    return (iins*4) + ASEC_START( isec_text );
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FindSymbol"


/* return a pointer to the symbol, NULL if we couldn't find it */
extern SYMBOL *GetSymbol(ADDRESS address) {
    UINT32 is;

    /* first find the symbol table */
    FOREACH_SYM(is) {
	if ( SYM_ADDRESS(is) == address )
	    return &SYM(is);
    }

    return NULL;
}

/* COFF appears to have all sorts of wacky symbol tables. 
 * ELF appears to have only one, so lots of this code will
 * not be used on ELF systems.
 */
extern STRING FindSymbol(ADDRESS address)
{
    UINT32 is;

    /* first find the symbol table */
    for (is=0; is < BIN_SYMBOL_NENT; is++) {
	if ( SYM_ADDRESS(is) == address )
	    return SYM_NAME(is);
    }
    
    return "++undef++";
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "IsFunctionStart"

extern BOOL IsFunctionStart(ADDRESS address)
{
    SYMBOL *s;

    s = GetSymbol(address);

    if (s != NULL) {
	/* found the symbol */
	return SYM_IS_PROC(s);
    }
    
    /* couldn't find the symbol */
    return 1;
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "LocateSection"

/* we allow the address to point to the end of a section - dangerous */

extern INDEX LocateSection(ADDRESS address)
{
    UINT32 is;
    
    FOREACH_SEC(is)
    {
        if( AsecContainsAddress(is,address) )
            return is;
    }
    
    is = BIN_SECTION_MAP[R_SN_TEXT];
    if( address==ASEC_START(is)+ASEC_SIZE(is) )
    {
#if 0        
        WARNING(("access to end of text section \n"));
#endif        
        return is;
    }
    else
    {
#if 0        
        WARNING(("Cannot locate address: %lx\n",address));
#endif        
    }
    
    return INVALID;
}



/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecNew"

extern INDEX AsecNew()
{
    CFATAL(CONFIG_USEC >= CONFIG_NSEC,("not enough sections\n"));
    return(CONFIG_USEC++);
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecSetSize"

extern void AsecSetSize(INDEX isec, INT64 size)
{
    ASEC_SIZE(isec) = size;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecGetSize"

extern INT64 AsecGetSize(INDEX isec)
{
    return ASEC_SIZE(isec);
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecContainsAddress"

extern BOOL AsecContainsAddress(INDEX isec,ADDRESS addr)
{
    if( addr >= ASEC_START(isec) && addr < ASEC_START(isec) + ASEC_SIZE(isec) )
        return 1;
    else
        return 0;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecMapAddress"

extern ADDRESS AsecMapAddress(INDEX isec,ADDRESS addr)
{
    CASSERT( AsecContainsAddress(isec,addr),("weird\n"));
    return addr - ASEC_START(isec) + (INT64)ASEC_DATA8(isec);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecIsReadOnly"

extern BOOL AsecIsReadOnly(INDEX isec)
{
    if( ASEC_IS_RDATA(isec)  ||
        ASEC_IS_TEXT(isec)   ||
        ASEC_IS_INIT(isec)   ||
        ASEC_IS_FINI(isec)   ||
	ASEC_IS_GOT(isec)    ||
        ASEC_IS_RCONST(isec) )
        return 1;
    else
        return 0;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecWriteData32"

extern void AsecWriteData32(INDEX isec, NUMBER index,INT32 value)
{
    /*
    CASSERT( index*4 < ASEC_SIZE(isec),("weird\n"));
    */
    ASEC_DATA32(isec)[index] = value;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecReadData32"

extern INT32 AsecReadData32(INDEX isec, NUMBER index)
{
    CASSERT( index*4 < ASEC_SIZE(isec),("weird\n"));
    return ASEC_DATA32(isec)[index];
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecGetAddressData32"

extern ADDRESS AsecGetAddressData32(INDEX isec, NUMBER index)
{
    CASSERT( index*4 < ASEC_SIZE(isec),("weird\n"));
    return ASEC_START(isec) + 4 *index;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecIsCode"

extern BOOL AsecIsCode(INDEX isec)
{
    if( ASEC_IS_TEXT(isec)   ||
        ASEC_IS_INIT(isec)   ||
        ASEC_IS_FINI(isec)   )
        return 1;
    else
        return 0;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ReadData64"

extern INT64 ReadData64(ADDRESS address)
{
    UINT32 is = LocateSection(address);
    CASSERT(ASEC_DATA64(is),("illegal access\n"));
    return ASEC_DATA64(is)[(address-ASEC_START(is))>>3];
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "WriteData64"

extern void WriteData64(ADDRESS address, INT64 value)
{
    UINT32 is = LocateSection(address);
    CASSERT(ASEC_DATA64(is),("illegal access\n"));
    ASEC_DATA64(is)[(address-ASEC_START(is))>>3]  = value ;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ReadData32"

extern INT32 ReadData32(ADDRESS address)
{
    UINT32 is = LocateSection(address);
    CASSERT(ASEC_DATA32(is),("illegal access\n"));
    return ASEC_DATA32(is)[(address-ASEC_START(is))>>2];
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "WriteData32"

extern void WriteData32(ADDRESS address, INT32 value)
{
    UINT32 is = LocateSection(address);
    CASSERT(ASEC_DATA32(is),("illegal access\n"));
    ASEC_DATA32(is)[(address-ASEC_START(is))>>2] = value;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ReadData16"

extern INT16 ReadData16(ADDRESS address)
{
    UINT32 is = LocateSection(address);
    CASSERT(ASEC_DATA16(is),("illegal access\n"));
    return ASEC_DATA16(is)[(address-ASEC_START(is))>>1];
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FetchData8"

extern INT8 ReadData8(ADDRESS address)
{
    UINT32 is = LocateSection(address);
    CASSERT(ASEC_DATA8(is),("illegal access\n"));
    return ASEC_DATA8(is)[(address-ASEC_START(is))];
}

/*=======================================================
=======================================================*/
/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "PrintRELOC"

static char *sn[] =
{
    "R_SN_NULL  ",
    "R_SN_TEXT  ",
    "R_SN_RDATA ",
    "R_SN_DATA  ",
    "R_SN_SDATA ",
    "R_SN_SBSS  ",
    "R_SN_INIT  ",
    "R_SN_LIT8  ",
    "R_SN_LIT4  ",
    "R_SN_XDATA ",
    "R_SN_PDATA ",
    "R_SN_FINI  ",
    "R_SN_LITA  ",
    "R_SN_ABS   ",
    "R_SN_RCONST"
};

static char *type[] =
{
    "R_ABS       ",
    "R_REFLONG   ",
    "R_REFQUAD   ",
    "R_GPREL32   ",
    "R_LITERAL   ",
    "R_LITUSE    ",
    "R_GPDISP    ",
    "R_BRADDR    ",
    "R_HINT      ",
    "R_SREL16    ",
    "R_SREL32    ",
    "R_SREL64    ",
    "R_OP_PUSH   ",
    "R_OP_STORE  ",
    "R_OP_PSUB   ",
    "R_OP_PRSHIFT",
    "R_GPVALUE   ",
    "R_GPRELHIGH ",
    "R_GPRELLOW  ",
    "R_IMMED     ",
};



static void PrintRELOC(FILE *fp,RELOC *r)
{
#ifdef COFF
    fprintf(fp,
            "r_vaddr:    %016lx\n"
            "r_symndex:  %08x\n"
            "r_type:     %1x\n"
            "r_extern:   %1x\n"
            "r_offset:   %1x\n"
            "r_size:     %1x\n",
            r->r_vaddr,
            r->r_symndx,
            r->r_type,
            r->r_extern,
            r->r_offset,
            r->r_size);
    
    fprintf(fp,"r_symndx:   %s\n",
            r->r_symndx <= MAX_R_SN ? sn[r->r_symndx] : "SYMBOL");

    fprintf(fp,"r_type:     %s\n",
            r->r_type < MAX_R_TYPE ? type[r->r_type] : "ILLEGAL");
#endif

#ifdef ELF
    fprintf(fp,
            "r_symndex:  %016lx\n"
            "r_type:     %016lx\n"
            "r_offset:   %016lx\n",
            ELF64_R_SYM(r->r_info),
            ELF64_R_TYPE(r->r_info),
            r->r_offset);
#endif
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintRELOC2"

static void PrintRELOC2(FILE *fp,RELOC *r)
{

    fprintf(fp, "printreloc2 not defined at the moment ");
#ifdef COFF    
    fprintf(fp,"%016lx %s %s %s\n",
            r->r_vaddr,
            r->r_type < MAX_R_TYPE ? type[r->r_type] : "ILLEGAL",
            r->r_extern ? "EXT" : "INT",
            r->r_symndx <= MAX_R_SN ? sn[r->r_symndx] : "SYMBOL");
#endif
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "SysMarkRelocAndSymbol"

/* figure out what marking the reloc and symbol function do */
extern void SysMarkRelocAndSymbol(FILE *fp)
{
    INDEX is;
    INDEX isec;
    INDEX isec_text;
    INDEX iins;
    FILE *fp2;
    char line[81];
    char sec[81];
    INT64 *gaddr;
    INDEX goff;
    ADDRESS address;

    PRINT_FUN_BEG3();

#ifdef ELF

    FOREACH_SYM(is)  /* do symbols */
	{
	    SYMBOL *s = &SYM(is);
	    if( SYM_IS_PROC(s) && SYM_VALUE(is) != 0 )
	    {
            iins = Addr2Ins(SYM_VALUE(is));
            AINS_FLAGS(iins) |= FI_FUN_BEG | FI_BBL_BEG;
            if (iins)
                AINS_FLAGS(iins-1) |= FI_BBL_END | FI_FUN_END;
        }
    }




    /* read in the table we wrote using the patcher program */
    fp2 = fopen("altorelocationfile", "r");
    if ( fp2 == 0 )
	fprintf(stderr, "cannot open input file %s\n", "altorelocationfile");

    /*
      format is as follows:
       section name
       max # of relocs
       offset type section address addend
       offset type section address addend
       ....
       new section name
       ...
       EOF 
    */
    while (fgets(line, 80, fp2) != NULL) { 
        INDEX irel;
        INDEX iins;
	INDEX target;
        static ADDRESS immed;
	int reloc_sec;
	int nrelocs, relocType;
	INT64 offset, addr, addend;
	int section;

	sscanf(line, "%s", sec);
#if 0
	printf("relocations for [%s]\n", sec);
#endif
	reloc_sec = -1;
	FOREACH_SEC(isec) {  /* which section is being relocated */
	    if ( strcmp(&BIN_SHDRSTRTAB[ASEC_NAME(isec)], sec) == 0 ) { 
		printf("found section, %d\n", isec);
		reloc_sec = isec;
	    }
	}

	
	/* now we know which section to relocate, so let's do the relocation */
	fgets(line, 80, fp2);
	sscanf(line, "%d", &nrelocs);
#if 1
	printf("number of relocs found is %d\n", nrelocs);
#endif
	/* read lines until we get to the end of the section, this is denoted
	   by a 200 value for the type */
	while (fgets(line, 80, fp2) != NULL) {
	    sscanf(line, "%lx%d%d%lx%lx", &offset, &relocType, 
		   &isec, &addr, &addend);
	    if (relocType == 200) break;
	    /* valid relocation, so let's do it */
	    switch ( isec ) {
	      case 1: /* text section */
		  section = BIN_SECTION_MAP[R_SN_TEXT];
		  break;
	      case 2: /* data section */
		  section = BIN_SECTION_MAP[R_SN_DATA];
		  break;
	      case 3: /* rodata section */
		  section = BIN_SECTION_MAP[R_SN_RDATA];
		  break;
	      case 4: /* dtors section */
		  section = BIN_SECTION_MAP[R_SN_DTORS];
		  break;
	      case 5: /* ctors section */
		  section = BIN_SECTION_MAP[R_SN_CTORS];
		  break;
	      case 6: /* init section */
		  section = BIN_SECTION_MAP[R_SN_INIT];
		  break;
	      case 7: /* fini section */
		  section = BIN_SECTION_MAP[R_SN_FINI];
		  break;
	      default: 
		  printf("failed to determine reasonable section!");
	    }
	    address = ASEC_SCNHDR(reloc_sec).sh_addr + offset;
	    /* if addres to be relocated is in the data section, then we want to
	       make sure we make up for the addition of the 4*config_ntab */
	    if ( reloc_sec == BIN_SECTION_MAP[R_SN_DATA] ) { 
  		address += CONFIG_COUNTERSPACE;
 	    } 
	    isec = section;
#if 1
	    printf("relocating address %lx. section is %d\n", address, section);
	    printf("section headder address is %lx\n", ASEC_SCNHDR(reloc_sec).sh_addr);
#endif
	    switch( relocType )
	    {
	      case R_ALPHA_REFLONG:             /* Direct 32 bit */
	      case R_ALPHA_LITERAL:             /* GP relative 16 bit w/optimization */
	      case R_ALPHA_LITUSE:              /* Optimization hint for LITERAL */
	      case R_ALPHA_GPDISP:              /* Add displacement to GP */
	      case R_ALPHA_BRADDR:              /* PC+4 relative 23 bit shifted */
	      case R_ALPHA_HINT:                /* PC+4 relative 16 bit shifted */
	      case R_ALPHA_SREL16:              /* PC relative 16 bit */
	      case R_ALPHA_SREL32:              /* PC relative 32 bit */
	      case R_ALPHA_SREL64:              /* PC relative 64 bit */
	      case R_ALPHA_GPVALUE:
	      case R_ALPHA_GPRELHIGH:
	      case R_ALPHA_GPRELLOW:
	      case R_ALPHA_IMMED_GP_16:    
	      case R_ALPHA_IMMED_GP_HI32:
	      case R_ALPHA_IMMED_BR_HI32:
	      case R_ALPHA_COPY:                /* Copy symbol at runtime */
	      case R_ALPHA_GLOB_DAT:            /* Create GOT entry */
	      case R_ALPHA_JMP_SLOT:            /* Create PLT entry */
	      case R_ALPHA_RELATIVE:            /* Adjust by program base */

/* 	  case R_HINT: */
/* 	  case R_LITUSE: */
/* 	  case R_BRADDR: */
/* 	  case R_GPDISP: */
	      CASSERT(AsecIsCode(isec),("strange reloc in non code sec\n"));
	      break;
	      
	      
	      /* do not know what to do with those */
	      
	      case R_ALPHA_NONE:       /* No reloc */
	      case R_ALPHA_OP_PUSH:             /* OP stack push */
	      case R_ALPHA_OP_STORE:            /* OP stack pop and store */
	      case R_ALPHA_OP_PSUB:             /* OP stack subtract */
	      case R_ALPHA_OP_PRSHIFT:          /* OP stack right shift */
	      break;
	      
	      
	      case R_ALPHA_REFQUAD:             /* Direct 64 bit */
		  CASSERT(!AsecIsCode(isec),("refquad in code sec\n"));
		  
	      address = ReadData64( address );
	      printf("address read is %lx\n", address);
	      iins = Addr2Ins(address);

	      printf("Ins2Addr of %d is %lx\n", iins, Ins2Addr(iins));
	      printf("reloc address is %lx\n", Ins2Addr(iins));

	      AINS_FLAGS(iins) |= F_REFQUAD;
	      AINS_FLAGS(iins) |= FI_BBL_BEG;
	      if( iins > 0 )
		  AINS_FLAGS(iins-1) |= FI_BBL_END;
	      
	      if( IsFunctionStart(address) & !(AINS_FLAGS(iins) & FI_FUN_BEG) )
	      {
		  WARNING(("REFQUAD for non function at %lx (%s)\n",
			   address,FindSymbol(address)));
		  
		  AINS_FLAGS(iins) |=  FI_FUN_BEG;
		  if( iins > 0 )
		      AINS_FLAGS(iins-1) |= FI_FUN_END;
	      }
	      break;
	      
	      case R_ALPHA_GPREL32:             /* GP relative 32 bit */
		  /* NEEDS WORK ??? */
		  CASSERT(!AsecIsCode(isec),("gprel32 in code sec\n"));
		  iins = Addr2Ins(SysGetGlobalPointer() +
				  ReadData32(address)); 

#if 0
		  printf("Ins2Addr of %d is %lx\n", iins, Ins2Addr(iins));
		  printf("reloc address is %lx\n", Ins2Addr(iins));
#endif
		  AINS_FLAGS(iins) |= F_GPREL32 | FI_BBL_BEG; 
		  if( iins ) 
		      AINS_FLAGS(iins-1) |= FI_BBL_END; 
		  break; 
		  
	      case R_ALPHA_IMMED_SCN_HI32:
		  iins = Addr2Ins( address );
		  CASSERT( AINS_CODE(iins) == I_ldah,("weird\n"));
		  immed =  (long)AINS_DATA(iins)<<16L;
		  break;

	      case R_ALPHA_IMMED_LO32:
		    iins = Addr2Ins( address );
		  CASSERT( AINS_CODE(iins) == I_lda,("weird\n"));
		  immed +=  (long)AINS_DATA(iins);
		  immed += ASEC_START(isec);
		  
		  /*
		    This relocation type typically occurs with
		    jsr in the fini or init section
		    we reduce those to bsr here rather than letting the
		    partial evaluator do the job! 
		  */
		  
		  WARNING(("IMMED   for function at %lx (%s) (REFQUAD)\n",immed,FindSymbol(immed)));
		  CASSERT( AINS_REG_C(iins) == REG_PV,("unexpected IMMED reloc reg\n"));
		  iins++;
		  CASSERT( AINS_CODE(iins) == I_jsr,("unexpected IMMED reloc ins\n"));
		  target = Addr2Ins(immed);
		  AINS_CODE(iins) = I_bsr;
		  AINS_REG_B(iins) = REG_ZERO;
		  AINS_DATA(iins) = target - (iins+1);
		  
		  AINS_FLAGS(target) |= F_REFQUAD;
		  AINS_FLAGS(target) |= FI_BBL_BEG;
		  if( iins > 0 )
		      AINS_FLAGS(target-1) |= FI_BBL_END;
		  AINS_FLAGS(target) |=  FI_FUN_BEG;
		  if( iins > 0 )
		      AINS_FLAGS(target-1) |= FI_FUN_END;
	      break;
	      
	  default:
	      ERROR(("unexpected reloc\n"));
	      PrintRELOC2(fp,&ASEC_AREL(isec,irel));
	    }

	}
    }

#if 1
    printf("************************************\n");
    printf(" marking the global offset table now\n");
    printf("************************************\n");
    is = BIN_SECTION_MAP[R_SN_GOT];
    gaddr = ASEC_DATA64(is);
    isec_text = BIN_SECTION_MAP[R_SN_TEXT];
    printf("text section is %d\n", isec_text);

    for (goff = 0; goff < ASEC_SCNHDR(is).sh_size / 8; goff++) {
	address = gaddr[goff];
	
	/* if the address is in the text section, we need to patch
	   it. */
#if 1
	printf("virtual address is currently %lx\n", ASEC_SCNHDR(is).sh_addr + (goff * 8));
	printf("real address is %lx\n", address);
#endif
	if ((address >= ASEC_SCNHDR(isec_text).sh_addr) &&
	    (address <= ASEC_SCNHDR(isec_text).sh_addr + ASEC_SCNHDR(isec_text).sh_size)) {
	    iins = Addr2Ins(address);
	
#if 0
	    printf("Ins2Addr of %d is %lx\n", iins, Ins2Addr(iins));
	    printf("basic block of instruction is %d\n", AINS(iins).Bblock);
	    printf("reloc address is %lx\n", Ins2Addr(iins));
#endif
	    AINS_FLAGS(iins) |= F_REFQUAD;
	    AINS_FLAGS(iins) |= FI_BBL_BEG;
	    if( iins > 0 )
		AINS_FLAGS(iins-1) |= FI_BBL_END;
	    
	    if( IsFunctionStart(address) & !(AINS_FLAGS(iins) & FI_FUN_BEG) )
	    {
		WARNING(("REFQUAD for non function at %lx (%s)\n",
			 address,FindSymbol(address)));
		
		AINS_FLAGS(iins) |=  FI_FUN_BEG;
		if( iins > 0 )
		    AINS_FLAGS(iins-1) |= FI_FUN_END;
	    }
	}
    }
#endif
#endif

    fprintf(fp,"\t[MarkRelocAndSymbol]\n");
    PRINT_FUN_END3();
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "SysRelocate"

/* This will actually perform the relocations */
extern void SysRelocate(FILE *fp)
{
    INDEX isec;
    ADDRESS address;
    ADDRESS newaddress;
    FILE *fp2;
    char line[80];
    char sec[80];





    fp2 = fopen("altorelocationfile", "r");
    if ( fp2 == 0 )
	fprintf(stderr, "cannot open input file %s\n", "altorelocationfile");

    /*
      format is as follows:
       section name
       max # of relocs
       offset type section address addend
       offset type section address addend
       ....
       new section name
       ...
       EOF 
    */
    while (fgets(line, 80, fp2) != NULL) { 
	int reloc_sec;
	int nrelocs, relocType;
	INT64 offset, addr, addend;
	int section;

	sscanf(line, "%s", sec);
	printf("relocations for [%s]\n", sec);
	reloc_sec = -1;
	FOREACH_SEC(isec) {  /* which section is being relocated */
	    if ( strcmp(&BIN_SHDRSTRTAB[ASEC_NAME(isec)], sec) == 0 ) { 
		printf("found section, %d\n", isec);
		reloc_sec = isec;
	    }
	}

	
	/* now we know which section to relocate, so let's do the relocation */
	fgets(line, 80, fp2);
	sscanf(line, "%d", &nrelocs);
	printf("number of relocs found is %d\n", nrelocs);
	/* read lines until we get to the end of the section, this is denoted
	   by a 200 value for the type */
	while (fgets(line, 80, fp2) != NULL) {
	    sscanf(line, "%lx%d%d%lx%lx", &offset, &relocType, 
		   &isec, &addr, &addend);
	    section = 0;
	    if (relocType == 200) break;
	    /* valid relocation, so let's do it */
	    switch ( isec ) {
	      case 1: /* text section */
		  section = BIN_SECTION_MAP[R_SN_TEXT];
		  break;
	      case 2: /* data section */
		  section = BIN_SECTION_MAP[R_SN_DATA];
		  break;
	      case 3: /* rodata section */
		  section = BIN_SECTION_MAP[R_SN_RDATA];
		  break;
	      case 4: /* dtors section */
		  section = BIN_SECTION_MAP[R_SN_DTORS];
		  break;
	      case 5: /* ctors section */
		  section = BIN_SECTION_MAP[R_SN_CTORS];
		  break;
	      case 6: /* init section */
		  section = BIN_SECTION_MAP[R_SN_INIT];
		  break;
	      case 7: /* fini section */
		  section = BIN_SECTION_MAP[R_SN_FINI];
		  break;
	      default: 
		  printf("failed to determine reasonable section!\n");
	    }
	    address = ASEC_SCNHDR(reloc_sec).sh_addr + offset;

	    printf(" read from address %lx\n", address);
	    /* if address to be relocated is in the data section, then we want to
	       make sure we make up for the addition of the 4*config_ntab */
 	    if ( reloc_sec == BIN_SECTION_MAP[R_SN_DATA] ) { 
 		printf("adding offset back in for address \n"); 
  		address += CONFIG_COUNTERSPACE;  
 	    } 
	    isec = section;
	    
	    
            if( relocType ==  R_ALPHA_REFQUAD )
            {
                newaddress = ReadData64(address);
		printf("relocating address %lx \n", newaddress);
                newaddress = TranslateAddress(newaddress);
		printf("moving address to %lx \n", newaddress);
                WriteData64(address,newaddress);
            }
	    
            if( relocType ==  R_ALPHA_GPREL32 )
            {
		newaddress = ReadData32(address);
		printf("relocating address(32) %lx \n", newaddress);
                newaddress += SysGetGlobalPointer();
		printf("relocating address %lx \n", newaddress);
                newaddress = TranslateAddress(newaddress);
		printf("moving address to %lx \n", newaddress);
                WriteData32(address,(INT32)(newaddress - SysGetGlobalPointer()));

                /* NEEDS WORK */ 
            }
	}
    }
    fclose(fp2);
}

/*=======================================================
=======================================================*/
/*=======================================================
=======================================================*/
/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "PrintSCNHDR"

#ifdef COFF
static void PrintSCNHDR(FILE *fp, SCNHDR *s)
{
    CASSERT(fp,("nil FILE pointer\n"));
    CASSERT(s,("nil SCNHDR pointer\n"));

    fprintf(fp,
            "s_name:     %.8s\n"
            "s_paddr:    %016lx\n"
            "s_vaddr:    %016lx\n"
            "s_size:     %016lx\n"
            "s_scnptr:   %016lx\n"
            "s_relptr:   %016lx\n"
            "s_lnnoptr:  %016lx\n"
            "s_nreloc:   %04hx\n"
            "s_nlnno:    %04hx\n"
            "s_flags:    %08x\n",
            s->s_name,
            s->s_paddr,
            s->s_vaddr,
            s->s_size,
            s->s_scnptr,
            s->s_relptr,
            s->s_lnnoptr,
            s->s_nreloc,
            s->s_nlnno,
            s->s_flags);

    fprintf(fp,"s_flags:    ");

    if( IS_STYP_REG(s->s_flags) )      fprintf(fp,"STYP_REG ");
    if( IS_STYP_DSECT(s->s_flags) )    fprintf(fp,"STYP_DSECT ");
    if( IS_STYP_NOLOAD(s->s_flags) )   fprintf(fp,"STYP_NOLOAD ");
    if( IS_STYP_GROUP(s->s_flags) )    fprintf(fp,"STYP_GROUP ");
    if( IS_STYP_PAD(s->s_flags) )      fprintf(fp,"STYP_PAD ");
    if( IS_STYP_COPY(s->s_flags) )     fprintf(fp,"STYP_COPY ");
    if( IS_STYP_TEXT(s->s_flags) )     fprintf(fp,"STYP_TEXT ");
    if( IS_STYP_DATA(s->s_flags) )     fprintf(fp,"STYP_DATA ");
    if( IS_STYP_BSS(s->s_flags) )      fprintf(fp,"STYP_BSS ");
    if( IS_STYP_RDATA(s->s_flags) )    fprintf(fp,"STYP_RDATA ");
    if( IS_STYP_SDATA(s->s_flags) )    fprintf(fp,"STYP_SDATA ");
    if( IS_STYP_SBSS(s->s_flags) )     fprintf(fp,"STYP_SBSS ");
    if( IS_STYP_UCODE(s->s_flags) )    fprintf(fp,"STYP_UCODE ");
    if( IS_STYP_GOT(s->s_flags) )      fprintf(fp,"STYP_GOT ");
    if( IS_STYP_DYNAMIC(s->s_flags) )  fprintf(fp,"STYP_DYNAMIC ");
    if( IS_STYP_DYNSYM(s->s_flags) )   fprintf(fp,"STYP_DYNSYM ");
    if( IS_STYP_REL_DYN(s->s_flags) )  fprintf(fp,"STYP_REL_DYN ");
    if( IS_STYP_DYNSTR(s->s_flags) )   fprintf(fp,"STYP_DYNSTR ");
    if( IS_STYP_HASH(s->s_flags) )     fprintf(fp,"STYP_HASH ");
    if( IS_STYP_CONFLICT(s->s_flags) ) fprintf(fp,"STYP_CONFLICT ");
    if( IS_STYP_REGINFO(s->s_flags) )  fprintf(fp,"STYP_REGINFO ");
    if( IS_STYP_FINI(s->s_flags) )     fprintf(fp,"STYP_FINI ");
    if( IS_STYP_INIT(s->s_flags) )     fprintf(fp,"STYP_INIT ");
    if( IS_STYP_COMMENT(s->s_flags) )   fprintf(fp,"STYP_COMMENT ");
    if( IS_STYP_RCONST(s->s_flags) )   fprintf(fp,"STYP_RCONST ");
    if( IS_STYP_XDATA(s->s_flags) )    fprintf(fp,"STYP_XDATA ");
    if( IS_STYP_PDATA(s->s_flags) )    fprintf(fp,"STYP_PDATA ");
    if( IS_STYP_LITA(s->s_flags) )     fprintf(fp,"STYP_LITA ");
    if( IS_STYP_LIT8(s->s_flags) )     fprintf(fp,"STYP_LIT8 ");
    if( IS_STYP_LIT4(s->s_flags) )     fprintf(fp,"STYP_LIT4 ");
    if( IS_S_NRELOC_OVFL(s->s_flags) ) fprintf(fp,"S_NRELOC_OVFL ");
    fprintf(fp,"\n");
}
#endif

#ifdef ELF
static void PrintSCNHDR(FILE *fp, SCNHDR *s)
{
    CASSERT(fp,("nil FILE pointer\n"));
    CASSERT(s,("nil SCNHDR pointer\n"));

    fprintf(fp,
            "sh_name:     %.8s\n"
            "sh_addr:     %016lx\n"
            "sh_flags:    %016lx\n"
            "sh_size:     %016lx\n"
            "sh_offset:   %016lx\n"
	    "sh_link:     %016x\n"
	    "sh_addralign %016lx\n"
	    "sh_info:     %016x\n"
	    "sh_entsize:  %016lx\n",
            &BIN_SHDRSTRTAB[s->sh_name],
            s->sh_addr,
	    s->sh_flags,
            s->sh_size,
            s->sh_offset,
	    s->sh_link,
	    s->sh_addralign,
	    s->sh_info,
	    s->sh_entsize);

    fprintf(fp,"s_flags:    ");

    if( IS_STYP_TEXT(s->sh_flags) )     fprintf(fp,"STYP_TEXT ");
    if( IS_STYP_DATA(s->sh_flags) )     fprintf(fp,"STYP_DATA ");
    if( IS_STYP_BSS(s->sh_flags) )      fprintf(fp,"STYP_BSS ");
    if( IS_STYP_RDATA(s->sh_flags) )    fprintf(fp,"STYP_RDATA ");
    if( IS_STYP_SDATA(s->sh_flags) )    fprintf(fp,"STYP_SDATA ");
    if( IS_STYP_SBSS(s->sh_flags) )     fprintf(fp,"STYP_SBSS ");
    if( IS_STYP_GOT(s->sh_flags) )      fprintf(fp,"STYP_GOT ");
    if( IS_STYP_DYNAMIC(s->sh_flags) )  fprintf(fp,"STYP_DYNAMIC ");
    if( IS_STYP_DYNSYM(s->sh_flags) )   fprintf(fp,"STYP_DYNSYM ");
    if( IS_STYP_REL_DYN(s->sh_flags) )  fprintf(fp,"STYP_REL_DYN ");
    if( IS_STYP_DYNSTR(s->sh_flags) )   fprintf(fp,"STYP_DYNSTR ");
    if( IS_STYP_HASH(s->sh_flags) )     fprintf(fp,"STYP_HASH ");
    if( IS_STYP_CONFLICT(s->sh_flags) ) fprintf(fp,"STYP_CONFLICT ");
    if( IS_STYP_REGINFO(s->sh_flags) )  fprintf(fp,"STYP_REGINFO ");
    if( IS_STYP_FINI(s->sh_flags) )     fprintf(fp,"STYP_FINI ");
    if( IS_STYP_INIT(s->sh_flags) )     fprintf(fp,"STYP_INIT ");
    if( IS_STYP_COMMENT(s->sh_flags) )   fprintf(fp,"STYP_COMMENT ");
    if( IS_STYP_RCONST(s->sh_flags) )   fprintf(fp,"STYP_RCONST ");
    if( IS_STYP_XDATA(s->sh_flags) )    fprintf(fp,"STYP_XDATA ");
    if( IS_STYP_PDATA(s->sh_flags) )    fprintf(fp,"STYP_PDATA ");
    if( IS_STYP_LITA(s->sh_flags) )     fprintf(fp,"STYP_LITA ");
    if( IS_STYP_LIT8(s->sh_flags) )     fprintf(fp,"STYP_LIT8 ");
    if( IS_STYP_LIT4(s->sh_flags) )     fprintf(fp,"STYP_LIT4 ");
    if( IS_S_NRELOC_OVFL(s->sh_flags) ) fprintf(fp,"S_NRELOC_OVFL ");
    fprintf(fp,"\n");
}
#endif
/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "PrintSec"

extern void PrintSec(FILE *fp, INDEX isec)
{
    PrintSCNHDR(fp,&ASEC_SCNHDR(isec));
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintSym"

static char *stTab[] =
{
    "stNil",
    "stGlobal",
    "stStatic",
    "stParam",
    "stLocal",
    "stLabel",
    "stProc",
    "stBlock",
    "stEnd",
    "stMember",
    "stTypedef",
    "stFile",
    "stStaticProc",
    "stConstant",
    "stStaParam",
    "stBase",
    "stVirtBase",
    "stTag",
    "stInter"
};

static char *scTab[] =
{
"scText",
"scData",
"scBss",
"scRegister",
"scAbs",
"scUndefined",
"scCdbLocal",
"scBits",
"scCdbSystem",
"scDbx",
"scRegImage",
"scInfo",
"scUserStruct",
"scSData",
"scSBss",
"scRData",
"scVar",
"scCommon",
"scSCommon",
"scVarRegister",
"scVariant",
"scFileDesc",
"scSUndefined",
"scInit",
"scBasedVar",
"scReportDesc",
"scXData",
"scPData",
"scFini",
"scRConst"
};
            
/*=======================================================
=======================================================*/

#ifdef COFF
static void PrintSym(FILE *fp, SYMR *s, STRING name)
{
    fprintf(fp,"VAL: %016lx  NM: %08x (%-20s) TY: %1x (%-10s) CL: %1x (%-10s)  AX: %08x\n",
            s->value,
            s->iss,
            name,
            s->st,
            s->st < 21 ? stTab[s->st] : "ILLEGAL",  /* stMax is too big */
            s->sc,
            s->sc < scMax ? scTab[s->sc] : "ILLEGAL",
            s->index);
}
#endif

#ifdef ELF
static void PrintSym(FILE *fp, SYMBOL *s, STRING name)
{
    fprintf(fp,"NM: %08x (%-20s)  VAL: %016lx  SZ: %016lx  INFO: %016x  OTHER: %016x  SHNDX: %016x",
            s->st_name,
	    (char *)&BIN_STRING_TAB[s->st_name],
            s->st_value,
	    s->st_size,
	    s->st_info,
	    s->st_other,
            s->st_shndx);

    fprintf(fp, " info means :  BIND: " );
    
    switch (ELF64_ST_BIND(s->st_info)) {
      case STB_LOCAL  : fprintf(fp, " LOCAL  ");
	  break;
      case STB_GLOBAL : fprintf(fp, " GLOBAL ");
	  break;
      case STB_WEAK   : fprintf(fp, " WEAK   ");
	  break;
      case STB_LOPROC : fprintf(fp, " LOPROC ");
	  break;
      case STB_HIPROC : fprintf(fp, " HIPROC ");
	  break;
      default:
	  fprintf(fp, " UNKNOWN ");
    }

    fprintf(fp, "  TYPE:  ");
    
    switch (ELF64_ST_TYPE(s->st_info)) {
      case STT_NOTYPE  : fprintf(fp, " NOTYPE ");
	  break;
      case STT_OBJECT  : fprintf(fp, " OBJECT ");
	  break;
      case STT_FUNC    : fprintf(fp, " FUNC   ");
	  break;
      case STT_SECTION : fprintf(fp, " SECTION ");
	  break;
      case STT_FILE    : fprintf(fp, " FILE    ");
	  break;
      case STT_LOPROC  : fprintf(fp, " LOPROC  ");
	  break;
      case STT_HIPROC  : fprintf(fp, " HIPROC  ");
	  break;
      default:
	  fprintf(fp, " UNKOWN ");
    }
    fprintf(fp, "\n");

}

#endif            
/*=======================================================
=======================================================*/
/* don't have the esy or lsy in ELF */
#ifdef COFF
extern void PrintEsy(FILE *fp, INDEX iesy)
{
    PrintSym(fp,&AESY(iesy),AESY_NAME(iesy));
}
            
/*=======================================================
=======================================================*/

extern void PrintLsy(FILE *fp, INDEX ilsy)
{
    PrintSym(fp,&ALSY(ilsy),ALSY_NAME(ilsy));
}
#endif
    

/*=======================================================
=======================================================*/
#if 0
#undef __PROC__
#define __PROC__ "PrintPdt"

extern void PrintPdt(FILE *fp,INDEX ip)
{
    PDR *p = &APDR(ip);
                   
    fprintf(fp,
            "adr:        %016lx\n"
            "cbLineOff.: %016lx\n"
            "isym:       %08x (%s)\n"
            "iline:      %08x\n"
            "regmask:    %08x\n"
            "regoffset:  %08x\n"
            "iopt:       %08x\n"
            "fregmask:   %08x\n"
            "fregoffset: %08x\n"
            "frameoff.:  %08x\n"
            "lnLow:      %08x\n"
            "lnHigh:     %08x\n"
            "gp_prologue: %1x\n"
            "gp_used:    %1x\n"
            "reg_frame:  %1x\n"
            "localoff:   %1x\n"
            "framereg:   %1x\n"
            "reg_frame:  %04hx\n"
            "pcreg:      %04hx\n",
            p->adr,
            p->cbLineOffset,
            p->isym,
            ASYM_NAME(p->isym),
            p->iline,
            p->regmask,
            p->regoffset,
            p->iopt,
            p->fregmask,
            p->fregoffset,
            p->frameoffset,
            p->lnLow,
            p->lnHigh,
            p->gp_prologue,
            p->gp_used,
            p->reg_frame,
            p->localoff,
            p->framereg,
            p->reg_frame,
            p->pcreg);
}

            
#endif            
/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "PrintHDRR"

static void PrintHDDR(FILE *fp,HDRR *h)
{
    fprintf(fp,
            "magic:      %04hx\n"
            "vstamp:     %04hx\n"
            "ilineMax:   %08x\n"
            "idnMax:     %08x\n"
            "ipdMax:     %08x\n"
            "isymMax:    %08x\n"
            "ioptMax:    %08x\n"
            "iauxMax:    %08x\n"
            "issMax:     %08x\n"
            "issExtMax:  %08x\n"
            "ifdMax:     %08x\n"
            "crfd:       %08x\n"
            "iextMax:    %08x\n"
            "cbLine:     %016lx\n"
            "cbLineOff.: %016lx\n"
            "cbDnOff.:   %016lx\n"
            "cbPdOff.:   %016lx\n"
            "cbSymOff.:  %016lx\n"
            "cbOptOff.t: %016lx\n"
            "cbAuxOff.:  %016lx\n"
            "cbSsOff.:   %016lx\n"
            "cbFdOff.:   %016lx\n"
            "cbRfdOff.:  %016lx\n"
            "cbExtOff.:  %016lx\n",
            h->magic,
            h->vstamp,
            h->ilineMax,
            h->idnMax,
            h->ipdMax,
            h->isymMax,
            h->ioptMax,
            h->iauxMax,
            h->issMax,
            h->issExtMax,
            h->ifdMax,
            h->crfd,
            h->iextMax,
            h->cbLine,
            h->cbLineOffset,
            h->cbDnOffset,
            h->cbPdOffset,
            h->cbSymOffset,
            h->cbOptOffset,
            h->cbAuxOffset,
            h->cbSsOffset,
            h->cbFdOffset,
            h->cbRfdOffset,
            h->cbExtOffset);
}

    
/*=======================================================
=======================================================*/

#ifdef COFF
#undef __PROC__
#define __PROC__ "PrintAOUTHDR"

static void PrintAOUTHDR(FILE *fp, AOUTHDR *a)
{
    CASSERT(fp,("nil FILE pointer\n"));
    CASSERT(a,("nil AOUTHDR pointer\n"));
    
    fprintf(fp,
            "magic:      %04hx\n"
            "vstamp:     %04hx\n"
            "bldrev:     %04hx\n"
            "tsize:      %016lx\n"
            "dsize:      %016lx\n"
            "bsize:      %016lx\n"
            "entry:      %016lx\n"
            "text_start: %016lx\n"
            "data_start: %016lx\n"
            "bss_start:  %016lx\n"
            "gprmask:    %08x\n"
            "fprmask:    %08x\n"
            "gp_value:   %016lx\n",
            a->magic,
            a->vstamp,
            a->bldrev,
            a->tsize,
            a->dsize,
            a->bsize,
            a->entry,
            a->text_start,
            a->data_start,
            a->bss_start,
            a->gprmask,
            a->fprmask,
            a->gp_value);
}


#endif

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "PrintFILEHDR"

#ifdef COFF

static void PrintFILEHDR(FILE *fp,FILHDR *f)
{
    CASSERT(fp,("nil FILE pointer\n"));
    CASSERT(f,("nil FILHDR pointer\n"));
    
    fprintf(fp,
            "f_magic:    %04hx\n"
            "f_nscns:    %04hx\n"
            "f_timdat:   %08x\n"
            "f_symptr:   %016lx\n"
            "f_nsyms:    %08x\n"
            "f_opthdr:   %04hx\n"
            "f_flags:    %04hx\n",
            f->f_magic,
            f->f_nscns,
            f->f_timdat,
            f->f_symptr,
            f->f_nsyms,
            f->f_opthdr,
            f->f_flags);
    
    fprintf(fp,"f_flags:    ");
    if( f->f_flags & F_RELFLG ) fprintf(fp,"F_RELFLG ");
    if( f->f_flags & F_EXEC )   fprintf(fp,"F_EXEC ");
    if( f->f_flags & F_LNNO )   fprintf(fp,"F_LNNO ");
    if( f->f_flags & F_LSYMS )  fprintf(fp,"F_LSYMS ");
    if( f->f_flags & F_NO_SHARED ) fprintf(fp,"F_NO_SHARED ");
    if( f->f_flags & F_LOMAP ) fprintf(fp,"F_LOMAP ");
    fprintf(fp,"\n");
}

#endif
#ifdef ELF 
static void PrintFILEHDR(FILE *fp,FILHDR *f)
{
    CASSERT(fp,("nil FILE pointer\n"));
    CASSERT(f,("nil FILHDR pointer\n"));
    
    fprintf(fp,
	    "e_ident:     %04hx\n"
	    "e_type :     %04hx\n"
	    "e_machine:   %04hx\n"
	    "e_version:   %016x\n"
	    "e_entry:     %016lx\n"
	    "e_phoff:     %016lx\n"
	    "e_shoff:     %016lx\n"
	    "e_flags:     %016x\n"
	    "e_ehsize:    %016x\n"
	    "e_phentsize: %016x\n"
	    "e_phnum:     %016x\n"
	    "e_shentsize: %016x\n"
	    "e_shnum:     %016x\n"
	    "e_shstrndx:  %016x\n",
	    f->e_ident,
	    f->e_type,    
	    f->e_machine,
	    f->e_version,  
	    f->e_entry,  
	    f->e_phoff,    
	    f->e_shoff,    
	    f->e_flags,    
	    f->e_ehsize,   
	    f->e_phentsize,
	    f->e_phnum,
	    f->e_shentsize,
	    f->e_shnum,
	    f->e_shstrndx);
    
    fprintf(fp,"\n");
}
#endif

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAllHead"

extern void PrintAllHead(FILE *fp)
{
    INDEX isec;

    PRINT_FUN_BEG1();

    PrintFILEHDR(fp,&BIN_FILHDR);
    PrintSep2(fp);
#ifdef COFF
    PrintAOUTHDR(fp,&BIN_AOUTHDR);
    PrintSep2(fp);
#endif
    for(isec=0;isec<BIN_NSECTIONS;isec++)
    {
        PrintSCNHDR(fp,&ASEC_SCNHDR(isec));
        PrintSep2(fp);
    }

    PRINT_FUN_END1();
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAllData"

extern void PrintAllData(FILE *fp)
{
    INDEX isec;

    PRINT_FUN_BEG1();

    for(isec=0; isec<BIN_NSECTIONS; isec++)
    {
        if( ASEC_IS_TEXT(isec) || ASEC_IS_BSS(isec) || ASEC_IS_SBSS (isec)) continue;

        PrintSCNHDR(fp,&ASEC_SCNHDR(isec));
        PrintHexDump(fp,ASEC_START(isec),(UINT8*)ASEC_DATA8(isec),ASEC_SIZE(isec));
    }

    PRINT_FUN_END1();
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAllCode"

extern void PrintAllCode(FILE *fp)
{
    INDEX isec;

    PRINT_FUN_BEG1();

    for(isec=0;isec<BIN_NSECTIONS;isec++)
    {
        INDEX iins;
                
        if( !ASEC_IS_TEXT(isec) ) continue;

        FOREACH_INS(iins)
        {
            PrintIns(fp,iins);
        }
    }

    PRINT_FUN_END1();
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAllReloc"

extern void PrintAllReloc(FILE *fp)
{
    INDEX isec;

    PRINT_FUN_BEG1();

    for(isec=0;isec<BIN_NSECTIONS;isec++)
    {
        INDEX irel;
        PrintSCNHDR(fp,&ASEC_SCNHDR(isec));

        for( irel=ASEC_HAS_RELOC_OVFL(isec) ? 1 : 0; irel<ASEC_NRELOCS(isec); irel++ )
        {
            PrintRELOC2(fp,&ASEC_AREL(isec,irel));
        }
    }

    PRINT_FUN_END1();
}

/*=======================================================
=======================================================*/
#if 0
#undef __PROC__
#define __PROC__ "PrintAllFdr"

void PrintAllFdr(FILE *fp)
{
    INDEX ifdr;

    PRINT_FUN_BEG1();

    FOREACH_FDR(ifdr)
    {
        PrintFdr(fp,ifdr);
        PrintNL(fp);
    }
    PRINT_FUN_END1();
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAllPdr"

void PrintAllPdr(FILE *fp)
{
    INDEX ipdr;

    PRINT_FUN_BEG1();

    FOREACH_PDR(ipdr)
    {
        PrintPdt(fp,ipdr);
        PrintNL(fp);
    }
    PRINT_FUN_END1();
}
#endif
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAllSym"

extern void PrintAllSym(FILE *fp)
{
    INDEX isym;

    PRINT_FUN_BEG1();

#ifdef COFF
    fprintf(fp,"\nLOCALS:\n\n");
    
    FOREACH_LSY(isym)
    {
        PrintLsy(fp,isym);
    }

    fprintf(fp,"\nEXTERNALS:\n\n");
        
    FOREACH_ESY(isym)
    {
        PrintEsy(fp,isym);
    }
#endif

#ifdef ELF
    FOREACH_SYM(isym) {
	PrintSym(fp, &SYM(isym), &BIN_STRING_TAB[SYM(isym).st_name]);
    }
#endif
	    
    PRINT_FUN_END1();
}


