/*=======================================================
=======================================================*/

#include <stdio.h>
#include <stdlib.h>

#define FILE_PDF 0x0a213e4644503c21L
#define FILE_ADR 0x0f0e0010L
#define FILE_CNT 0x0f0e0a11L
#define FILE_CNT2 0x0f0e0a33L



/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ComputeReadProfilingInformation"

#define USE_NEW_ATOM_PIXIE 1
typedef unsigned char  UINT8;
typedef   signed char  INT8;
typedef unsigned short UINT16;
typedef   signed short INT16;
typedef unsigned int   UINT32;
typedef   signed int   INT32;
typedef unsigned long  UINT64;
typedef   signed long  INT64;

typedef INT8  BOOL;
typedef INT32 INDEX;
typedef INT32 NUMBER;
typedef INT64 ADDRESS;
typedef char *STRING;
typedef UINT16 FLAGS;
typedef INT8  DATATYPE;
typedef INT8  EDGETYPE;
typedef UINT64 RESLIST;
typedef INT8  RES;
typedef INT32    LATENCY;
typedef UINT8   PIPELINES;
typedef UINT64 REGLIST;
#define REGSET REGLIST
typedef INT8  REG;



#define CONFIG_NINDIR       20
#define ALTO_MAGIC 0xcaffe

typedef struct {
    ADDRESS addr;
    ADDRESS count;
} indir;


typedef struct {
    INT64 iins;
    INDEX ins;
    REG reg1;
    REG reg2;
    INT64 totalcount;
    indir values[CONFIG_NINDIR];
} Indir;



Indir *aIndir;
NUMBER aIndirFuns;
Indir *aValues;
NUMBER aValueCount;

/*=======================================================
=======================================================*/

BOOL flag_important_only =0;
BOOL flag_floats = 0;
double cut_off_percentile = 10.0;
INT64 cut_off_count = 0;
/*=======================================================
=======================================================*/

static int cmp(const void *a, const void *b)
{
    indir *v1 = (indir *)a;
    indir *v2 = (indir *)b;

    return (v2->count - v1->count);
}



/*=======================================================
 we want to have the value profile create an ascii file
 so we can look at the values used 
=======================================================*/
#undef __PROC__
#define __PROC__ "ComReadValueProfilingInformation"


extern void ComReadValueProfilingInformation(FILE *fp, char *CONFIG_INFILE)
{
    FILE *fp_adr;
    FILE *valprof;
    char fn_adr[256];
    INT64 cnt;
    INT64 bcount = 0;
    INT64 ecount = 0;
    INT64 fcount = 0;
    INT64 mapcount = 0;
    INT64 totalcount;
    int i;
    int j;
    int temp;
    INT64 counter;


    /*
      open address file and advance file pointer
      to beginning of address data
      */
    
    strcpy(fn_adr,CONFIG_INFILE);
    
    fp_adr = fopen(fn_adr,"r");
    if( fp_adr==0 )
    {
        
        fprintf(stderr,"Cannot open adr file: %s, no value profiles\n",fn_adr);
        return;
    }
    
    fread(&cnt,sizeof(cnt),1,fp_adr);


    if ( cnt != ALTO_MAGIC ) {
	return;
    }


    fread(&bcount,sizeof(bcount),1,fp_adr);
    fread(&ecount,sizeof(ecount),1,fp_adr);
    fread(&fcount,sizeof(fcount),1,fp_adr);
    fread(&mapcount,sizeof(mapcount),1,fp_adr);
    
    printf("using alto profiling, checking counters\n");
    printf("number of bbl counters is %ld\n", bcount);
    printf("number of edg counters is %ld\n", ecount);
    printf("number of fun counters is %ld\n", fcount);
    printf("number of iins map is %ld\n", mapcount);
    
    if (bcount != 0) {
	fprintf(stderr,"bbl count is not zero in value profile\n");
    }
    
    if (ecount != 0) {
	fprintf(stderr,"Edg count is not zero in value profile\n");
    }
    
    aValues = malloc(sizeof(Indir) * fcount);
    aValueCount = fcount;
    
    sprintf(fn_adr,"%s.valprof",CONFIG_INFILE);
    
    valprof = fopen(fn_adr,"w");
    if (valprof==0) 
	fprintf(stderr, "Cannot open valprof file: %s\n",fn_adr);
    
    
    for (i=0; i < fcount; i++) {
	fread(&counter, sizeof(counter), 1, fp_adr);
	/* bottom half of the word is the instruction, top is register */
	aValues[i].iins = counter;

	aValues[i].reg1 = (counter >> 32) & 0xffff;
	aValues[i].reg2 = (counter >> 48) - 1;
	aValues[i].ins=  counter & 0xffffffff;

	fread(&counter, sizeof(counter), 1, fp_adr);
	
	totalcount = 0;
	for (j=0; j < CONFIG_NINDIR; j++) {
	    fread(&counter, sizeof(counter), 1, fp_adr);
	    aValues[i].values[j].addr = counter;
	    fread(&counter, sizeof(counter), 1, fp_adr);
	    aValues[i].values[j].count = counter;
	    totalcount += counter;
	}
	aValues[i].totalcount = totalcount;

	qsort(aValues[i].values, CONFIG_NINDIR - 1, sizeof(indir), &cmp);
    }
    
    for (temp=0; temp < fcount; temp++)
    {
        if( aValues[temp].totalcount < cut_off_count ) continue;
        
        printf("======================================================================\n");
        printf("id: %20ld=0x%016lx  cnt: %ld  reg1: %2d  reg2 %2d\n", 
               aValues[temp].iins, aValues[temp].iins, aValues[temp].totalcount,aValues[temp].reg1, aValues[temp].reg2);
        if( flag_important_only )
            if( 100.0*aValues[temp].values[0].count/(aValues[temp].totalcount+1) < cut_off_percentile )
            {
                printf("UNINTERRSTING\n");
                continue;
            }
        
	for (j=0; j < CONFIG_NINDIR; j++)
        {
            if( aValues[temp].values[j].count == 0 ) continue;
	    printf("val: %16lx  (%f) cnt: %20ld (%4.1f%%)\n",
                   aValues[temp].values[j].addr, *(double*) &aValues[temp].values[j].addr, aValues[temp].values[j].count,
                   100.0*aValues[temp].values[j].count/(aValues[temp].totalcount+1));
	}
    }
    printf("======================================================================\n");
    
}

/*=======================================================
=======================================================*/

static void Usage()
{
    printf("usage: valprofdump  [-i] value-profiling-file\n"
           "\n"
           "Prints the contents of value-profiling-file\n"
           "\n"
           "-c num\tonly print items with totalcount > c\n"
           "-i\tonly print important entries\n");
}

/*=======================================================
=======================================================*/

extern int main(int argc, char *argv[])
{

    int i;
    
    while( EOF != (i = getopt(argc,argv,"ihc:")) )
    {
        switch(i)
        {
        case 'c':
            cut_off_count = atoi(optarg);
            break;
        case 'i':
            flag_important_only = 1;
            break;
        case 'f':
            flag_floats = 1;
            break;
        case 'h':
        case '?':
        default:
            Usage();
            break;
        }
    }

    argv += optind;     
    argc -=optind;

    ComReadValueProfilingInformation(stdout,argv[0]);
    return 0;
}


/*=======================================================
=======================================================*/





